\2048.xpl	16-Mar-2018
\Yet another version of Gabriele Cirulli's 2048 game

int	Board(4,4), Score, Best, Moved;
def	Xc = (640/8  - 4*6)/2,	\text coordinates to center the displayed board
	Yc = (480/16 - 4*3)/2;
def	Esc=$1B, UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D; \scan codes


proc	DrawTile(X0, Y0, Power, Color);	\Draw a numbered tile or empty location
int	X0, Y0, Power, Color;	\character coordinates and tile's power of 2
int	X, Y, Att;
[\Set color for each tile (or empty) value
\      2^0  2^1  2^2  2^3  2^4  2^5  2^6  2^7  2^8  2^9  2^10 2^11
Att:= [$80, $70, $60, $5F, $4F, $3F, $2F, $1F, $C0, $D0, $E0, $F0];
Attrib(Att(Power));
for Y:= Y0 to Y0+2 do
	[Cursor(X0+Xc, Y+Yc);  Text(6, "      ")];
if Power > 0 then			\display tile's number
	[Cursor(X0+1+Xc, Y0+1+Yc);
	if Power <= 9 then ChOut(6, ^ );\center numbers, somewhat
	if Power <= 3 then ChOut(6, ^ );
	IntOut(6, 1<<Power);		\2^Power
	];
X:= (X0+Xc)*8;  Y:= (Y0+Yc)*16;		\outline tile with Color
Move(X, Y);
Line(X+6*8-1, Y, Color);
Line(X+6*8-1, Y+3*16-1, Color);
Line(X, Y+3*16-1, Color);
Line(X, Y, Color);
if Color # 0 then Sound(0, 3, 1);	\delay if not black (0)
];	\DrawTile


proc	MoveTiles(X0, Y0, DX, DY);	\Shift tiles, add adjacents, shift again
int	X0, Y0, DX, DY;
int	M, N, X, Y;

	proc	Shift;			\Shift all tiles in a single row or col.
	[for M:= 1 to 3 do
	    [X:= X0;  Y:= Y0;
	    for N:= 1 to 3 do
		[if Board(X,Y)=0 & Board(X+DX,Y+DY)#0 then
		   [Board(X,Y):= Board(X+DX,Y+DY);  Board(X+DX,Y+DY):= 0;
		   Moved:= true];
		X:= X+DX;  Y:= Y+DY;
		];
	    ];
	];	\Shift

[Shift;
X:= X0;  Y:= Y0;		\add identical adjacent tiles into a new tile
for N:= 1 to 3 do
	[if Board(X,Y)=Board(X+DX,Y+DY) & Board(X,Y)#0 then
		[Board(X,Y):= Board(X,Y)+1;  Board(X+DX,Y+DY):= 0;
		Score:= Score + 1<<Board(X,Y);
		DrawTile(X*6, Y*3, Board(X,Y), $F \BWhite\);
		Moved:= true];
	X:= X+DX;  Y:= Y+DY;
	];
Shift;
];	\MoveTiles


proc	PlayGame;		\Play one game
int	X, Y, C;
[SetVid($12);			\set 640x480 graphics with 16 colors
for Y:= 0 to 3 do		\initialize board to empty
    for X:= 0 to 3 do
	Board(X,Y):= 0;
Score:= 0;
Moved:= true;			\force first tile to be inserted

loop	[for Y:= 0 to 3 do	\is there an empty location?
	    for X:= 0 to 3 do
		if Board(X,Y) = 0 then Y:= 5;
	if Y = 4 then		\no empty location found
		[Cursor(Xc, 4*3+Yc);  Text(0, "Game Over");  quit];
	if Moved then		\insert a new tile (2^1=2 or 2^2=4)
		[repeat X:= Ran(4);  Y:= Ran(4) until Board(X,Y) = 0;
		Board(X,Y):= if Ran(10) then 1 else 2;
		DrawTile(X*6, Y*3, Board(X,Y), $F \BWhite\);
		];

	for Y:= 0 to 3 do	\draw board with its tiles
	    for X:= 0 to 3 do
		DrawTile(X*6, Y*3, Board(X,Y), 0);

	Attrib($07);		\display Score and legends
	Cursor(60, 0);  Text(6, "Score");
	Cursor(60, 1);  IntOut(6, Score);
	Cursor(70, 0);  Text(6, "Best");
	if Score > Best then Best:= Score;
	Cursor(70, 1);  IntOut(6, Best);
	Cursor(60, 28);  Text(6, "^{  ^x  ^y  ^z  Esc");

	for Y:= 0 to 3 do	\search for winning location (2^11 = 2048)
	    for X:= 0 to 3 do
		if Board(X,Y) = 11 then
		    [Cursor(Xc, 4*3+Yc);  Text(0, "You Win!");  quit];

	Moved:= false;
	repeat C:= ChIn(1) until C#0;	\ignore scan code prefix byte
	case C of
	  LtArrow: for Y:= 0 to 3 do MoveTiles(0, Y, +1, 0);
	  RtArrow: for Y:= 0 to 3 do MoveTiles(3, Y, -1, 0);
	  DnArrow: for X:= 0 to 3 do MoveTiles(X, 3, 0, -1);
	  UpArrow: for X:= 0 to 3 do MoveTiles(X, 0, 0, +1);
	  Esc, $03, ^q, ^Q: [SetVid($03);  exit] \Ctrl+C also restores text mode
	other	[];
	];	\loop
X:= ChIn(1);			\wait for keystroke before starting next game
];	\PlayGame


[TrapC(true);			\prevent Ctrl+C from not restoring text mode (3)
Best:= 0;
loop PlayGame;
]
