\btc.xpl	13-Aug-2016	loren.blaney@gmail.com
\Block Transfer Computations
\Modified for RPi.

inc	codesr;			\intrinsic declarations

def	VMode = $105,		\video mode: VESA (256-color) graphics
	ImW=1024, ImH=768;	\graphic screen dimensions (pixels)

inc	graphics.inc;		\routines and definitions for graphics

def	FloorY = ImH-20;	\position of floor that block slides along

def	FlyR = 200,		\radius of flywheel
	FlyX=FlyR+20,		\position of center coordinates
	FlyY=ImH/2;

def	PullR = 100,		\radius of pulley wheels (pixels)
	PullX = ImW-3*PullR-20,	\position of center of top pulley wheel
	PullY = PullR+20;

def	BlkW=100, BlkH=100;	\block dimensions

def	HopX=ImW/2-55, HopY=-160, \hopper position and dimensions
	HopW=BlkW+6, HopH=ImH-BlkH-23;

int	BlkX,			\horiz position of left edge of (sliding) block
	PokeX, PokeY,		\position of right end of poker
	StackY;			\position of bottom of stack of blocks in hopper
real	Rotation,		\increasing flywheel rotation angle (radians)
	RopeTug;		\amount that rope is pulled by flywheel

def	Phase = 1.9;		\Rotation offset where block starts sliding



proc	DrawBlock(X0, Y0);	\Draw a block with upper-left corner at X0,Y0
def	C = BWhite;		\color
int	X0, Y0;
begin
DrawRectangle(X0, Y0, BlkW, BlkH, C);
end;	\DrawBlock



proc	DrawHopper;		\Draw the hopper filled with blocks
def	C = BWhite;		\color
int	N;
begin
MovePen(HopX, 0);		\draw sides of hopper
DrawLine(HopX, HopH-1, C);
MovePen(HopX+HopW-1, 0);
DrawLine(HopX+HopW-1, HopH-1, C);

if BlkX>=HopX+3 & BlkX<=HopX+BlkW+3 then
	begin			\block is under stack of blocks in hopper
	StackY:= HopH+2;	\stack is above block (at bottom of hopper)
	end
else if PokeX >= HopX+3 then
	begin			\poker is under hopper
	StackY:= StackY+2;	\fall to poker
	if StackY >= PokeY then StackY:= PokeY-1;
	end
else	begin			\poker is not under hopper
	StackY:= StackY+2;	\fall to floor
	if StackY > FloorY then StackY:= FloorY;
	end;

for N:= 1, 8 do			\draw blocks in hopper
	DrawBlock(HopX+3, StackY-(BlkH+1)*N+1);
end;	\DrawHopper



proc	DrawWeight(X, Y, C);	\Draw trapezoid representing the weight
int	X, Y,			\top center coordinates
	C;			\color
def	W1=100, W2=200, H=100;	\width of top, bottom, and height
begin
MovePen(X, Y);
DrawLine(X-W1/2, Y, C);
DrawLine(X-W2/2, Y+H, C);
DrawLine(X+W2/2, Y+H, C);
DrawLine(X+W1/2, Y, C);
DrawLine(X, Y, C);
end;	\DrawWeight



proc	DrawWheel(X0, Y0, R, Rot, C); \Draw a 3-hole pulley wheel
int	X0, Y0, R;		\center coordinates, radius
real	Rot;			\rotation (radians)
int	C;			\color
int	I, X, Y;
begin
DrawCircle(X0, Y0, R, C);	\rim
DrawCircle(X0, Y0, 7, C);	\axle

for I:= 0, 3-1 do		\draw holes in wheel
	begin
	X:= X0 + fix(float(R/2)*Cos(Rot));
	Y:= Y0 + fix(float(R/2)*Sin(Rot));
	DrawCircle(X, Y, R/4, C);
	Rot:= Rot + Pi/1.5;
	end;
end;	\DrawWheel



proc	DrawPulley;		\Draw pulley mechanism
int	X, Y;
real	Rot;
begin
Rot:= RopeTug/float(PullR);		\rotation of wheel (radians)

DrawWheel(PullX, PullY, PullR, -Rot, BWhite);	\draw top wheel
MovePen(PullX, PullY);				\draw its support
DrawLine(PullX, 0, BWhite);

Y:= PullY + 353 + fix(RopeTug)/3;	\vertical posn of rope at bottom wheel
MovePen(PullX-PullR, PullY);		\draw rope down left side
DrawLine(PullX-PullR, Y, BWhite);

DrawWheel(PullX, Y, PullR, -2./3.*Rot, BWhite); \draw lower wheel

MovePen(PullX+PullR, Y);		\draw rope up wheel's right side
DrawLine(PullX+PullR, PullY+PullR+20, BWhite);

X:= PullX + 2*PullR;			\draw middle wheel
DrawWheel(X, PullY+PullR+20, PullR, Rot/3., BWhite);

MovePen(X, PullY+PullR+20);		\draw middle wheel support
DrawLine(X, 0, BWhite);

MovePen(X+PullR, PullY+PullR+20);	\draw rope down far right side
DrawLine(X+PullR, Y+PullR+20, BWhite);

MovePen(PullX, Y);			\draw rope holding bar
DrawLine(PullX, Y+PullR+20, BWhite);

MovePen(PullX-20, Y+PullR+20);		\draw horizontal bar
DrawLine(X+PullR+20, Y+PullR+20, BWhite);

X:= (PullX + PullR);			\draw rope holding weight
MovePen(X, Y+PullR+20);
DrawLine(X, Y+PullR+40, BWhite);

DrawWeight(X, Y+PullR+40, BWhite);	\draw weight

if Cos(Rotation+Phase) < 0.0 then	\block is no longer moving right
    if Cos(Rotation+Phase+0.9) < 0.0 then \ and block is moving downward
	DrawBlock(X-50, Y+PullR+40+102);
end;	\DrawPulley



proc	DrawPivotArm;		\Draw pivot arm
def	D = 330,		\distance of pivot from center of flywheel
	L = 680,		\length of pivot arm
	S = 80.0,		\distance of slider from axle
	X0=220, Y0=ImH/2-D,	\pivot point
	C = BWhite;		\color
def	Poker = 368;		\poker length
int	X, Y, Z;
begin
X:= fix(S*Cos(Rotation));
Y:= fix(S*Sin(Rotation));
Z:= sqrt(sq(X) + sq(Y+D));

\220, ImH/2 = center coordinates of flywheel
DrawCircle(X+220, Y+ImH/2, 7, C); \sliding pivot

X:= L*X/Z + X0;
PokeY:= L*(Y+D)/Z + Y0;
PokeX:= X + Poker;		\right end of poker

DrawCircle(X0, Y0, 7, C);	\pivot at top of arm
MovePen(X0, Y0);
DrawLine(X, PokeY, C);
DrawLine(PokeX, PokeY, C);	\draw poker
DrawCircle(X, PokeY, 7, C);	\pivot on poker

BlkX:= HopX+BlkW+10;		\out of the way, to the right of the hopper
if Cos(Rotation+Phase) >= 0.0 then
	begin			\block is moving to the right
	BlkX:= X+Poker;
	if BlkX > HopX+3 then
		DrawBlock(BlkX, FloorY-BlkH);
	end;
end;	\DrawPivotArm



proc	Tangent(X0, Y0, R, X1, Y1, AX2, AY2);
\Return the lower (larger Y value) tangent point (X2,Y2) of a line drawn from
\ an external point (X1,Y1) to a circle centered at X0,Y0 with radius R.
real	X0, Y0, R, X1, Y1, AX2, AY2;
real	X, Y, K, A, B, C, D, E;
begin
K:= (R*R + X1*X0 - Y0*Y0 + Y1*Y0 - X0*X0) / (Y1-Y0);
A:= (X0-X1) / (Y1-Y0);
B:= 2.*A*K - 2.*X0 -2.*A*Y0;
C:= X0*X0 - 2.*K*Y0 + K*K + Y0*Y0 - R*R;
D:= sqrt(B*B - 4.*(A*A+1.)*C);
E:= 2.*(A*A+1.);

AX2(0):= (-B + D) / E;
AY2(0):= A*AX2(0) + K;

X:= (-B - D) / E;
Y:= A*X + K;
if Y > AY2(0) then		\use solution with larger Y
	begin
	AX2(0):= X;
	AY2(0):= Y;
	end;
end;	\Tangent



proc	DrawPullRope;
def	Stroke = float(FlyR),	\amount of +/- pull on rope (pixels)
	Phi = -4./5.*Pi,	\attach rope to a flywheel spoke
	RopeLen = 500.0,
	C = BWhite;		\color
real	RopeX, RopeY,		\position of pulled end of rope
	TanX, TanY,		\tangent point on pulley wheel
	Dist;			\distance from tan point to end of rope
begin
RopeX:= Stroke*Cos(Rotation+Phi) + float(FlyX);	\rope at flywheel
RopeY:= Stroke*Sin(Rotation+Phi) + float(FlyY);

Tangent(float(PullX), float(PullY), float(PullR), RopeX, RopeY,
	 addr TanX, 0, addr TanY, 0);

MovePen(fix(RopeX), fix(RopeY));
DrawLine(fix(TanX), fix(TanY), C);
DrawCircle(fix(RopeX), fix(RopeY), 7, C);

Dist:= sqrt(sq(TanX-RopeX) + sq(TanY-RopeY));
RopeTug:= RopeLen - Dist;
end;	\DrawPullRope



proc	DrawFlyWheel;		\Draw a flywheel
def	R1 = 30,		\radius of flywheel's hub
	C = BWhite;		\color
int	I, X, Y;
real	Rot;
begin
DrawCircle(FlyX, FlyY, FlyR, C);\rim
DrawCircle(FlyX, FlyY, R1, C);	\hub
DrawCircle(FlyX, FlyY, 7, C);	\axle

Rot:= Rotation;
for I:= 0, 5-1 do		\draw spokes
	begin
	X:= FlyX + fix(float(FlyR)*Cos(Rot));
	Y:= FlyY + fix(float(FlyR)*Sin(Rot));
	MovePen(X, Y);
	X:= FlyX + fix(float(R1)*Cos(Rot));
	Y:= FlyY + fix(float(R1)*Sin(Rot));
	DrawLine(X, Y, C);
	Rot:= Rot + 2.*Pi/5.;
	end;
DrawPullRope;
end;	\DrawFlyWheel

\-------------------------------------------------------------------------------

begin	\Main
SetVid(VMode);			\set graphic mode

Rotation:= 0.0;			\initialize
StackY:= HopH;
BlkX:= HopX+BlkW+10;		\out of the way, to the right of the hopper
repeat	Erase(Blue);		\clear Screen buffer
	DrawFlyWheel;
	DrawPivotArm;
	DrawPulley;		\draw everything in Screen buffer
	DrawHopper;
	Rotation:= Rotation + 0.05; \rotate flywheel
	WaitVB;			\regulate speed
	Paint(0, 0, ImW, ImH, Screen, ImW); \copy buffer to display
until	KeyHit;			\terminate on keystroke

SetVid(3);			\restore normal text mode
end;	\Main
