\congame.xpl	20-Feb-2018
\Concentration Card Game - match pairs of cards
\
\Revisions:
\29-May-2017: Modified for RPi.
\20-Feb-2018: Modified to work with all color depths, e.g: Jessie's 32-bit color

int	Cards,			\number of cards remaining to be matched
	Deck(52),		\deck of cards, contains 0..51 and -1
	Mouse,			\info array from GetMouse intrinsic
	SecOld,			\updates time display when seconds change
	Time0;			\time at start of game (microseconds)

def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,  \attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite;



proc	Exit;			\Make a clean exit
[SetVid($03);			\restore normal text mode
exit;
];	\Exit



proc	Int2Out(Dev, N);	\Output 2-digit positive integer
int	Dev, N;
[if N < 10 then ChOut(Dev, ^0);	\leading zero
IntOut(Dev, N);
];	\Int2Out



proc	TimeOut(Dev, Time);	\Output Time, e.g: 05:43:03
int	Dev, Time;
int	Hour, Min, Sec;
[if Time < 0 then return;
Time:= Time/60;
Sec:= rem(0);
Hour:= Time/60;
Min:= rem(0);
Int2Out(Dev, Hour);
ChOut(Dev, ^:);
Int2Out(Dev, Min);
ChOut(Dev, ^:);
Int2Out(Dev, Sec);
];	\TimeOut



proc	TickTock;		\Display time ticking away
int	Timer;
[Timer:= (GetTime - Time0) / 1000000;	\works for 35 minutes
if Timer # SecOld then		\seconds have changed
	[SecOld:= Timer;
	ShowMouse(false);
	Cursor(35, 13);
	Attrib(BWhite<<4 + Black);
	TimeOut(6, Timer);
	ShowMouse(true);
	];
];	\TickTock



proc	ShowCard(X, Y, Card);	\Show Card at char coordinates X,Y
int	X, Y, Card;
int	Num, Suit;
char	Str;

	proc	CardStock(Ch);	\Draw the basic card stock pattern for front
	int	Ch;		\ or back side
	int	I, J;
	[for J:= 0 to 2 do
		[Cursor(X, Y+J);
		for I:= 0 to 5-1 do ChOut(6, Ch);
		];
	];	\CardStock

[if Card < 0 then		\display the back of the card
	[Attrib(BWhite<<4 + Blue);
	CardStock($B1);
	return;
	];

Str:= "234567891JQKA ";		\else display front of card
Num:= Card/4;
Suit:= Rem(0);

Attrib(BWhite<<4 + (if Suit<2 then LRed else Black));
CardStock(^ );

Cursor(X+2, Y+1);		\show suit symbol in center of card
ChOut(6, Suit+3);

Cursor(X, Y);			\show card number, at top corner
ChOut(6, Str(Num));

if Num = 10-2 then		\"10" requires two digits
	[ChOut(6, ^0);
	Cursor(X+3, Y+2);	\also show number at bottom corner
	Text(6, "10");
	]
else	[Cursor(X+4, Y+2);	\show single digit at bottom corner
	ChOut(6, Str(Num));
	];
];	\ShowCard

\-------------------------------------------------------------------------------

proc	Play;			\Play the game
int	MX, MY,			\mouse char coordinates
	N, X, Y,		\card number and card coordinates
	C1, C1X, C1Y,		\first selected card
	C2, C2X, C2Y;		\second selected card


func	GetCard;
\Given mouse coordinates MX,MY, return card N at card position X,Y
[N:= 0;
for Y:= 0 to 6-1 do		\for all card positions...
    for X:= 0 to 10-1 do
	[if Y#2 & Y#3 ! (X<3!X>6) then		\skip center
		[if MX<5+X*7+6 & MY<1+Y*4+4 then return;
		N:= N+1;
		];
	];
N:= 52-1;			\don't go beyond last card
X:= 9;  Y:= 5;			\return last card
];	\GetCard



proc	PickCard;		\Click on a card and turn it over
[ShowMouse(true);
repeat	MoveMouse;
	Mouse:= GetMouse;
until	Mouse(2) = 0;		\wait for button release
repeat	if KeyHit then Exit;	\any keystroke terminates prog.
	TickTock;
	MoveMouse;
	Mouse:= GetMouse;	\wait for button press
until	Mouse(2);
MX:= Mouse(0) / 8;		\convert pixel coordinates to character cells
MY:= Mouse(1) / 16;
ShowMouse(false);

GetCard;
if Deck(N) >= 0 then ShowCard(5+X*7, 1+Y*4, Deck(N));	\turn the card face up
];	\PickCard



[	\Play
PickCard;
C1:= N;  C1X:= X;  C1Y:= Y;
PickCard;
C2:= N;  C2X:= X;  C2Y:= Y;

if Deck(C1)/4 = Deck(C2)/4  &  C1 # C2  &  Deck(C1) >= 0  &  Deck(C2) >= 0 then
	[			\face values match and they aren't the same card
	Cards:= Cards-2;	\ and they have not already been matched
	Deck(C1):= -1;		\mark cards as matched (left face up)
	Deck(C2):= -1;
	]
else	[			\cards don't match (or illegal move)
	Sound(0, 36, 1);	\delay then turn cards face down
	if Deck(C1) >= 0 then ShowCard(5+C1X*7, 1+C1Y*4, -1);
	if Deck(C2) >= 0 then ShowCard(5+C2X*7, 1+C2Y*4, -1);
	];
];	\Play

\-------------------------------------------------------------------------------

proc	Deal;			\Lay out all cards facing down
int	X, Y;			\card coordinates
[for Y:= 0 to 6-1 do
    for X:= 0 to 10-1 do
	if Y#2 & Y#3 ! (X<3!X>6) then
		ShowCard(5+X*7, 1+Y*4, -1);
];	\Deal



proc	Shuffle;		\Shuffle the card deck
int	I, J, K, T;
[for J:= 0 to 10-1 do		\go thru deck 10 times
    for I:= 0 to 52-1 do	\for all the cards...
	[K:= Ran(52);		\randomly pick a card
	T:= Deck(I);		\swap cards in the deck at locations I and K
	Deck(I):= Deck(K);
	Deck(K):= T;
	];
];	\Shuffle



proc	Init;			\Initialize
int	I, J;
char	Str;
[SetVid($12);
for I:= 0 to 52-1 do Deck(I):= I;	\set up the deck
Shuffle;
Cards:= 52;

Attrib(Green<<4 + White);	\paint the background green
for J:= 0 to 25-1 do
	[Cursor(0, J);
	for I:= 0 to 80-2 do Chout(6, ^ );
	];

Str:= "CONCENTRATION ";
for I:= 0 to 13-1 do
	[Cursor(26+I+I, 10);
	Attrib(Green<<4 + ((I+4)&$0F));
	ChOut(6, Str(I));
	];

Time0:= GetTime;		\get starting time
SecOld:= -1;			\force immediate display of timer
];	\Init

\-------------------------------------------------------------------------------

[	\Main
Init;
Deal;
repeat Play until Cards <= 0;

\Show winning screen before exiting
ShowMouse(true);
repeat	MoveMouse;
	Mouse:= GetMouse;
until	Mouse(2) = 0;		\wait for button release
repeat	MoveMouse;		\wait for button or keystroke
	Mouse:= GetMouse;
until	Mouse(2) or KeyHit;
Exit;
]	\Main
