\sch.xpl	30-Nov-2017
\Instruction scheduler. This moves ldr r_, instructions to the beginning of a
\ basic block. Moved instructions are more likely to execute in a single cycle
\ rather than in 3 or 4 cycles, which is the case when the loaded register is 
\ used in the immediately following instruction.
\
\This program is free software; you can redistribute it and/or modify it under
\ the terms of the GNU General Public License version 2 as published by the
\ Free Software Foundation.
\This program is distributed in the hope that it will be useful, but WITHOUT
\ ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
\ FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
\ details.
\You should have received a copy of the GNU General Public License along with
\ this program (in the file LICENSE.TXT); if not, write to the Free Software
\ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\
\The authors can be reached at: loren.blaney@gmail.com
\
\REVISIONS:
\ 12-Aug-2016: Version 1.0 release.
\ 30-Jun-2017: Version 2.1 release. Only the version number is changed.
\ 30-Nov-2017: Version 2.2 release. Reverse order of arguments in StrFind.

string	0;			\enable null-terminated string convention
int	AsmLine(50000, 80),	\lines of assembly code for a basic block
	PLocList(50000);	\list of used (pseudo) memory locations
def	OD = 3;			\output device number
def	LF=$0A, CR=$0D, EOF=$1A;

\-------------------------------------------------------------------------------

func	StrCopy(A, B);		\Copy ASCIIZ string: A <- B
char	A, B;			\WARNING: Make sure string A has adequate space
int	I;			\(Don't want any infamous buffer overflows)
[for I:= 0 to -1>>1 -1 do
	[A(I):= B(I);
	if B(I) = 0 then return A;
	];
];	\StrCopy



func	StrLen(A);		\Return number of characters in an ASCIIZ string
char	A;
int	I;
for I:= 0 to -1>>1 -1 do
	if A(I) = 0 then return I;



func	StrFind(A, B);		\Search for ASCIIZ string B in string A
\Returns address of first occurrence of string B in A, or zero if B is not found
char	A, B;			\strings to be compared
int	LA, LB, I, J;
[LA:= StrLen(A);
 LB:= StrLen(B);
 for I:= 0 to LA-LB do
    [for J:= 0 to LB-1 do
	if B(J) # A(J+I) then J:= LB+1;
    if J = LB then return A+I;	\found
    ];
return 0;
];	\StrFind



func	ReadLine(LineStr, SizeLimit);	\Read line of text from the input device
char	LineStr;			\ and return it as a string in LineStr.
int	SizeLimit;			\Return 'false' when EOF is reached.
int	I, C;
[I:= 0;						\read one line from file
loop	[repeat C:= ChIn(3) until C#CR;		\strip off carriage return, & LF
	if C=EOF then [LineStr(I):= 0;  return false];
	if C=LF then C:= 0;			\terminate string
	LineStr(I):= C;
	if C = 0 then return true;
	if I < SizeLimit-1 then I:= I+1;	\leave room for terminating zero
	];
];	\ReadLine

\-------------------------------------------------------------------------------

proc	OpenFiles;	\Open input and output files specified on command line
int	InHand,		\input file handle
	OutHand,	\output file handle
	I, Ch;
char	CmdTail;

	proc	Parse;		\Parse command tail and set I/O handles
	char	ExtIn, ExtOut;
	int	P, ExtFlag, I;
	[ExtIn:= ".s";   ExtOut:= ".ss";
	P:= 1;
	ExtFlag:= false;
	loop	[if CmdTail(P) = ^. then  \parse command tail for extension
			[ExtFlag:= true;  quit];
		if CmdTail(P) = ^; then quit;	\ignore semicolon
		if P > CmdTail(0) then quit;	\if quit then P points to 0
		P:= P+1;
		];
	\Back up over any trailing spaces (caused by switches)
	repeat P:= P-1 until CmdTail(P) # ^ ;
	P:= P+1;
	if not ExtFlag then			\append input file extension
		for I:= 0 to 2 do CmdTail(P+I):= ExtIn(I);
	InHand:= FOpen(CmdTail+1, 0);
	for I:= 0 to 3 do CmdTail(P+I):= ExtOut(I);
	OutHand:= FOpen(CmdTail+1, 1);
	];	\Parse


[CmdTail:= Reserve($80+4);
I:= 1;				\get file name from command line
repeat	Ch:= ChIn(8);
	CmdTail(I):= Ch;
	I:= I+1;
until	Ch=0 ! Ch=EOF;
CmdTail(0):= I-2;	\number of characters in CmdTail not including last 0

Parse;					\parse command line and set file handles
FSet(InHand, ^I);
FSet(OutHand, ^O);
];	\OpenFiles

\-------------------------------------------------------------------------------

proc	DropComment(Str);	\Chop off any comment or .ascii string from Str
char	Str;
char	P;
[for P:= Str to Str+1000 do
	case P(0) of 0, ^@, ^": [P(0):= 0;  return]
	other [];
];	\DropComment



func	PLoc(Str);		\Convert asm string to a pseudo memory location
char	Str;
int	Reg, Base, Off;
[OpenO(8);
Text(8, Str);
OpenI(8);
Reg:= IntIn(8);			\str r0, [r11, #4]
Base:= IntIn(8);		\str r0, [r1]
Off:= IntIn(8);			\IntIn returns 0 if EOF is reached
return Base<<16 + Off;
];	\PLoc



func	Okay(I);
\Return false if memory location in present line is = to any previous lines.
\This must detect duplicate memory locations, but false positives are okay,
\ such as: ldr r12, base1   ldr r2, [r12]   ldr r12 base2   ldr r3 [r12]
\ r2 and r3 will appear to be loaded from the same address, but that's okay.
int	I;		\index to present line
int	L, J;
[L:= PLoc(AsmLine(I));
for J:= 0 to I-1 do
	if L = PLocList(J) then return false;
return true;
];	\Okay



func	GetRegs(Str);		\Return bit array of registers used in string
char	Str;
int	I, Reg, RegBits;
[RegBits:= 0;
for I:= 0 to 79 do
	[if Str(I) = 0 then return RegBits;
	 if Str(I)=^R then [Text(0, "Error: Uppercase R");  CrLf(0);  exit];
	 if Str(I)=^r & Str(I+1)>=^0 & Str(I+1)<=^9 then	\found register
		[Reg:= Str(I+1) - ^0;
		if Str(I+2)>=^0 & Str(I+2)<=^9 then
			Reg:= Reg*10 + Str(I+2) - ^0;
		RegBits:= RegBits ! 1<<Reg;
		];
	];
Text(0, "Error: Line too long
");
exit;
];	\GetRegs

\-------------------------------------------------------------------------------

int	HaveEOF,		\flag: end-of-file has been reached
	LineNum,		\input assembly code line number
	RegsUsed,		\bit array of registers used by prior code
	Regs,			\bit array of registers used on current line
	LockMask,		\bit array of registers preventing ldr's moving
	Rd,			\register destination number for ldr instruction
	L;
char	AsmStr(80*4),		\copy of assembly line with comments removed
	P;

[Text(0, "
-- CODE SCHEDULER FOR XPL0, VERSION 2.2 --

");
OpenFiles;
OpenO(OD);  OpenI(3);
HaveEOF:= false;
RegsUsed:= -1;			\assume worst case to begin
repeat	LineNum:= 0;		\read in a basic block of asm code
	repeat			\read in a line of text and save it in a string
		P:= ReadLine(AsmLine(LineNum), 80*4);
		StrCopy(AsmStr, AsmLine(LineNum));	\remove comments from
		DropComment(AsmStr);			\ a working copy
		if P=0 then HaveEOF:= true
		else LineNum:= LineNum+1;
	until	HaveEOF or AsmStr(0) = ^K or		\starts with "K"
		StrFind(AsmStr, "	b") # 0;	\or branch instr

	\	    pls1 11
	\	    crp2 1098 7654 3210
	LockMask:= %0100_0000_1111_1111;
	for L:= 0 to LineNum-2 do	\ignore last line like K12: ldr r12...
		[StrCopy(AsmStr, AsmLine(L));		\remove comments from
		DropComment(AsmStr);			\ a working copy
		P:= StrFind(AsmStr, "ldr	r");
		if P#0 then \0123	45	have ldr Rd instruction
			[Rd:= P(5) - ^0;	\get register Rd's number
			if P(6)>=^0 & P(6)<=^9 then
				Rd:= Rd*10 + P(6) - ^0;
			]
		else	Rd:= 0;

		Regs:= GetRegs(AsmStr);
		if P#0 and
		   (1<<Rd & RegsUsed) = 0 and
		   (Regs & RegsUsed & LockMask) = 0 and
		   Okay(L) then
			[Text(OD, AsmLine(L));		\output original line
			Text(OD, "	@ *****");	\indicate line moved
			CrLf(OD);
			AsmLine(L, 0):= 0;		\delete the line
			]
		else	[LockMask:= LockMask ! 1<<Rd;	\unmoved line
			RegsUsed:= RegsUsed ! Regs;	\ record all regs used
			];
		PLocList(L):= PLoc(AsmStr);		\record any memory loc
		];
	StrCopy(AsmStr, AsmLine(L));		\if last line read in starts "K"
	DropComment(AsmStr);			\ then it's actually the first
	RegsUsed:= GetRegs(AsmStr);		\ line of the next block

	\Output all remaining (undeleted) lines
	for L:= 0 to LineNum-1 do		\write basic block of asm code
		[if AsmLine(L, 0) then		\line was not deleted
			Text(OD, AsmLine(L));
		CrLf(OD);
		];
until	HaveEOF;
Close(OD);
]
