\xplr.xpl	2-Mar-2018
\XPL0 COMPILER FOR THE RASPBERRY PI
\COPYRIGHT 1984-2018 P.J.R. BOYLE
\FLOATING POINT VERSION BY LOREN BLANEY
\IBM NATIVE LANGUAGE VERSION BY LARRY FISH
\
\This program is free software; you can redistribute it and/or modify it under
\ the terms of the GNU General Public License version 2 as published by the
\ Free Software Foundation.
\This program is distributed in the hope that it will be useful, but WITHOUT
\ ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
\ FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
\ details.
\You should have received a copy of the GNU General Public License along with
\ this program (in the file LICENSE.TXT); if not, write to the Free Software
\ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\
\You can reach the authors at: loren.blaney@gmail.com
\
\REVISIONS:
\26-Feb-2014: Modified for RPi. Based on XPLN.XPL version 3.0.
\21-Apr-2014: Automatically include 'code' declarations (at level -1).
\ Make inline asm work for push {r0}, etc.
\ 17-Feb-2015: Allow many more global variables by increasing range of HPI.
\ Increase heap space to 64 MB. Align heap to 16-byte boundary. Fix glitch
\ displayed for 'include' name when extension is given.
\04-Oct-2015: Fix @ address operator on indexed reals e.g: @A(3). Mask reserved
\ word hashes (e.g: CODSYM) to 16 bits.
\12-Aug-2016: Release version 1.0.
\29-May-2017: Version 2.0: Fix define sqrt, sq, rem. Check for integer too big.
\ Increase RLMAX to 1600. Check for valid calls to intrinsics.
\30-Jun-2017: Version 2.1: codesrx adds GetPalette (80).
\02-Mar-2018: Version 2.2: codesrx adds Floor, Ceil, Pow, and SetHexDigits.
\ Fix bug where in-line asm didn't always access local variables.

code	\THE REQUIRED INTRINSICS:
	ABS=0		REM=2		RESERVE=3	SWAP=4
	CHIN=7		CHOUT=8		CRLF=9		INTIN=10
	INTOUT=11	TEXT=12		OPENI=13	OPENO=14
	CLOSE=15	TRAP=17,	GETERR=22,	HEXOUT=27,
	FSET=24,	FOPEN=29,	FCLOSE=32,	FIX=50;
code real		FLOAT=49,	RLRES=46,	SQRT=53;

int	INHAND,		\INPUT FILE HANDLE
	OUTHAND;	\OUTPUT FILE HANDLE

int	COMFLG,		\PUT COMMENTS IN CODE
	IMMTYP,		\IMMEDIATE TYPE
	IMMVAL,		\IMMEDIATE VALUE
	LOCAL,		\LOCAL LABEL NUMBER
	OLDLEV,		\OLD LEVEL
	TOS;		\KEEPS TRACK OF WHERE TOS IS (STACKED, INREG, OR IMAGE)

def	STACKED,	\TOS IS ON THE STACK
	INREG,		\TOS IS IN THE REGISTER
	IMAGE;		\TOS HAS NOT BEEN GENERATED YET

char	FNAME;		\NAME PART OF INPUT FILE (EXCLUDING DRIVE AND PATHNAME)
			\(ASSIGNS UNIQUE NAMES TO PROCEDURE CODE SEGMENTS)
def	TV=0, KB=0, NULDEV=7, EOF=$1A, BEL=$07, EOL=$0A; \I/O STUFF
def	SYMAX=1600,	\SIZE OF THE SYMBOL TABLE
	SIGCHAR=16,	\NO. OF SIGNIFICANT CHARS IN AN IDENT
	HASHMSK=$FF,	\MASK FOR HASHES
	EMTPNT=-1,	\EMPTY SYMBOL POINTER
	BOXNUM=256,	\NUMBER OF BOXES

	RLMAX=1600,	\SIZE OF REAL-CONSTANT SYMBOL TABLE
	RLSIZE=8,	\NO. OF BYTES IN A REAL NUMBER
	INTSIZE=4,	\number of bytes in an integer
	QUITMAX=160;	\MAXIMUM NO. OF 'QUIT'S IN A 'LOOP'

int	ERRCNT	\ERROR COUNTER
	LSTDEV	\LISTING OUTPUT DEVICE NUMBER
	SRCDEV	\SOURCE INPUT DEVICE NUMBER
	BINDEV	\BINARY OUTPUT DEVICE NUMBER
	SEGCNT	\COUNTS AND CREATE UNIQUE SEGMENT NUMBER
	DEEPER	\TRUE WHEN NESTING SEGMENTS DEEPER
	CODFLG	\FLAG: TRUE IF CODE HAS BEEN GENERATED
	HASMAIN	\FLAG: THERE IS A STATEMENT IN THE MAIN PROCEDURE
	CONDITIONAL  \FLAG: CONDITIONAL COMPILE
	STRTERM	\flag: string termination (0=null, nonzero=MSB)
	CHAR	\CURRENT CHARACTER.  MOST OF THE TIME IT
		\ CONTAINS THE TERMINATOR OF THE CURRENT ATOM
	ATOM	\PRESENT ATOM DESCRIPTOR
		\CONTAINS RESERVED WORD HASH OR THE ASCII FOR
		\ A SPECIAL CHARACTER; 0 IF THE ATOM IS A
		\ CONSTANT OR AN IDENTIFIER
	ATYPE;	\PRESENT ATOM TYPE DESCRIPTOR
def	\ATYPE\ SPECIAL,IDENTIFIER,INTCON,REALCON;
char	IDENT;	\ARRAY--CURRENT IDENTIFIER NAME
int	HASH	\CURRENT IDENTIFIER HASH CODE
	LABCNT,	\LABEL COUNT (FROM NEWLAB)
	IATOM,	\VALUE OF CURRENT INTEGER CONSTANT
	REMAINDER; \value of remainder from constant integer division
real	RLATOM;	\REAL CONSTANT FROM PROC "RATOM"

int	IDTYPE;	\PRESENT IDENTIFIER TYPE DESCRIPTOR
def	UNDEF=0,	\UNDEFINED ID (NO. ORDER IS CRITICAL)
	ADDRVAR=1,	\ADDRESS VARIABLE ID (TYPE = INTEGER)
	INVAR=3,	\INTEGER VARIABLE ID (ODD NOS.=INTEGER)
	RLVAR=4,	\REAL VARIABLE ID
	INCON=5,	\INTEGER CONSTANT ID
	RLCON=6,	\REAL CONSTANT ID
	INPROC=7,	\INTEGER PROCEDURE ID
	RLPROC=8,	\REAL PROCEDURE ID
	INFPROC=9,	\INTEGER FORWARD PROCEDURE ID
	RLFPROC=10,	\REAL FORWARD PROCEDURE ID
	INOPT=11,	\INTEGER OPTIMIZED PROCEDURE ID
	RLOPT=12,	\REAL OPTIMIZED PROCEDURE ID
	INEPRO=13,	\INTEGER XPL EXTERNAL PROCEDURE
	RLEPRO=14,	\REAL XPL EXTERNAL PROCEDURE
	ININT=15,	\INTEGER INTRINSIC ID
	RLINT=16,	\REAL INTRINSIC ID
	INEXT=17,	\INTEGER EXTERNAL ASSEMBLY ROUTINE ID
	RLEXT=18,	\REAL EXTERNAL ASSEMBLY ROUTINE ID
	INSEG=19,	\INTEGER SEGMENT VARIABLE ID
	RLSEG=20,	\REAL SEGMENT VARIABLE ID
	ADSEG=21,	\ADDRESS VARIABLE ID
	SHSEG=22;	\SHORT SEGMENT VARIABLE ID

int	LEV	\LEVEL OF CURRENT IDENTIFIER
	VAL	\VALUE OR ADDRESS OF CURRENT IDENTIFIER
	SYMNUM	\POSITION IN "SYMTBL" OF CURRENT IDENTIFER
	FACTYP;	\FACTOR (OR OPERAND) TYPE (REAL OR INTEGER)
def	\FACTYP\ REAL,INTEGER;
int	FIXES	\ARRAY--'QUIT' FIXES STILL OUTSTANDING
	LEVEL	\STATIC LEVEL OF CURRENT PROCEDURE
	NOSYM	\CURRENT NUMBER OF SYMBOLS IN SYMBOL TABLE
	FIXCNT	\COUNT OF THE NUMBER OF OUTSTANDING 'QUIT'S
	STKLOD	\NO. OF INTEGERS LEFT ON STACK BY 'FOR' ! 'CASE'
	OPTPROC	\BOOLEAN--GENERATE AN OPTIMIZED PROCEDURE CALL
	NORLSY	\CURRENT NUMBER OF REAL CONSTANTS IN TABLE
	LASTOP	\PREVIOUS OPCODE
	HASLAB	\KEEPS TRACK OF LABEL GENERATION
	II;	\SCRATCH
char	HEXDIGIT; \ARRAY OF HEX DIGITS (0 - F)
	\ -- SYMBOL TABLE ARRAYS --
char	SYMBOL	\IDENTIFIER NAME (IDENT)
	SYMTYP	\TYPE DESCRIPTORS (IDTYPE)
	SYMLEV;	\LEVEL (LEV)

int	SYMVAL	\VALUE OR ADDRESS (VAL)
	SYMPNT	\LIST LINKAGE POINTERS
	BOX;	\HASH BOXES (SYMBOL LIST HEADERS)

real	RLTBL;	\REAL CONSTANT TABLE

int	HANPTR,		\POINTER TO OLD INCLUDE HANDLES
	OLDHAN;		\ARRAY OF OLD INCLUDE HANDLES

def	HANMAX=8;	\MAXIMUM NESTING DEPTH OF INCLUDES

\RESERVED WORD HASHES:
def	ADRSYM=$88E4,	BEGSYM=$84C7,	CASEYM=$8053,	CODSYM=$8184,
	DEFSYM=$9CC6,	DOSYM=$0CEF,	ELSEYM=$99F3,	ENDSYM=$99A4,
	EXITYM=$9B69,	EXTNYM=$9B74,	FALSYM=$944C,	FFUNYM=$94B5,
	FORSYM=$9592,	FPRSYM=$9672,	FUNSYM=$96CE,	GESYM=$0C85,
	GETSYM=$90D4,	IFSYM=$0D46,	INTSYM=$A9B4,	LESYM=$0DE5,
	LOOPYM=$BD8F,	NOTSYM=$B594,	OFSYM=$0D86,	PROCYM=$CE2F,
	QUITYM=$CAC9,	REALYM=$C4C1,	REPSYM=$C4D0,	RETSYM=$C4D4,
	THENYM=$DD65,	TRUSYM=$DE35,	UNTSYM=$D9B4,	WHILYM=$D169,
	CHARYM=$8161,	EPRSYM=$9A72,	EFUNYM=$98B5,	PUBSYM=$CEC2,
	OTHSYM=$B2E8,	INCSYM=$A9A3,	LSLSYM=$BE0C,	LSRSYM=$BE92,
	SEGSYM=$C0C7,	SHTSYM=$C16F,	CONSYM=$818E,	ABSSYM=$8833,
	REMSYM=$C4CD,	SWAPYM=$C281,	EXTSYM=$9B74,	PORTYM=$CD92,
	FIXSYM=$9558,	FLOSYM=$95EF,	SQRTYM=$C252,	SQSYM=$0E11,
	STRSYM=$C2F2,	ASMSYM=$8A0D,	TOSYM=$0EEF,	DOWNYM=$9D97,
	ASRSYM=$8A12,	ANDSYM=$89A4,	ORSYM=$0D92,	XORSYM=$ED92;

proc	ERROR; int N;	\SEND ERROR MESSAGE TO THE TV
int	ERR,CH,I;
char	STRING;
def	MAXERR=74;	\MAXIMUM ERROR NUMBER
begin
\\I:= "MASM 6.11 LINE TOO LONG BUG!";
ERR:=RESERVE((MAXERR+1)*INTSIZE);
for I:=0,MAXERR do ERR(I):="? ";	\UNUSED ERROR NOS. ="?"

ERR(1):="TOO MANY VARIABLES ";
ERR(2):="TOO MANY REAL CONSTANT NAMES ";
ERR(3):="TOO MANY NAMES ";
ERR(4):="TOO MANY 'QUITS' ";
ERR(5):="TOO MANY STATIC LEVELS ";
ERR(6):="NUMBER OUT OF RANGE ";
ERR(7):=ERR(6);		\FOR INTRINSIC DECLARATIONS
ERR(10):="UNDECLARED NAME ";
ERR(11):="NAME ALREADY DECLARED ";
ERR(20):="ILLEGAL START OF A STATEMENT ";	\IN "ASSIGN"
ERR(21):="^":=^"* ";
ERR(22):="'THEN'* ";
ERR(23):="'DO'* ";
ERR(24):="'TO' OR 'DOWNTO'* ";
ERR(26):="ILLEGAL FACTOR ";	\UNRECOGNIZABLE SPECIAL FACTOR
ERR(27):="STATEMENT STARTING WITH A CONSTANT "; \IN "ASSIGN"
ERR(28):="'UNTIL'* ";
ERR(29):="'OTHER'* ";
ERR(30):="'ELSE'* ";
ERR(31):="DIGIT* ";
ERR(33):="INTEGER VARIABLE* ";	\IN A 'FOR' STATEMENT
ERR(38):="^">^"* ";		\ASR ->>
ERR(39):="^"(^"* ";
ERR(40):="^"=^"* ";
ERR(41):="^";^"* ";
ERR(42):="CONSTANT* ";		\IN "GETCON"
ERR(43):="VARIABLE* ";		\FOR AN 'ADDR' OPERATOR
ERR(44):="^")^"* ";
ERR(45):="NAME* ";
ERR(46):="MIXED MODE ";
ERR(47):="INTEGER* ";
ERR(48):="'OF'* ";
ERR(49):="^":^"* ";
ERR(50):="^"]^"* ";
ERR(51):="NO ARGUMENTS DECLARED ";
ERR(52):="STATEMENT STARTING WITH 'ELSE' ";
ERR(53):="STATEMENT STARTING WITH 'OTHER' ";
ERR(54):="ILLEGAL INTRINSIC CALL ";
ERR(60):="'QUIT' NOT IN A 'LOOP' ";
ERR(61):="EOF* ";
ERR(62):="EOF INSIDE A BLOCK ";
ERR(63):="EOF INSIDE A STRING ";
ERR(65):="'FPROC' & ITS 'PROC' NOT AT SAME LEVEL ";
ERR(66):="'FPROC' REFERENCE NOT FOUND ";
ERR(67):="'PROC' OR 'FUNC'* ";
ERR(68):="'EPROC'S AND 'PUBLIC'S MUST BE GLOBAL ";
ERR(69):="'INCLUDE'S NESTED TOO DEEP ";
ERR(70):="BAD FILE SPEC ";
ERR(71):="FILE NOT FOUND ";
ERR(72):="'INT', 'REAL', 'CHAR' or 'ADDR'* ";
ERR(73):="DIVIDE BY ZERO ";
ERR(74):="MATH ERROR IN A CONSTANT EXPRESSION ";

I:= 0;
if LSTDEV=8 then
	begin
	OPENI(8);
	loop	begin
		CH:=CHIN(8);		\INCLUDES CAUSE NESTED EOFs
		if CH=EOF then
			begin
			I:=I+1;
			if I>=HANMAX then quit;
			end
		else [I:=0; CHOUT(0,CH)];
		end;
	CRLF(0);
	end;

CHOUT(TV,BEL); CHOUT(TV,$0A);	\(DAMN LINEFEEDS!)
CRLF(TV);
TEXT(TV,"***** ERROR NO. "); INTOUT(TV,N); TEXT(TV," *****");
CRLF(TV);
STRING:=ERR(N);
I:=0;
loop	[CH:=STRING(I);		\OUTPUT MESSAGE
	if CH>=$80 then quit;
	if CH=^* then TEXT(TV," EXPECTED BUT NOT FOUND")
		else CHOUT(TV,CH);
	I:=I+1];
CRLF(TV);
TEXT(TV,"ATTEMPT TO CONTINUE (Y/N)? ");
OPENI(KB);
case CHIN(KB) of ^N,^n: [CLOSE(LSTDEV); exit 1] other [];
BINDEV:=NULDEV;		\THERE SHALL BE NO OUTPUT FILE
ERRCNT:=ERRCNT+1;
end;	\ERROR



proc	GETCH;	\GET A CHARACTER FROM THE SOURCE DEVICE
\ FILTERS OUT COMMENTS
begin
CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR);
if CHAR=^\ then				\FILTER OUT COMMENTS
	begin
	CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR);
	if CHAR=^\ then			\FILTER OUT REST OF LINE
		loop   [CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR);
			if CHAR=EOL then return;
			if CHAR=EOF then return]
	else	loop   [if CHAR=EOL then return;
			if CHAR=EOF then return;
			CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR);
			if CHAR=^\ then [CHAR:= ^ ; return]];
	end;
end;	\GETCH



proc	GETCH_;	\GET NON-UNDERLINE CHARACTER FROM THE SOURCE DEVICE
repeat GETCH until CHAR#^_;

\------------------------ ROUTINES TO HANDLE INCLUDES --------------------------

procedure FALLBACK;	\TERMINATE AN INCLUDE AND FALL BACK TO PREVIOUS HANDLE
begin
HANPTR:=HANPTR-1;
FCLOSE(INHAND);
INHAND:=OLDHAN(HANPTR);
\ONLY THE MAIN FILE GETS A BIG BUFFERS
FSET(INHAND,if HANPTR=0 then ^I else ^i);
if LEVEL<0 then LEVEL:= 0;	\for explicit 'code' declarations
end;	\FALLBACK



procedure INCLUDE;	\SET UP AN INCLUDE FILE
char NAME; define NAMMAX=80;
integer NEWHAND, I;


	procedure GETC;
	\GET CHARACTER WITH NO FILTERING
	[CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR)];


	function GETNAME;
	\READ FILE SPECIFICATION
	integer EXTFLG, I,K;
	char DEFEXT;
	begin
	DEFEXT:=".xpl";
	EXTFLG:=false;

	\EAT LEADING SPACES AND CONTROL CHARS
	while CHAR<=$20 do [if CHAR=EOF then ERROR(61); GETC];

	\COPY FILE NAME INTO 'NAME'
	K:=0;
	loop	begin
		case CHAR of
		  ^. : EXTFLG:=true;
		  EOL: return false;
		  ^; : quit
		other  [];

		NAME(K):=CHAR;
		K:=K+1;
		if K>=NAMMAX  then return false;
		GETC;
		end;

	if K=0 then return false;	\deal with empty filename

	if EXTFLG then			\deal with default extensions
		NAME(K-1):= NAME(K-1) ! $80
	else	begin
		if (K+4) >= NAMMAX then return false;
		for I:= 0 to 3 do NAME(K+I):= DEFEXT(I);
		end;
	return true;
	end;	\GETNAME


begin	\INCLUDE
NAME:=RESERVE(NAMMAX);
if HANPTR>=HANMAX then
	[ERROR(69); while CHAR#^; & CHAR#EOF do GETC; return];
if not GETNAME then
	[ERROR(70); while CHAR#^; & CHAR#EOF do GETC; return];

\OPEN FILE
TRAP($FFFB);
NEWHAND:=FOPEN(NAME,0);
TRAP($FFFF);
if GETERR=3 then [ERROR(71); return];

\SAVE OLD FILE HANDLE ON STACK
OLDHAN(HANPTR):=INHAND;
HANPTR:=HANPTR+1;
INHAND:=NEWHAND;

\INCLUDE FILES ALWAYS USE SMALL BUFFERS
FSET(INHAND,^i);

TEXT(0,"INCLUDING: ");
for I:=2,HANPTR do TEXT(0,"   ");
TEXT(0,NAME); CRLF(0);
end;	\INCLUDE



proc	INCLCODE;		\Include 'code' declarations
int	NEWHAND;
char	NAME;
begin
NAME:= "codesr.xpl";
TRAP($FFFB);
NEWHAND:= FOPEN(NAME,0);
TRAP($FFFF);
if GETERR=3 then
	[TEXT(TV, "codesr.xpl not found");
	ERROR(71); return];
OLDHAN(HANPTR):= INHAND;	\save old file handle on stack
HANPTR:= HANPTR+1;
INHAND:= NEWHAND;
FSET(INHAND,^i);		\include files always use small buffers

TEXT(0,"INCLUDING: ");
TEXT(0,NAME); CRLF(0);
end;	\INCLCODE

\-------------------------------------------------------------------------------

fproc	CONEXPRESS;
fproc	LOOKUP, TTXT, NEWLEVEL;	\FOR 'ASM'


proc	RATOM;		\READ AN ATOM
\OUTPUTS:  ATOM, ATYPE, IDENT, HASH, IATOM, CHAR, RLATOM.
\ (THIS PROCEDURE IS OPTIMIZED FOR SPEED.)
int	LEN,NEG,EXP,I,INTOVF,Braces;
real	FRACT,DENOM;

	proc	RFRACT;	\READ THE FRACTIONAL PART OF A REAL NO.
	begin
	ATYPE:=REALCON; ATOM:=0;
	GETCH_;
	FRACT:=FLOAT(0); DENOM:=FLOAT(10);	\(10.0 IS NOT SO PORTABLE)
	while CHAR>=^0 & CHAR<=^9 do
		[FRACT:=FRACT +FLOAT(CHAR-^0) /DENOM;
		DENOM:=DENOM*FLOAT(10);
		GETCH_];
	RLATOM:=RLATOM +FRACT;
	end;	\RFRACT


	proc	REXP;	\READ AN EXPONENT IF ANY
	if CHAR=^E ! CHAR=^e then
		begin
		ATYPE:=REALCON;
		GETCH_;
		if CHAR=^- then [NEG:=true; GETCH_]
			else NEG:=false;
		if CHAR=^+ then GETCH_;
		EXP:=0;
		if CHAR<^0 ! CHAR>^9 then ERROR(31);
		while CHAR>=^0 & CHAR<=^9 do
			[EXP:=EXP *10 +CHAR-^0; GETCH_];
		if NEG then EXP:= -EXP;
		while EXP>0 do
			[RLATOM:=RLATOM *FLOAT(10); EXP:=EXP-1];
		while EXP<0 do
			[RLATOM:=RLATOM /FLOAT(10); EXP:=EXP+1];
		end;	\REXP


	proc	DoAsmLine;	\Output a line of assembly code
	int	HaveComment;
	begin
	HaveComment:= false;
	loop	begin
		if CHAR=^@ then HaveComment:= true;
		if not HaveComment then CHOUT(BINDEV,CHAR);
		GETCH;
		if CHAR>=^A & CHAR<=^Z & not HaveComment then
			begin
			RATOM;			\(ATYPE=IDENTIFIER)
			LOOKUP;
			case IDTYPE of
			  ADDRVAR, INVAR, RLVAR:
				[if LEV=0 then TTXT("[r10")
				else TTXT("[r11"); \(no intermediate levels)
				TTXT(", #"); INTOUT(BINDEV,VAL); TTXT("]")];
			  INCON:
				[TTXT("#"); INTOUT(BINDEV,VAL)];
			  UNDEF:
				ERROR(10)
			other	ERROR(26);
			end;
		if CHAR=EOF then [ERROR(62); exit 1];
		if not HaveComment then
			[if CHAR=^{ then Braces:= Braces+1;
			 if CHAR=^} then Braces:= Braces-1;
			];
		if Braces<=0 then quit;		\don't output final } to BINDEV
		if CHAR=EOL then quit;
		end;
	CRLF(BINDEV);
	GETCH;
	end;	\DoAsmLine


begin	\RATOM
while CHAR<=$20\SPACE\ do
	begin	\SKIP SPACES, TABS, RETURNS, LF'S, & FF'S, ETC.
		\DON'T GO PAST EOF
	\IF HANPTR=0 THEN IT'S A HARD EOF
	if CHAR=EOF then if HANPTR>0 then FALLBACK
	else [ATYPE:=SPECIAL; ATOM:=EOF; return];
	GETCH;
	end;
if CHAR>=^a then if CHAR<=^z then			\RESERVED WORD
	[ATYPE:= SPECIAL;
	ATOM:= CHAR; GETCH;
	ATOM:= ATOM<<5|CHAR; GETCH;
	if CHAR>=^a & CHAR<=^z then [ATOM:= ATOM<<5|CHAR; GETCH];
	ATOM:= ATOM & $FFFF;
	while CHAR>=^a & CHAR<=^z do GETCH;
	case ATOM of
	  TRUSYM: [ATYPE:=INTCON; ATOM:=0; IATOM:=true];
	  FALSYM: [ATYPE:=INTCON; ATOM:=0; IATOM:=false];
	  CONSYM: begin
		  RATOM;
		  CONEXPRESS;
		  if FACTYP=INTEGER then CONDITIONAL:= IATOM else ERROR(47);
		  while ATOM=^; do RATOM;		\EAT SEMI, IF ANY
		  loop	begin				\EAT ATOMS TIL COND=TRUE
			if CONDITIONAL \#0\ then quit;
			if ATYPE=SPECIAL then
			  if ATOM=EOF then quit
			  else if ATOM=^" then		\ignore 'con' in strings
			    begin
			    if CHAR#^" then		\null string ("")
			      loop begin
				 CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR);
				 if CHAR=EOF then [ERROR(63); exit 1];
				 if CHAR=^^ then
					[CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR)]
				 else	if CHAR=^" then quit;
				 end;
			    GETCH;
			    end;
			RATOM;
			end;
		  end;
	  STRSYM:
		begin
		RATOM;
		CONEXPRESS;
		if FACTYP=INTEGER then STRTERM:= IATOM else ERROR(47);
		while ATOM=^; do RATOM;		\eat semicolon, if any
		end;
	  ASMSYM:
	    if CONDITIONAL then
	  	begin				\INSERT IN-LINE ASSEMBLY CODE
	  	LASTOP:= -1;			\instead of call to COMMENT
	  	HASLAB:=false;
	  	CODFLG:=true;
		NEWLEVEL(LEVEL);		\make sure r11 is set for locals
		while CHAR=$20\SPACE\ ! CHAR=$09\TAB\ do
			[CHOUT(BINDEV,CHAR); GETCH]; \don't skip EOL
		Braces:= 1;
		if CHAR=^{ then
			begin
			GETCH;		\eat {
			Braces:= 1;
			repeat DoAsmLine until Braces<=0;
			GETCH;		\eat }
			end
		else	DoAsmLine;
		RATOM;		\return atom following 'asm' line(s)
		end;
	  INCSYM: [INCLUDE; GETCH; RATOM]
	other	  [];
	return];
if CHAR>=^A then if CHAR<=^Z ! CHAR=^_ then		\IDENTIFIER
	begin
	ATYPE:=IDENTIFIER; ATOM:=0;
	IDENT(0):=CHAR; HASH:=CHAR; GETCH;
	LEN:=1;
	loop	begin
		if CHAR>=^a & CHAR<=^z then CHAR:= CHAR & $DF;	\UPPERCASE
		case of
		  CHAR>=^A & CHAR<=^Z,  CHAR>=^0 & CHAR<=^9,  CHAR=^_ :
			begin
			if LEN <SIGCHAR then
				[IDENT(LEN):=CHAR;
				HASH:=HASH+CHAR;
				LEN:=LEN+1];
			GETCH;
			end
		other quit;
		end;

	for LEN:=LEN,SIGCHAR-1 do
		[IDENT(LEN):=^ ; HASH:=HASH+^ ];
	HASH:=HASH & HASHMSK;
	return;
	end;
if CHAR>=^0 then if CHAR<=^9 then			\UNSIGNED INTEGER
	begin
	ATYPE:=INTCON;		\ASSUME INTEGER UNTIL SHOWN OTHERWISE
	ATOM:=0;
	INTOVF:=false;
	IATOM:=CHAR-^0;
	GETCH_;
	loop	begin
		I:=IATOM;
		if CHAR<^0 ! CHAR>^9 then quit;
		I:=IATOM*10+CHAR-^0;
\\		if IATOM>3276 ! IATOM=3276 & CHAR>^7 then	\"I" OVERFLOWED
\\			[INTOVF:=true; quit];			\OK IF IT'S REAL
		if IATOM>214748364 ! IATOM=214748364 & CHAR>^7 then
			[INTOVF:=true; quit];	\overflowed, but ok if it's real
		IATOM:=I;
		GETCH_;
		end;
	RLATOM:=FLOAT(IATOM);
	IATOM:=I;		\(CAN'T FLOAT(32768))
	while CHAR>=^0 & CHAR<=^9 do	\MORE DIGITS MUST BE REAL
		[RLATOM:=RLATOM*FLOAT(10) + FLOAT(CHAR-^0);
		GETCH_];
	if CHAR=^. then RFRACT;				\UNSIGNED REAL
	REXP;
	if ATYPE=INTCON & INTOVF & IATOM#$8000 then
		if CONDITIONAL then ERROR(6);
	return;
	end;
case CHAR of
^.:	[RLATOM:=FLOAT(0);				\UNSIGNED REAL
	RFRACT;
	REXP;
	return];
^$:	begin						\UNSIGNED HEX INTEGER
	ATYPE:=INTCON; ATOM:=0;
	GETCH_;
	case of
	  CHAR>=^0 & CHAR<=^9: IATOM:=CHAR-^0;
	  CHAR>=^A & CHAR<=^F: IATOM:=CHAR-$37;
	  CHAR>=^a & CHAR<=^f: IATOM:=CHAR-$57
	other [\DIGIT EXPECTED\ ERROR(31); return];
	loop	[GETCH_;
		case of
		  CHAR>=^0 & CHAR<=^9: I:=CHAR-^0;
		  CHAR>=^A & CHAR<=^F: I:=CHAR-$37;
		  CHAR>=^a & CHAR<=^f: I:=CHAR-$57
		other return;
\\		if IATOM&$F000 then
\\			if CONDITIONAL then ERROR(6);
		if IATOM > $FFF_FFFF then
			if CONDITIONAL then ERROR(6);
		IATOM:=IATOM*16+I];
	end;
  ^%:	begin						\UNSIGNED BINARY INTEGER
	ATYPE:=INTCON; ATOM:=0;
	GETCH_;
	if CHAR>=^0 & CHAR<=^1 then IATOM:=CHAR-^0
	else [\DIGIT EXPECTED\ ERROR(31); return];
	loop	[GETCH_;
		if CHAR>=^0 & CHAR<=^1 then I:=CHAR-^0
		else return;
		if IATOM<0 then		\(if IATOM > $7FFF ... unsigned)
			if CONDITIONAL then ERROR(6);
		IATOM:=IATOM*2+I];
	end;
^^:	[ATYPE:=INTCON;	\META CHARACTER = INTEGER CONSTANT
	ATOM:=0;
	CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR);
	IATOM:=CHAR;
	GETCH;
	return];
^":	[ATYPE:=SPECIAL;				\SPECIAL CHARACTER
	ATOM:=CHAR;\(' AND BACKSLASH HAVE NO EFFECT IN STRINGS)
	CHAR:=CHIN(SRCDEV); CHOUT(LSTDEV,CHAR);
	return]
other	[];
ATYPE:=SPECIAL;						\SPECIAL CHARACTER
ATOM:=CHAR;
GETCH;
case CHAR of
  ^=:	case ATOM of
	^::	[GETCH; ATOM:=GETSYM];
	^>:	[GETCH; ATOM:=GESYM];
	^<:	[GETCH; ATOM:=LESYM]
	other	[];
  ^<:	[if ATOM = ^< then [GETCH; ATOM:= LSLSYM]];
  ^>:	begin
	if ATOM = ^> then [GETCH; ATOM:= LSRSYM]
	else if ATOM = ^- then
		[GETCH;
		if CHAR#^> then ERROR(38);
		GETCH;
		ATOM:= ASRSYM];
	end
other	[];
end;	\RATOM

\-------------------------------------------------------------------------------

proc	SKIPIT;
\SKIP THE REST OF A STATEMENT FOR ERROR RECOVERY
while ATOM#EOF & ATOM#^; & ATOM#ENDSYM & ATOM#^] &
	ATOM#BEGSYM & ATOM#^[ do RATOM;



proc	HEXB; int I;	\OUTPUT HEX BYTE (OPTIMIZED FOR SPEED)
begin
CHOUT(BINDEV,HEXDIGIT( (I&$FF)/16 ));
CHOUT(BINDEV,HEXDIGIT(REM(0)));
end;	\HEXB



proc	SYMOUT(SYM);	\OUTPUT SYMBOL NAME AT 'SYM'
int	SYM;
int	I, C;
begin
for I:= 0, SIGCHAR-1 do
	begin
	C:= SYMBOL(SYM);
	if C=$20 then return;
	CHOUT(BINDEV, C);
	SYM:= SYM +SYMAX;
	end;
end;	\SYMOUT

\===================================================================

proc TTXT(STR);		\OUTPUT ASSEMBLY STRINGS
char STR;
int I,CH;
begin
I:=0;
loop	begin
	CH:=STR(I);
 	if CH>127 then quit;
	if CH=^| then CRLF(BINDEV) else CHOUT(BINDEV,CH);
	I:=I+1;
	end;
CH:=CH&$7F;
if CH=^| then CRLF(BINDEV) else CHOUT(BINDEV,CH);
end;


procedure LOCOUT;
\OUTPUT CURRENT LOCAL LABEL
[TEXT(BINDEV,"LL"); INTOUT(BINDEV,LOCAL)];


proc GENIMAGE;
begin
if IMMTYP=^L then	\generate address to a label for strings and constant
	begin		\ arrays whose total space can be addressed with 2 bytes
	TTXT("	add	r0, r8, #(L"); INTOUT(BINDEV,IMMVAL);
	TTXT("-dseg)&0xFF @0B IMM|");
	TTXT("	add	r0, #(L"); INTOUT(BINDEV,IMMVAL); TTXT("-dseg)&0xFF00|");
	TTXT("	add	r0, #(L"); INTOUT(BINDEV,IMMVAL); TTXT("-dseg)&0xFF0000|");
	end
else if (~IMMVAL)>=0 & (~IMMVAL)<=$FF then 
	[TTXT("	mvn	r0, #0x"); HEXB(~IMMVAL&$00FF);
	TTXT("		@0B IMM|")]		\-1..-256
     else
	begin
	TTXT("	mov	r0, #0x"); HEXB(IMMVAL&$00FF);
	TTXT("		@0B IMM|");
	if IMMVAL>>8 then
		[TTXT("	orr	r0, #0x"); HEXB(IMMVAL>>8); TTXT("00|")];
	if IMMVAL->>16 then
		[TTXT("	orr	r0, #0x"); HEXB(IMMVAL->>16); TTXT("0000|")];
	if IMMVAL->>24 then
		[TTXT("	orr	r0, #0x"); HEXB(IMMVAL->>24); TTXT("000000|")];
	end;
end;


proc STKTOS;
\THIS IS INVOKED AT THE START OF A GEN IF THE OP
\ NEEDS TO BE SURE TOS HAS BEEN STACKED
begin
case TOS of
IMAGE:	begin
	GENIMAGE;
	TTXT("	push	{r0}|");
	end;
INREG:	TTXT("	push	{r0}|")
other	[];
TOS:=STACKED;
end;	\STKTOS


proc PULTOS;
\THIS IS INVOKED AT THE START OF AN OP THAT NEEDS TO HAVE TOS IN THE AC
begin
case TOS of
IMAGE:	 GENIMAGE;
STACKED: TTXT("	pop	{r0}|")
other	 [];
TOS:=INREG;
end;	\PULTOS


procedure NEWLEVEL(L);	\LOAD A NEW LEVEL IF NEEDED
int L;
begin
if OLDLEV#L then
	begin
	TTXT("	ldr	r11, [r8, #base"); INTOUT(BINDEV,L/2); TTXT("-dseg]|");
	OLDLEV:=L;
	end;
end;


procedure COMMENT(OPC);
integer OPC, COMTB1,COMTB2;
begin
LASTOP:=OPC;
HASLAB:=false;
CODFLG:=true;
if not COMFLG then return;

COMTB1:=
["EXIT","LOD","LDX","STO","STX","CAL","RET","JMP","JPC","HPI","ARG","IMM",
"CML","ADD","SUB","MUL","DIV","NEG","EQ","NE","GE","GT","LE","LT","FOR",
"INP","OR","AND","NOT","EOR","DBA","STD","DBX","ADR","LDI","LDA","IMS",
"CJP","JSR","RTS","DRP","CEXT","FLOD","FSTO","FIMM","FADD","FSUB","FMUL",
"FDIV","FNEG","FEQ","FNE","FGE","FGT","FLE","FLT","TRA","TRX","TRI","STT",
"MARK","ASR","LSL","LSR","LDSI","LDSB","LDSR","STSI","STSB","STSR","LSHORT",
"SSHORT","MKARRAY","ABS","FABS","REM","SWAP","EXT","PIN","POUT"];

COMTB2:=["TXT","FLT","INT","END","NUL"];

CRLF(BINDEV); TEXT(BINDEV,"; "); INTOUT(BINDEV,OPC); TEXT(BINDEV," - ");
TEXT(BINDEV,if OPC>$FA then COMTB2(OPC-$FB) else COMTB1(OPC));
CRLF(BINDEV);
end;	\COMMENT

\-------------------------------------------------------------------------------

proc DOEXIT;
begin
TTXT("	ldr	sp, [r8, #stkptr-dseg]	@00 EXIT|");
TTXT("	pop	{pc}|");
end;


proc GENEXIT;
\$00
[COMMENT(0); DOEXIT];


proc GENLOD(L,O);		\$01
int L,O;

   proc GENOFF(R);
   int  R;
   begin
   if O >= 1<<24 then ERROR(1);				\16Meg	   0x01000000
   if O < $400 then
	[TTXT("	ldr	r0, [r"); INTOUT(BINDEV,R); TTXT(", #");
	INTOUT(BINDEV,O); TTXT("]		@01 LOD|");
	]
   else [TTXT("	add	r1, r"); INTOUT(BINDEV,R);	\add r1, r11, #0x5600
	TTXT(", #0x"); HEXB(O>>8); TTXT("00	@01 LOD|");
	if O >= 1<<16 then				\add r1,    #0x340000
		[TTXT("	add	r1,    #0x"); HEXB(O>>16); TTXT("0000|")];
	TTXT("	ldr	r0, [r1, #0x00"); HEXB(O); TTXT("]|");
	];						\ldr r0, [r1, #0x0078]
   end;

begin	\GENLOD
COMMENT($01);
STKTOS;
if L=0 then GENOFF(10)
else [NEWLEVEL(L); GENOFF(11)];
TOS:=INREG;
end;


proc GENLDX;			\$02
begin
COMMENT($02);
PULTOS;
TTXT("	pop	{r1}			@02 LDX|");
TTXT("	ldrb	r0, [r0, r1]|");
TOS:=INREG;
end;


proc GENSTO(L,O);		\$03
int L,O;

   proc GENOFF(R);
   int  R;
   begin
   if O >= 1<<24 then ERROR(1);				\16Meg	   0x01000000
   if O < $400 then
	[TTXT("	str	r0, [r"); INTOUT(BINDEV,R); TTXT(", #");
	INTOUT(BINDEV,O); TTXT("]		@03 STO|");
	]
   else [TTXT("	add	r1, r"); INTOUT(BINDEV,R);	\add r1, r11, #0x5600
	TTXT(", #0x"); HEXB(O>>8); TTXT("00	@03 STO|");
	if O >= 1<<16 then				\add r1,    #0x340000
		[TTXT("	add	r1,    #0x"); HEXB(O>>16); TTXT("0000|")];
	TTXT("	str	r0, [r1, #0x00"); HEXB(O); TTXT("]|");
	];						\str r0, [r1, #0x0078]
   end;

begin	\GENSTO
COMMENT($03);
PULTOS;
if L=0 then GENOFF(10)
else [NEWLEVEL(L); GENOFF(11)];
TOS:=STACKED;
end;


proc GENSTX;			\$04
begin
COMMENT($04);
PULTOS;
TTXT("	pop	{r1}			@04 STX|");
TTXT("	strb	r0, [r1]|");
TOS:=STACKED;
end;


procedure DOCALL(L);
int L;
begin
STKTOS;
TTXT("	bl	L"); INTOUT(BINDEV,L); \CRLF(BINDEV);
TOS:=STACKED;
OLDLEV:=-1;	\CALL MAY HAVE BOMBED IT
end;


proc GENCAL(L);			\$05
int L;
begin
COMMENT($05);
DOCALL(L);
TTXT("			@05 CAL|");
end;


proc GENRET(L);			\$06
int L;
begin
if LASTOP=$06 then return;	\AVOID DUPLICATE RETS
COMMENT($06);
STKTOS;
if L=0 then DOEXIT
else	begin
	TTXT("	ldr	r9, [r8, #base"); INTOUT(BINDEV,L/2);
	TTXT("-dseg]	@06 RET|"); 
	TTXT("	pop	{r0}|");
	TTXT("	str	r0, [r8, #base"); INTOUT(BINDEV,L/2); TTXT("-dseg]|"); 
	TTXT("	pop	{pc}|");
	end;
end;


proc GENJMP(L);			\$07
int L;
begin
COMMENT($07);
STKTOS;		\CLEAN UP?
TTXT("	b	L"); INTOUT(BINDEV,L); TTXT("			@07 JMP|");
TOS:=STACKED;
end;


proc GENJPC(L);			\$08 jump on false (0)
int L;
begin
COMMENT($08);
PULTOS;
TTXT("	cmp	r0, #0|");
TTXT("	beq	L"); INTOUT(BINDEV,L); TTXT("			@08 JPC|");
TOS:=STACKED;
end;


proc GENHPI(L,V);		\$09
int L,V;	\V=number of bytes to reserve
begin		\WARNING: V can appear negative because of array declarations
COMMENT($09);
if V>>24 then ERROR(1);
if V&$FF then
	[TTXT("	add	r9, #0x0000");  HEXB(V);
	TTXT("		@09 HPI|")];
if V>>8&$FF then
	[TTXT("	add	r9, #0x00");  HEXB(V>>8);  TTXT("00|")];
if V>>16&$FF then
	[TTXT("	add	r9, #0x");  HEXB(V>>16);  TTXT("0000|")];
end;


proc GENBASE(L,V);		\$09-1/2
int L,V;	\WARNING: V is unsigned and can appear negative
begin
COMMENT($09);
STKTOS;		\WE NEED THE AC
TTXT("	ldr	r0, [r8, #base"); INTOUT(BINDEV,L/2); TTXT("-dseg]|"); 
TTXT("	push	{r0, lr}|	str	r9, [r8, #base"); INTOUT(BINDEV,L/2);
TTXT("-dseg]|"); 
end;


proc GENOPTPROC;		\$09-3/4 Save link register for optimized procs
begin
COMMENT($09);
TTXT("	push	{lr}|");
end;


proc GENARG(V);			\$0A
int V;
begin
COMMENT($0A);
PULTOS;
V:= V-INTSIZE;
if V>255 then ERROR(1);
TTXT("	str	r0, [r9, #"); INTOUT(BINDEV,V); TTXT("]		@0A ARG|");
TOS:= STACKED;
if V<=0 then return;

if V<=16 then
	begin
	if V>12 then TTXT("	pop	{r0}|	str	r0, [r9, #12]|");
	if V> 8 then TTXT("	pop	{r0}|	str	r0, [r9, #8]|");
	if V> 4 then TTXT("	pop	{r0}|	str	r0, [r9, #4]|");
		     TTXT("	pop	{r0}|	str	r0, [r9]|");
	end
else	begin
	TTXT("	mov	r1, #"); INTOUT(BINDEV,V);
	TTXT("-4|");
	LOCOUT; TTXT(":	pop	{r0}|");
	TTXT("	str	r0, [r9, r1]|");
	TTXT("	subs	r1, #4|");
	TTXT("	bge	"); LOCOUT; CRLF(BINDEV);
	LOCAL:=LOCAL+1;
	end;
end;	\GENARG


proc	GENFARG;		\$0Ax; HANDLE REAL ARGUMENT
\This reverses the words in a real so that when GENARG pops them off the stack
\ into the heap they'll be forwards. Might be a problem for external subroutines
begin
COMMENT($0A);
STKTOS;
TTXT("	ldr	r0, [sp]|	ldr	r1, [sp, #4]|");	\swap words
TTXT("	str	r0, [sp, #4]|	str	r1, [sp]|");		\ on the stack
TOS:=STACKED;
end;


proc GENIMM(F,V);		\$0B & $24 (both 32-bit and 8-bit)
int F,V;
begin
COMMENT($0B);
STKTOS;
IMMTYP:=F; IMMVAL:=V;
TOS:=IMAGE;
end;


proc GENCML(LABEL,TYPE,FUNC,ARG); \$0C 
int LABEL,TYPE,FUNC,ARG;
int IntTbl;
begin
COMMENT($0C);
if ARG>0 &  TYPE=ININT then PULTOS else STKTOS;
TTXT("	bl	intr"); INTOUT(BINDEV, LABEL);
if LABEL=12 & STRTERM=0 then CHOUT(BINDEV, ^a);	\use alternate Text intrinsic
TTXT("			@0C CML|");
\OLDLEV:=-1;		intrinsics don't alter r11
TOS:=if (TYPE=ININT) & FUNC  then INREG else STACKED;

include codesrx.xpl;
\\	true,  4,	\\var:=\Abs(int)=0
if IntTbl(LABEL*2)#FUNC then ERROR(54);
\\CrLf(0); IntOut(0, Arg); Text(0, "/"); IntOut(0, IntTbl(Label*2+1)); CrLf(0);
if IntTbl(LABEL*2+1)#ARG then ERROR(54);
end;


proc GENADD;			\$0D
begin
PULTOS;
TTXT("	pop	{r1}			@0D ADD|");
TTXT("	add	r0, r1|");
TOS:=INREG;
end;


proc GENSUB;			\$0E; TOS:= NOS-TOS
begin
PULTOS;
TTXT("	pop	{r1}			@0E SUB|");
TTXT("	rsb	r0, r1|");
TOS:=INREG;
end;


proc GENMUL;			\$0F
begin
PULTOS;
TTXT("	pop	{r1}			@0F MUL|");
TTXT("	mul	r0, r1|");
TOS:=INREG;
end;


proc GENDIV;			\$10; tos:= nos/tos
begin
PULTOS;
TTXT("	bl	sdivx			@10 DIV|");
TOS:=INREG;
end;


proc GENNEG;			\$11
begin
COMMENT($11);
PULTOS;
TTXT("	rsb	r0, #0			@11 NEG|");
TOS:=INREG;
end;


proc FCMP(T);			\$32-$37
int T;
int CMPT;
begin
CMPT:=[
"	subeq	r0, #1|",
"	subne	r0, #1|",
"	subge	r0, #1|",
"	subgt	r0, #1|",
"	suble	r0, #1|",
"	sublt	r0, #1|"];
STKTOS;
TTXT("	fldd	d0, [sp]		@3_ FCMP|");
TTXT("	fldd	d1, [sp, #8]|");
TTXT("	fcmpd	d1, d0|");
TTXT("	add	sp, #16|");		\pop tos and nos
TTXT("	fmstat|");
TTXT("	mov	r0, #0|");
TTXT(CMPT(T-$32));
TOS:=INREG;
end;


proc ICMP(T);			\$12-$17
int T;
int CMPT;
begin
CMPT:=[
"	subeq	r0, #1|",
"	subne	r0, #1|",
"	subge	r0, #1|",
"	subgt	r0, #1|",
"	suble	r0, #1|",
"	sublt	r0, #1|"];
PULTOS;
TTXT("	pop	{r1}			@1_ CMP|");
TTXT("	cmp	r1, r0|");
TTXT("	mov	r0, #0|");
TTXT(CMPT(T-$12));
TOS:=INREG;
end;


proc GENCMP(T);			\SELECT INTEGER OR REAL COMPARES
int T;
begin
COMMENT(T);
if T>$17 then FCMP(T) else ICMP(T);
end;


proc GENINP(L,O,DOWNTO);	\$19, $4A
int L,O,DOWNTO;

   proc GENLDR(R);
   int  R;
   begin
   if O >= 1<<24 then ERROR(1);				\16Meg	   0x01000000
   if O < $400 then
	[TTXT("	ldr	r0, [r"); INTOUT(BINDEV,R); TTXT(", #");
	INTOUT(BINDEV,O); TTXT("]|");			\ldr r0, [r11, #1023]
	]
   else [TTXT("	add	r1, r"); INTOUT(BINDEV,R);	\add r1, r11, #0x5600
	TTXT(", #0x"); HEXB(O>>8); TTXT("00|");
	if O >= 1<<16 then				\add r1,    #0x340000
		[TTXT("	add	r1,    #0x"); HEXB(O>>16); TTXT("0000|")];
	TTXT("	ldr	r0, [r1, #0x00"); HEXB(O); TTXT("]|");
	];						\ldr r0, [r1, #0x0078]
   end;

   proc GENSTR(R);
   int  R;
   begin
   if O < $400 then
	[TTXT("	str	r0, [r"); INTOUT(BINDEV,R); TTXT(", #");
	INTOUT(BINDEV,O); TTXT("]|");			\str r0, [r11, #1023]
	]
   else [TTXT("	str	r0, [r1, #0x00"); HEXB(O); TTXT("]|");
	];						\str r0, [r1, #0x0078]
   end;

   proc	GENOFF(R);
   int  R;
   begin
   GENLDR(R);
   if DOWNTO then
	TTXT("	sub	r0, #1			@4A DEP|")
   else TTXT("	add	r0, #1			@19 INP|");
   GENSTR(R);
   end;

begin	\GENINP
COMMENT(if DOWNTO then $4A else $19);
STKTOS;
if L=0 then GENOFF(10)
else [NEWLEVEL(L); GENOFF(11)];
end;


proc GENFOR(L,O,LAB,DOWNTO); \$18, $49
int L,O,LAB,DOWNTO;

   proc GENOFF(R);
   int  R;
   begin
   if O >= 1<<24 then ERROR(1);				\16Meg	   0x01000000
   if O < $400 then
	[TTXT("	ldr	r0, [r"); INTOUT(BINDEV,R); TTXT(", #");
	INTOUT(BINDEV,O); TTXT("]|");			\ldr r0, [r11, #1023]
	]
   else [TTXT("	add	r1, r"); INTOUT(BINDEV,R);	\add r1, r11, #0x5600
	TTXT(", #0x"); HEXB(O>>8); TTXT("00|");
	if O >= 1<<16 then				\add r1,    #0x340000
		[TTXT("	add	r1,    #0x"); HEXB(O>>16); TTXT("0000|")];
	TTXT("	ldr	r0, [r1, #0x00"); HEXB(O); TTXT("]|");
	];						\ldr r0, [r1, #0x0078]
   end;

begin	\GENFOR
COMMENT(if DOWNTO then $49 else $18);
TTXT(if DOWNTO then "					@49 FORDN|"
	       else "					@18 FOR|");
if L=0 then GENOFF(10)
else [NEWLEVEL(L); GENOFF(11)];
TTXT("	ldr	r1, [sp]|");
TTXT("	cmp	r0, r1|");
TTXT(if DOWNTO then "	bge	L" else "	ble	L");
INTOUT(BINDEV,LAB);
TTXT("|	add	sp, #4|");
TOS:=STACKED;
end;


proc GENOR;			\$1A
begin
COMMENT($1A);
PULTOS;
TTXT("	pop	{r1}			@1A OR|");
TTXT("	orr	r0, r1|");
TOS:=INREG;
end;


proc GENAND;			\$1B
begin
COMMENT($1B);
PULTOS;
TTXT("	pop	{r1}			@1B AND|");
TTXT("	and	r0, r1|");
TOS:=INREG;
end;


proc GENNOT;			\$1C
begin
COMMENT($1C);
PULTOS;
TTXT("	mvn	r0, r0			@1C NOT|");
TOS:=INREG;
end;


proc GENEOR;			\$1D
begin
COMMENT($1D);
PULTOS;
TTXT("	pop	{r1}			@1D EOR|");
TTXT("	eor	r0, r1|");
TOS:=INREG;
end;


proc GENDBA;			\$1E: TOS:= NOS + TOS*4
begin
COMMENT($1E);
PULTOS;
TTXT("	pop	{r1}			@1E DBA|");
TTXT("	add	r0, r1, r0, lsl #2|");
TOS:=INREG;
end;


proc GENSTD;			\$1F: store TOS at address in NOS
begin
COMMENT($1F);
PULTOS;
TTXT("	pop	{r1}			@1F STD|");
TTXT("	str	r0, [r1]|");
TOS:=STACKED;
end;


proc GENDBX;			\$20: Load(TOS*4+NOS)
begin
COMMENT($20);
PULTOS;
TTXT("	pop	{r1}			@20 DBX|");
TTXT("	ldr	r0, [r1, r0, lsl #2]|");
TOS:=INREG;
end;


proc GENADR(L,O);		\$21
int L,O;	\WARNING: O is unsigned and can appear negative

   proc GENOFF(R);
   int  R;
   begin
   if O >= 1<<24 then ERROR(1);				\16Meg	   0x01000000
   if O < $100 then
	[TTXT("	add	r0, r"); INTOUT(BINDEV,R); TTXT(", #0x");
	HEXB(O); TTXT("		@21 ADR|");		\add r0, r11, #0x78
	]
   else [TTXT("	add	r0, r"); INTOUT(BINDEV,R);	\add r0, r11, #0x5600
	TTXT(", #0x"); HEXB(O>>8); TTXT("00	@21 ADR|");
	if O >= 1<<16 then				\add r0,    #0x340000
		[TTXT("	add	r0,    #0x"); HEXB(O>>16); TTXT("0000|")];
	TTXT("	add	r0,      #0x00"); HEXB(O); TTXT("|");
	];						\add r0,      #0x0078
   end;

begin
COMMENT($21);
STKTOS;
if L=0 then GENOFF(10)
else [NEWLEVEL(L); GENOFF(11)];
TOS:=INREG;
end;


\proc GENLDI;	not used


proc GENCJP(L);			\$25
int L;
begin
COMMENT($25);
PULTOS;
TTXT("	ldr	r1, [sp]		@25 CJP|");
TTXT("	cmp	r0, r1|");
TTXT("	bne	L"); INTOUT(BINDEV,L); CRLF(BINDEV);
TOS:=STACKED;
end;


proc GENJSR(L);			\$26
int L;
begin
COMMENT($26);
DOCALL(L);
TTXT("			@26 JSR|");
end;


proc GENRTS;			\$27
begin
if LASTOP=$27 then return;	\AVOID DUPLICATE RTSS
COMMENT($27);
STKTOS;
TTXT("	pop	{pc}			@27 RTS|");
TOS:=STACKED;
end;


proc GENDRP;			\$28
begin
COMMENT($28);
if TOS=STACKED then		\discard from stack
	TTXT("	add	sp, #4			@28 DRP|");
TOS:=STACKED;			\discard if INREG or IMAGE
end;


proc GENCEXT(LABEL,TYPE,FUNC,ARG); \$29: CALL EXTERNAL XPL ROUTINE
int LABEL,TYPE,FUNC,ARG;
begin
COMMENT($29);
STKTOS;
TTXT("	bl	_"); SYMOUT(LABEL); CRLF(BINDEV);
OLDLEV:=-1;
TOS:=STACKED;
end;


proc GENAEXT(LABEL,TYPE,FUNC,ARG);	\$29A: CALL EXTERNAL ASSEMBLY ROUTINE
int LABEL,TYPE,FUNC,ARG;
begin
COMMENT($29);
STKTOS;
TTXT("	bl	"); SYMOUT(LABEL); CRLF(BINDEV);
OLDLEV:=-1;
TOS:=STACKED;
end;


proc GENFLOD(L,O);		\$2A
int L,O;

   proc GENOFF(R);
   int  R;
   begin
   if O >= 1<<24 then ERROR(1);				\16Meg	    0x01000000
   if O < $100 then
	[TTXT("	ldrd	r0, [r"); INTOUT(BINDEV,R); TTXT(", #");
	INTOUT(BINDEV,O); TTXT("]		@2A FLOD|");
	]
   else [TTXT("	add	r2, r"); INTOUT(BINDEV,R);	\add r1, r11,  #0x5600
	TTXT(", #0x"); HEXB(O>>8); TTXT("00	@2A FLOD|");
	if O >= 1<<16 then				\add r1,     #0x340000
		[TTXT("	add	r2,    #0x"); HEXB(O>>16); TTXT("0000|")];
	TTXT("	ldrd	r0, [r2, #0x00"); HEXB(O); TTXT("]|");
	];						\ldrd r0, [r1, #0x0078]
   end;

begin
COMMENT($2A);
STKTOS;
if L=0 then GENOFF(10)
else [NEWLEVEL(L); GENOFF(11)];
TTXT("	push	{r0, r1}|");
TOS:=STACKED;
end;


proc GENFSTO(L,O);		\$2B
int L,O;

   proc GENOFF(R);
   int  R;
   begin
   if O >= 1<<24 then ERROR(1);				\16Meg	    0x01000000
   if O < $100 then
	[TTXT("	strd	r0, [r"); INTOUT(BINDEV,R); TTXT(", #");
	INTOUT(BINDEV,O); TTXT("]|");
	]
   else [TTXT("	add	r2, r"); INTOUT(BINDEV,R);	\add r1, r11,  #0x5600
	TTXT(", #0x"); HEXB(O>>8); TTXT("00|");
	if O >= 1<<16 then				\add r1,     #0x340000
		[TTXT("	add	r2,    #0x"); HEXB(O>>16); TTXT("0000|")];
	TTXT("	strd	r0, [r2, #0x00"); HEXB(O); TTXT("]|");
	];						\strd r0, [r1, #0x0078]
   end;

begin
COMMENT($2B);
STKTOS;
TTXT("	pop	{r0, r1}		@2B FSTO|");
if L=0 then GENOFF(10)
else [NEWLEVEL(L); GENOFF(11)];
TOS:=STACKED;
end;


proc GENFIMM(T,V);		\$2C
int T,V;
begin
COMMENT($2C);
STKTOS;		\CLEAN UP
if T=^C then
	begin
	TTXT("	add	r1, r8, #(L"); INTOUT(BINDEV,V);
	TTXT("-dseg)&0xFF @2C FIMM|");
	TTXT("	add	r1, #(L"); INTOUT(BINDEV,V); TTXT("-dseg)&0xFF00|");
	TTXT("	ldrd	r0, [r1]|");
	TTXT("	push	{r0, r1}|");
	end
else if T=^L then
	begin
	TTXT("	add	r0, r8, #(L"); INTOUT(BINDEV,V);
	TTXT("-dseg)&0xFF	@2C FIMM|");
	TTXT("	add	r0, #(L"); INTOUT(BINDEV,V); TTXT("-dseg)&0xFF00|");
	TTXT("	mov	r1, #0x3F000000|");
	TTXT("	orr	r1, #0x00F00000|");
	TTXT("	push	{r0, r1}|");
	end
else ERROR(10);
end;


proc GENFADD;			\$2D; tos:= nos + tos
begin
STKTOS;
TTXT("	vpop	{d0}			@2D FADD|");
TTXT("	fldd	d1, [sp]|");				\get what was nos
TTXT("	faddd	d0, d1, d0|");
TTXT("	fstd	d0, [sp]|");				\push
TOS:=STACKED;
end;


proc GENFSUB;			\$2E; pop NOS into d1, TOS:= NOS-TOS
begin
STKTOS;
TTXT("	vpop	{d0}			@2E FSUB|");
TTXT("	fldd	d1, [sp]|");				\get what was nos
TTXT("	fsubd	d0, d1, d0|");
TTXT("	fstd	d0, [sp]|");				\push
TOS:=STACKED;
end;


proc GENFMUL;			\$2F
begin
STKTOS;
TTXT("	vpop	{d0}			@2F FMUL|");
TTXT("	fldd	d1, [sp]|");				\get what was nos
TTXT("	fmuld	d0, d1, d0|");
TTXT("	fstd	d0, [sp]|");				\push
TOS:=STACKED;
end;


proc GENFDIV;			\$30
begin
STKTOS;
TTXT("	vpop	{d0}			@30 FDIV|");
TTXT("	fldd	d1, [sp]|");				\get what was nos
TTXT("	fdivd	d0, d1, d0|");
TTXT("	fstd	d0, [sp]|");				\push
TOS:=STACKED;
end;


proc GENMTH(T);			\SELECT REAL OR INTEGER MATH OPS
int T;
begin
COMMENT(T);
case T of
  $0D: GENADD;
  $0E: GENSUB;
  $0F: GENMUL;
  $10: GENDIV;
  $2D: GENFADD;
  $2E: GENFSUB;
  $2F: GENFMUL;
  $30: GENFDIV
other  [];
end;	\GENMTH


proc GENFNEG;			\$31
begin
COMMENT($31);
STKTOS;
TTXT("	fldd	d0, [sp]		@31 FNEG|");
TTXT("	fnegd	d0, d0|");
TTXT("	fstd	d0, [sp]|");
end;


proc GENTRA;			\$38; tos:= 8*tos + nos
begin
COMMENT($38);
PULTOS;
TTXT("	pop	{r1}			@38 TRA|");
TTXT("	add	r0, r1, r0, lsl #3|");
TOS:=INREG;
end;


proc GENTRX;			\$39; ftos:= (tos*8 + nos)
begin
COMMENT($39);
PULTOS;
TTXT("	pop	{r1}			@39 TRX|");
TTXT("	ldr	r0, [r1, r0, lsl #3]|");
TOS:=INREG;
end;


proc GENTRI;			\$3A
\THIS ROUTINE ADDS 8*TOS + NOS, AND USES THE SUM AS AN ADDRESS TO FETCH A VALUE.
\ NOS IS INTEGER PACKED INTO A REAL.
begin
COMMENT($3A);
PULTOS;				\get the address in low half of (real) nos 
TTXT("	ldr	r1, [sp]		@3A TRI|");
TTXT("	add	r2, r1, r0, lsl #3|");
TTXT("	fldd	d0, [r2]|");
TTXT("	fstd	d0, [sp]|");	\replace tos
TOS:=STACKED;
end;


proc GENSTT;			\$3B
\STORE THE 8-BYTE REAL IN TOS INTO THE ABSOLUTE ADDRESS POINTED TO BY NOS.
\ THIS OPCODE IS USED TO STORE INDEXED REAL NUMBERS.
begin
COMMENT($3B);
STKTOS;
TTXT("	pop	{r0-r2}			@3B STT|");
TTXT("	stmia	r2, {r0-r1}|");
TOS:=STACKED;
end;


proc GENMARK;			\$3C
COMMENT($3C);


proc GENASR;			\$3D
begin
COMMENT($3D);
PULTOS;
TTXT("	pop	{r1}			@3D ASR|");
TTXT("	mov	r0, r1, asr r0|");
TOS:=INREG;
end;


proc GENLSL;			\$3E
begin
COMMENT($3E);
PULTOS;
TTXT("	pop	{r1}			@3E LSL|");
TTXT("	mov	r0, r1, lsl r0|");
TOS:=INREG;
end;


proc GENLSR;			\$3F tos:= nos>>tos
begin
COMMENT($3F);
PULTOS;
TTXT("	pop	{r1}			@3F LSR|");
TTXT("	mov	r0, r1, lsr r0|");
TOS:=INREG;
end;

\-------------------------------------------------------------------------------

proc GENLDSI;			\$40: segment
begin
COMMENT($40);
PULTOS;
TTXT("	SAL	AX,1|	MOV	BX,AX|	POP	DS|");
TTXT("	MOV	AX,[BX]|	MOV	BX,SS|	MOV	DS,BX|");
TOS:=INREG;
end;


proc GENLDSB;			\$41: segment
begin
COMMENT($41);
PULTOS;
TTXT("	MOV	BX,AX|	POP	DS|	MOV	AL,[BX]|");
TTXT("	XOR	AH,AH|	MOV	BX,SS|	MOV	DS,BX|");
TOS:=INREG;
end;


proc GENLDSR;			\$42: segment
begin
COMMENT($42);
PULTOS;
TTXT("	CALL	LDSRDO|");
TOS:=STACKED;
end;


proc GENSTSI;			\$43: segment
begin
COMMENT($43);
PULTOS;
TTXT("	POP	BX|	SAL	BX,1|	POP	DS|");
TTXT("	MOV	[BX],AX|	MOV	BX,SS|	MOV	DS,BX|");
TOS:=STACKED;
end;


proc GENSTSB;			\$44: segment
begin
COMMENT($44);
PULTOS;
TTXT("	POP	BX|	POP	DS|");
TTXT("	MOV	[BX],AL|	MOV	BX,SS|	MOV	DS,BX|");
TOS:=STACKED;
end;


proc GENSTSR;			\$45: segment
begin
COMMENT($45);
STKTOS;
TTXT("	CALL	STSRDO|");
TOS:=STACKED;
end;


proc GENLSHORT;			\$46: segment
begin
COMMENT($46);
PULTOS;
TTXT("	CALL	LSHORT|");
TOS:=STACKED;
end;


proc GENSSHORT;			\$47: segment
begin
COMMENT($47);
STKTOS;
TTXT("	CALL	SSHORT|");
TOS:=STACKED;
end;


proc GENARY;			\$48: ARRAY DECLARATION
begin
COMMENT($48);
STKTOS;
TTXT("	bl	mkarray|");
OLDLEV:=-1;
TOS:=STACKED;
end;

\-------------------------------------------------------------------------------

proc GENABS;			\$49: ABSOLUTE VALUE FUNCTION
begin
COMMENT($49);
PULTOS;
TTXT("	cmp	r0, #0			@49 ABS|");
TTXT("	rsbmi	r0, #0|");
TOS:=INREG;
end;


proc GENFABS;			\$4A: ABSOLUTE VALUE FUNCTION OF A REAL
begin
COMMENT($4A);
STKTOS;
TTXT("	fldd	d0, [sp]		@4A FABS|");
TTXT("	fabsd	d0, d0|");
TTXT("	fstd	d0, [sp]|");
end;


proc GENREM;			\$4B: REMAINDER OF LAST INTEGER DIVIDE
begin
COMMENT($4B);
PULTOS;		\DISCARD TOS
TTXT("	ldr	r0, [r8, #remain-dseg]	@4B REM|");
TOS:=INREG;
end;


proc GENSWAP;			\$4C: SWAP BYTES FUNCTION
begin
COMMENT($4C);
PULTOS;
TTXT("	rev16	r0, r0			@4C SWAP|");
TOS:=INREG;
end;


proc GENEXT;			\$4D: SIGN EXTEND FUNCTION
begin
COMMENT($4D);
PULTOS;
TTXT("	sxtb	r0, r0			@4D EXT|");
TOS:=INREG;
end;


proc GENPIN;			\$4E, FUNCTION TO READ A BYTE FROM A PORT
begin				\TOS:= port(TOS);
COMMENT($4E);
PULTOS;
TTXT("	MOV	DX,AX|	IN	AL,DX|	XOR	AH,AH|");
TOS:=INREG;
end;


proc GENPOUT;			\$4F, FUNCTION TO WRITE A BYTE TO A PORT
begin
COMMENT($4F);
PULTOS;
TTXT("	POP	DX|	OUT	DX,AL|");
TOS:=STACKED;
end;


proc	GENFIX;		\$50, Fix function
begin			\INTEGER:= FIX(REAL);
COMMENT($50);
TTXT("	vpop	{d0}			@50 FIX|");
TTXT("	ftosid	s2, d0|");
TTXT("	fmrs	r0, s2|");
TOS:=INREG;
end;


proc	GENFLOAT;	\$51, Float function
begin			\REAL:= FLOAT(INTEGER);
COMMENT($51);
PULTOS;
TTXT("	fmsr	s2, r0			@51 FLOAT|");
TTXT("	fsitod	d0, s2|");
TTXT("	vpush	{d0}|");
TOS:=STACKED;
end;


proc	GENFSQRT;	\$52, Sqrt function of a real
begin
COMMENT($52);
TTXT("	fldd	d0, [sp]		@52 FSQRT|");
TTXT("	fsqrtd	d0, d0|");
TTXT("	fstd	d0, [sp]|");
end;


proc	GENSQRT;	\$53, Sqrt function of an integer
begin
COMMENT($53);
PULTOS;
TTXT("	fmsr	s2, r0			@53 SQRT|");
TTXT("	fsitod	d0, s2|");	\float
TTXT("	fsqrtd	d0, d0|");
TTXT("	ftosid	s2, d0|");	\fix
TTXT("	fmrs	r0, s2|");
TOS:=INREG;
end;


proc	GENFSQ;		\$54, Square a real
begin
COMMENT($54);
STKTOS;
TTXT("	fldd	d0, [sp]		@54 FSQ|");
TTXT("	fmuld	d0, d0, d0|");
TTXT("	fstd	d0, [sp]|");
end;	\GENFSQ


proc	GENSQ;		\$55, Square an integer
begin
COMMENT($55);
PULTOS;
TTXT("	mul	r0, r0, r0		@55 SQ|");
TOS:=INREG;
end;


proc GENEND;			\$FE
begin
COMMENT($FE);
TTXT("	.bss|	.align	4|heaplo:	.skip	0x4000000|");	\64 MB
TTXT("	.end|");		\This must match the size set up in natrpi.s
end;


proc GENSTART;
begin
TTXT("	.include	^"natrpi.s^"|	.text|	.align	2|progrm:|");
end;


proc GENNUL;			\$FF
[COMMENT($FF); CRLF(BINDEV)];


proc DSTART;
TTXT("	.data|	.align	2|");


proc DEND;
TTXT("	.text|	.align	2|");

\-------------------------------------------------------------------------------

proc STARTSEG(LEV,SEG);
\START A NEW CODE SEGMENT
int LEV,SEG;

	proc PUTSYMBOL;
	\OUTPUT SYMBOLS
	begin
	TTXT("CSEG");
	if LEV#0 then
		begin
		CHOUT(BINDEV,^_);
		TEXT(BINDEV,FNAME);
		CHOUT(BINDEV,^_);
		INTOUT(BINDEV,SEG);
		end;
	end;

begin
return;		\there is only one code segment in the 32-bit version

TTXT("@curseg	ENDS||");
PUTSYMBOL;
TTXT("	SEGMENT DWORD PUBLIC 'CODE'|	ASSUME	CS:");
PUTSYMBOL;
CRLF(BINDEV);
end;	\STARTSEG


proc DLABEL(N);			\MAKE DATA LABEL
int N;
begin
if HASLAB then GENNUL;
TTXT("|L");
INTOUT(BINDEV,N);
HASLAB:=true;
CHOUT(BINDEV,^:);
end;


proc FLABEL(N);			\(unused) MAKE A FAR CODE LABEL
int N;
begin
STKTOS;
DLABEL(N);
TTXT("	LABEL	FAR|");
LASTOP:=-1;
OLDLEV:=-1;
end;


proc CLABEL(N);			\MAKE A CODE LABEL
int N;
begin
STKTOS;

if HASLAB then GENNUL;
TTXT("|L");
INTOUT(BINDEV,N);
HASLAB:=true;

CHOUT(BINDEV,^:);
LASTOP:=-1;
OLDLEV:=-1;
end;


func NEWLAB;
begin
LABCNT:=LABCNT+1;
return LABCNT;
end;


proc DOGENFIMM(T,V);
int T,V;		\args ignored
int SPC, I;
char P;
begin
DSTART;
SPC:= NEWLAB;
DLABEL(SPC);
P:=addr RLATOM;
TTXT("	.byte	0x"); HEXB(P(0));
for I:= 1, 7 do [TTXT(",0x"); HEXB(P(I))];
CRLF(BINDEV);
DEND;
GENFIMM(^C,SPC);
end;

\======================================================================

proc	LOOKUP;		\LOOKUP IDENTIFIER IN SYMBOL TABLE
\INPUTS: IDENT, HASH
\OUTPUTS: IDTYPE, VAL, LEV, SYMNUM.
\IF TWO IDENTIFIERS OF THE SAME NAME ARE IN THE SYMBOL TABLE
\ THEN THE MOST RECENT ENTRY IS USED.
int	I,K,PNTR;
begin
PNTR:=BOX(HASH);
loop	begin
	if PNTR=EMTPNT then [IDTYPE:=UNDEF; quit];
	I:=0; K:=PNTR;
	while IDENT(I)=SYMBOL(K) & I<SIGCHAR do
		[I:=I+1; K:=K+SYMAX];
	if I=SIGCHAR then	\FOUND
		[IDTYPE:=SYMTYP(PNTR);
		VAL:=SYMVAL(PNTR);
		LEV:=SYMLEV(PNTR);
		SYMNUM:=PNTR;		\(FOR FORWARD PROC)
		quit];
	PNTR:=SYMPNT(PNTR);
	end;
end;	\LOOKUP



proc	INSERT; int STYP,SLEV,SVAL;
\INSERT THE CURRENT IDENTIFIER INTO THE SYMBOL TABLE
\INPUTS:  STYP, SLEV, SVAL, IDENT, HASH, NOSYM, SYMBOL, & BOX.
int	I,K;
begin
LOOKUP;
if IDTYPE#UNDEF then if LEV=LEVEL then \COLLISION\ ERROR(11);
if NOSYM>=SYMAX then \TABLE FULL\ [ERROR(3); NOSYM:=SYMAX-1];
K:=NOSYM;
for I:=0,SIGCHAR-1 do [SYMBOL(K):=IDENT(I); K:=K+SYMAX];
SYMTYP(NOSYM):=STYP;
SYMLEV(NOSYM):=SLEV;
SYMVAL(NOSYM):=SVAL;
SYMPNT(NOSYM):=BOX(HASH);		\LINK BACK
BOX(HASH):=NOSYM;
NOSYM:=NOSYM+1;
end;	\INSERT



proc	GETCON;	\GET A CONSTANT--EITHER BY VALUE OR BY NAME
int	NEG;
begin
if ATOM=^+ then RATOM;
if ATOM=^- then [NEG:=true; RATOM] else NEG:=false;
case ATYPE of
  INTCON: [if NEG then IATOM:=-IATOM; FACTYP:=INTEGER];
  REALCON:[if NEG then RLATOM:=-RLATOM; FACTYP:=REAL];
  IDENTIFIER:
	begin
	LOOKUP;
	case IDTYPE of
	 INCON:	[IATOM:=if NEG then -VAL else VAL;
		FACTYP:=INTEGER];
	 RLCON:	[RLATOM:=if NEG then-RLTBL(VAL)else RLTBL(VAL);
		FACTYP:=REAL]
	other	ERROR(42);
	end
other	ERROR(42);
end;	\GETCON

\-------------------------------------------------------------------------------

proc	CONEXPRESS;	\EVALUATE CONSTANT EXPRESSIONS
\OUTPUTS FACTYP, IATOM, RLATOM
int	SFACTYP,ITEMP,IFVAR;
real	RTEMP;


	proc INTTEST;		\TEST FOR INTEGER ERRORS
	if FACTYP # INTEGER then ERROR(47);


	proc MIXTEST(TYPE);	\TEST FOR MIXED MODE ERRORS
	int TYPE;
	if TYPE#FACTYP then ERROR(46);


	proc	CFACTOR;
	begin
	if ATOM=^( then
		begin
		RATOM;
		CONEXPRESS;
		if ATOM#^) then ERROR(44);		
		end
	else if ATOM=ABSSYM then
		begin					\absolute value function
		RATOM;
		if ATOM#^( then ERROR(39) else RATOM;
		CONEXPRESS;
		if FACTYP=INTEGER then IATOM:= abs(IATOM) else RLATOM:= abs(RLATOM);
		if ATOM#^) then ERROR(44);
		end
	else if ATOM=SQRTYM then
		begin					\square-root function
		RATOM;
		if ATOM#^( then ERROR(39) else RATOM;
		CONEXPRESS;
		if FACTYP=INTEGER then IATOM:= sqrt(IATOM) else RLATOM:= sqrt(RLATOM);
		if ATOM#^) then ERROR(44);
		end
	else if ATOM=SQSYM then
		begin					\square function
		RATOM;
		if ATOM#^( then ERROR(39) else RATOM;
		CONEXPRESS;
		if FACTYP=INTEGER then IATOM:= sq(IATOM) else RLATOM:= sq(RLATOM);
		if ATOM#^) then ERROR(44);
		end
	else if ATOM=SWAPYM then
		begin					\swap bytes function
		RATOM;
		if ATOM#^( then ERROR(39) else RATOM;
		CONEXPRESS;
		if FACTYP#INTEGER then ERROR(47);
		IATOM:= swap(IATOM);
		if ATOM#^) then ERROR(44);
		end
	else if ATOM=REMSYM then
		begin					\remainder
		RATOM;
		if ATOM#^( then ERROR(39) else RATOM;
		CONEXPRESS;
		if FACTYP#INTEGER then ERROR(47);
		IATOM:= REMAINDER;
		if ATOM#^) then ERROR(44);
		end
	else if ATOM=FIXSYM then
		begin					\fix function
		RATOM;
		if ATOM#^( then ERROR(39) else RATOM;
		CONEXPRESS;
		if FACTYP#REAL then ERROR(37);
		IATOM:= FIX(RLATOM);
		FACTYP:= INTEGER;
		if ATOM#^) then ERROR(44);
		end
	else if ATOM=FLOSYM then
		begin					\float function
		RATOM;
		if ATOM#^( then ERROR(39) else RATOM;
		CONEXPRESS;
		if FACTYP#INTEGER then ERROR(47);
		RLATOM:= FLOAT(IATOM);
		FACTYP:= REAL;
		if ATOM#^) then ERROR(44);
		end
	else	GETCON;
	RATOM;
	end;	\CFACTOR



	proc	CSHIFTEXP;
	int	ITEMP;
	begin
	CFACTOR;
	ITEMP:=IATOM;
	case ATOM of
	LSLSYM:	begin
		INTTEST;
		RATOM;
		CFACTOR;
		INTTEST;
		IATOM:=ITEMP << IATOM;
		end;	
	LSRSYM:	begin
		INTTEST;
		RATOM;
		CFACTOR;
		INTTEST;
		IATOM:=ITEMP >> IATOM;
		end;
	ASRSYM:	begin
		INTTEST;
		RATOM;
		CFACTOR;
		INTTEST;
		IATOM:=ITEMP ->> IATOM;
		end
	other	[];
	end;	\CSHIFTEXP



	proc	CTERM;
	int	SFACTYP;
	int	ITEMP;
	real	RTEMP;
	begin
	CSHIFTEXP;
	SFACTYP:=FACTYP;
	loop	begin
		ITEMP:=IATOM; RTEMP:=RLATOM;
		case ATOM of
		  ^*:	begin
			RATOM;
			CSHIFTEXP;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:=ITEMP * IATOM
				else RLATOM:=RTEMP * RLATOM;
			end;
		  ^/:	begin
			RATOM;
			CSHIFTEXP;
			MIXTEST(SFACTYP);
			if IATOM=0 ! RLATOM=0.0 then ERROR(73) \Divide by zero
			else if FACTYP=INTEGER then
				[IATOM:=ITEMP / IATOM;  REMAINDER:=Rem(0)]
			else RLATOM:=RTEMP / RLATOM;
			end
		other	quit;
		end;
	end;	\CTERM



	proc	CALGEXP;	\ALGEBRIAC EXPRESSION
	int	SFACTYP;
	int	ITEMP;
	real	RTEMP;
	begin
	CTERM;
	SFACTYP:=FACTYP;
	loop	begin
		ITEMP:=IATOM; RTEMP:=RLATOM;
		case ATOM of
		  ^+:	begin
			RATOM;
			CTERM;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:=ITEMP + IATOM
			else RLATOM:=RTEMP + RLATOM;
			end;
		  ^-:	begin
			RATOM;
			CTERM;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:=ITEMP - IATOM
				else RLATOM:=RTEMP - RLATOM;
			end
		other	quit;
		end;
	end;	\CALGEXP



	proc	CLOGEXP;	\'NOT' AND COMPARISONS
	int	SFACTYP;
	int	ITEMP;
	real	RTEMP;
	begin
	if ATOM=NOTSYM ! ATOM=^~ then	\UNARY 'NOT' OPERATOR
		begin
		RATOM;
		CLOGEXP;
		INTTEST;
		IATOM:= ~IATOM;
		end
	else	begin
		CALGEXP;
		SFACTYP:=FACTYP;
		ITEMP:=IATOM; RTEMP:=RLATOM;
		case ATOM of
		 ^=:	[RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:= ITEMP = IATOM
			else IATOM:= RTEMP = RLATOM;
			FACTYP:=INTEGER];
		 ^#:	[RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:= ITEMP # IATOM
			else IATOM:= RTEMP # RLATOM;
			FACTYP:=INTEGER];
		 ^>:	[RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:= ITEMP > IATOM
			else IATOM:= RTEMP > RLATOM;
			FACTYP:=INTEGER];
		 ^<:	[RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:= ITEMP < IATOM
			else IATOM:= RTEMP < RLATOM;
			FACTYP:=INTEGER];
		 GESYM:	[RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:= ITEMP >= IATOM
			else IATOM:= RTEMP >= RLATOM;
			FACTYP:=INTEGER];
		 LESYM:	[RATOM;
			CALGEXP;
			MIXTEST(SFACTYP);
			if FACTYP=INTEGER then IATOM:= ITEMP <= IATOM
			else IATOM:= RTEMP <= RLATOM;
			FACTYP:=INTEGER]
		other	[];
		end;
	end;	\CLOGEXP



	proc	CBOOLTERM;	\BOOLEAN "&" EXPRESSIONS
	int	ITEMP;
	begin
	CLOGEXP;
	loop	begin
		ITEMP:=IATOM;
		if ATOM=^& ! ATOM=ANDSYM then
			begin
			INTTEST;
			RATOM;
			CLOGEXP;
			INTTEST;
			IATOM:=ITEMP & IATOM;
			end
		else quit;
		INTTEST;
		end;
	end;	\CBOOLTERM



begin	\CONEXPRESS
TRAP(false);
if ATOM=IFSYM then		\'IF' EXPRESSION
	begin
	RATOM;
	CONEXPRESS;
	INTTEST;
	IFVAR:=IATOM;
	if ATOM#THENYM then ERROR(22);
	RATOM;
	CONEXPRESS;
	SFACTYP:=FACTYP;
	ITEMP:=IATOM; RTEMP:=RLATOM;
	if ATOM#ELSEYM then ERROR(30);
	RATOM;
	CONEXPRESS;
	MIXTEST(SFACTYP);
	if IFVAR then 
		if FACTYP=INTEGER then IATOM:=ITEMP
		else RLATOM:=RTEMP;
	end
else	begin
	CBOOLTERM;
	loop	begin
		ITEMP:=IATOM;
		case ATOM of
		  ^!, ORSYM:
			begin
			INTTEST;
			RATOM;
			CBOOLTERM;
			INTTEST;
			IATOM:=ITEMP ! IATOM;
			end;
		  ^|, XORSYM:
			begin
			INTTEST;
			RATOM;
			CBOOLTERM;
			INTTEST;
			IATOM:=ITEMP | IATOM;
			end
		other	quit;
		end;
	end;
TRAP(true);
if GETERR#0 then ERROR(74);	\GENERAL MATH ERROR
end;	\CONEXPRESS

\-------------------------------------------------------------------------------

fproc	BOOLEXP;



proc	PROCAL(FUNC);
int	FUNC;	\TRUE IF FUNCTION VS PROCEDURE
int	SVAL,SLEV,ARGCNT,SID,CURSYM;
begin
SVAL:=VAL; SLEV:=LEV; SID:=IDTYPE; CURSYM:=SYMNUM;
RATOM;
ARGCNT:=0;
if ATOM=^( then
	begin
	if SID>=INPROC & SID<=RLFPROC then	\NORMAL PROCEDURE CALL
		GENMARK;		\WITH ARGS
	repeat	RATOM; 
		BOOLEXP;
		ARGCNT:=ARGCNT + (if FACTYP=INTEGER then INTSIZE else RLSIZE);
		if FACTYP=REAL & SID#ININT & SID#RLINT then GENFARG;
	until ATOM#^,;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
case of
SID>=INPROC & SID<=RLFPROC:	\NORMAL PROCEDURE CALL
	[if ARGCNT>0 then GENARG(ARGCNT);
	GENCAL(SVAL)]

other case SID of
ININT,RLINT: GENCML(SVAL,SID,FUNC,ARGCNT);	\INTRINSIC PROCEDURE CALL
INOPT,RLOPT:					\OPTIMIZED PROCEDURE CALL
	[if ARGCNT>0 then ERROR(51); 		\NO ARG DECLARED
	GENJSR(SVAL)];
INEXT,RLEXT: GENAEXT(CURSYM,SID,FUNC,ARGCNT);	\EXTERNAL ASSEMBLY CALL
INEPRO,RLEPRO:					\EXTERNAL XPL CALL
	[if ARGCNT>0 then GENARG(ARGCNT);
	GENCEXT(CURSYM,SID,FUNC,ARGCNT)]
other	[];
end;	\PROCAL



proc	BOOLEXP;	\BOOLEAN EXPRESSION
\OUTPUTS FACTOR TYPE (FACTYP)
int	P1,P2,SFACTYP;



proc	FACTOR;



func	STRCON;	\Generate code for a string constant and return its address
int	SPC,	\label number of starting address of string
	CTR,	\count of characters on current line of output
	NCHAR,	\next character (one character look ahead)
	DONE;	\flag: terminating quote mark has been read in


	proc	GETNEXTCH;	\Get next chararacter
	begin
	CHAR:= NCHAR; NCHAR:= CHIN(SRCDEV);
	if NCHAR=EOF then [ERROR(63); exit 1];
	CHOUT(LSTDEV, NCHAR);
	if NCHAR=^^ then
		begin		\convert to control characters, except ^ and Del
		NCHAR:=CHIN(SRCDEV);
		if NCHAR=EOF then [ERROR(63); exit 1];
		CHOUT(LSTDEV,NCHAR);
		if NCHAR>=^@ & NCHAR<=^_ & NCHAR#^^ then NCHAR:=NCHAR-^@;
		if NCHAR>=^` & NCHAR<=^~ then NCHAR:=NCHAR-^`;
		end
	else if NCHAR=^" then	\terminating quote mark
		begin
		if STRTERM \#0\ then CHAR:= CHAR!$80;	\terminate with MSB set
		DONE:= true;
		end;
	end;	\GETNEXTCH


	proc	OCTOUT(N);	\Output 3-digit octal number
	int	N;
	begin
	CHOUT(BINDEV, ^\);
	CHOUT(BINDEV, (N>>6&7)+^0);
	CHOUT(BINDEV, (N>>3&7)+^0);
	CHOUT(BINDEV, (N&7)+^0);
	end;


begin	\STRCON		Enter with CHAR = first character in string (or close ")
COMMENT($FB);
SPC:= NEWLAB;		\make a label at the starting address of the string
DLABEL(SPC);
if CHAR#^" then		\in case of null string (i.e: ""; must be 0 terminated)
	begin
	if CHAR=^^ then
		begin		\convert to control characters, except ^ and Del
		CHAR:=CHIN(SRCDEV);
		if CHAR=EOF then [ERROR(63); exit 1];
		CHOUT(LSTDEV,CHAR);
		if CHAR>=^@ & CHAR<=^_ & CHAR#^^ then CHAR:=CHAR-^@;
		if CHAR>=^` & CHAR<=^~ then CHAR:=CHAR-^`;
		end;
	NCHAR:= CHAR;			\so GETNEXTCH restores CHAR
	DONE:= false;
	CTR:= 0;
	loop	begin
		GETNEXTCH;
		if CTR=0 then TEXT(BINDEV, "	.ascii	^"");	\new output line
		case of
		  CHAR=^":   [OCTOUT(CHAR); CTR:= CTR+4]; \asm delim
		  CHAR=^\:   [OCTOUT(CHAR); CTR:= CTR+4]; \asm meta
		  CHAR>=$20: [CHOUT(BINDEV, CHAR); CTR:= CTR+1];
		  CHAR=$09:  [CHOUT(BINDEV, CHAR); CTR:= CTR+1]	    \show tabs
		other	     [OCTOUT(CHAR); CTR:= CTR+4];
		if DONE then [CHOUT(BINDEV, ^"); quit];
		if CTR >= 40 then
			begin
			CHOUT(BINDEV, ^"); CRLF(BINDEV);
			CTR:= 0;
			end;
		end;	\loop
	CRLF(BINDEV);
	end;
if STRTERM = 0 then TTXT("	.byte	0|");	\terminate with null
GETCH;
FACTYP:= INTEGER;
return SPC;	\return label number for starting address of string
end;	\STRCON



func	ARRAYCON;	\CONSTANT ARRAYS
int	THISEL,NEXTEL,PNTR,SPC,I,INDIRECT,SFACTYP;
def	NULL=$FFFF;
char	ENTRY,R;

	proc ARRAYX;	\(THIS MUST BE AN OPTIMIZED PROCEDURE
	begin		\ FOR THE RESERVE TO WORK PROPERLY.)
	RATOM;
	INDIRECT:=true;
	case ATOM of
	  ^[:	[ENTRY:=ARRAYCON; RATOM];
	  ^":	[ENTRY:=STRCON; RATOM]
	other	begin
		INDIRECT:=false;
		CONEXPRESS;
		if FACTYP=INTEGER then ENTRY:=IATOM
		else	[ENTRY:=RESERVE(RLSIZE);  \FACTYP=REAL
			R:=addr RLATOM;
			for I:=0,RLSIZE-1 do ENTRY(I):=R(I)];
		end;
	NEXTEL:=RESERVE(3*INTSIZE);
	THISEL(1):=ENTRY;
	THISEL(2):=INDIRECT;
	THISEL(0):=NEXTEL;
	NEXTEL(0):=NULL;
	THISEL:=NEXTEL;
	end;


begin	\ARRAYCON
PNTR:=RESERVE(3*INTSIZE);
THISEL:=PNTR;
THISEL(0):=NULL;
ARRAYX;
while ATOM=^, do
	[SFACTYP:=FACTYP;
	ARRAYX;
	if FACTYP#SFACTYP then \MIXED MODE\ ERROR(46)];
if ATOM#^] then ERROR(50);
COMMENT(if FACTYP=INTEGER then $FD else $FC);

SPC:=NEWLAB;
DLABEL(SPC);
while PNTR(0)#NULL do		\DUMP LIST
	begin
	ENTRY:=PNTR(1);
	if FACTYP=INTEGER then
		begin
		TTXT("	.word	");
		if PNTR(2) \INDIRECT\ then CHOUT(BINDEV,^L);
		INTOUT(BINDEV,ENTRY);
		end
	else	begin	\(FACTYP=REAL)
		if PNTR(2) \INDIRECT\ then
			begin
			TTXT("	.word	L");
			INTOUT(BINDEV,ENTRY);
			TTXT("|	.word	0|" );
			end
		else	begin
			TTXT("	.byte	0x"); HEXB(ENTRY(0));
			TTXT(",0x"); HEXB(ENTRY(1));
			TTXT(",0x"); HEXB(ENTRY(2));
			TTXT(",0x"); HEXB(ENTRY(3));
			TTXT(",0x"); HEXB(ENTRY(4));
			TTXT(",0x"); HEXB(ENTRY(5));
			TTXT(",0x"); HEXB(ENTRY(6));
			TTXT(",0x"); HEXB(ENTRY(7));
			end;
		end;
	PNTR:=PNTR(0);
	CRLF(BINDEV);
	end;
return SPC;	\RETURN STARTING ADDRESS OF ARRAY
end;	\ARRAYCON



proc	SPECFAC;	\SPECIAL CHARACTER FACTOR
int	SVAL, SPC, R, SID, SATOM, I;
begin
case ATOM of
  ^(:	[RATOM;				\PARENTHESIZED EXPRESSION
	BOOLEXP;			\(FACTOR TYPE IS UNCHANGED)
	if ATOM#^) then ERROR(44);
	RATOM];
  ^":	[DSTART;			\STRING CONSTANT
	SVAL:= STRCON;
	DEND;
	GENIMM(^L, SVAL);
	RATOM];
  ^[:	[DSTART;			\CONSTANT ARRAY
	SVAL:= ARRAYCON;
	DEND;
	if FACTYP=INTEGER then GENIMM(^L, SVAL)
	else GENFIMM(^L, SVAL);		\FACTYP=REAL
	RATOM];
  ADRSYM, ^@:
	begin				\GET ABSOLUTE HEAP ADDRESS
	SATOM:=ATOM;
	RATOM;
	if ATYPE#IDENTIFIER then ERROR(45);
	LOOKUP;
	case IDTYPE of
	  INVAR, RLVAR, ADDRVAR:
		begin
		SID:= IDTYPE;
		RATOM;
		if ATOM=^( then			\INDEXED
			begin
			if SID=RLVAR & SATOM=^@ then GENIMM(^C,0);
			GENLOD(LEV, VAL);	\(EVEN FOR REALS)
			RATOM;
			BOOLEXP;		\1ST INDEX
			if FACTYP#INTEGER then ERROR(47);
			while ATOM=^, do	\MULTIPLE INDEXING
				begin
				case SID of
				  INVAR, ADDRVAR: GENDBX;
				  RLVAR: GENTRX
				other	 [];
				RATOM;
				BOOLEXP;
				if FACTYP#INTEGER then ERROR(47);
				end;
			case SID of
			  INVAR: GENDBA;
			  RLVAR: begin
				 FACTYP:=INTEGER;
				 if SATOM=^@ then
					[FACTYP:=REAL;
\\					for I:= 1, RLSIZE/INTSIZE-1 do
\\						GENIMM(^C,0);
					];
				 GENTRA;
				 end;
			  ADDRVAR: GENADD
			other	 [];
			if ATOM#^) then ERROR(44) else RATOM;
			end
		else	begin
			FACTYP:=INTEGER;
			if SID=RLVAR & SATOM=^@ then
				[FACTYP:=REAL;
				for I:= 1, RLSIZE/INTSIZE-1 do GENIMM(^C,0)];
			GENADR(LEV, VAL);
			end;
		end;
	  UNDEF:ERROR(10)		\(UNDECLARED NAME)
	other	ERROR(43);		\(VARIABLE EXPECTED)
	end;
  ABSSYM:
	begin				\ABSOLUTE VALUE FUNCTION
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP=INTEGER then GENABS else GENFABS;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
  REMSYM:
	begin				\REMAINDER OF LAST DIVIDE
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP#INTEGER then ERROR(47);
	GENREM;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
  SWAPYM:
	begin				\SWAP BYTES FUNCTION
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP#INTEGER then ERROR(47);
	GENSWAP;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
  EXTSYM:
	begin				\SIGN EXTEND FUNCTION
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP#INTEGER then ERROR(47);
	GENEXT;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
  PORTYM:
	begin				\READ BYTE FROM PORT
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP#INTEGER then ERROR(47);
	GENPIN;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
  FIXSYM:
	begin					\fix function
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP#REAL then ERROR(37);
	GENFIX;
	FACTYP:= INTEGER;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
  FLOSYM:
	begin					\float function
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP#INTEGER then ERROR(47);
	GENFLOAT;
	FACTYP:= REAL;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
  SQRTYM:
	begin					\sqrt function
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP=INTEGER then GENSQRT else GENFSQRT;
	if ATOM#^) then ERROR(44) else RATOM;
	end;
  SQSYM:
	begin					\square function
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP=INTEGER then GENSQ else GENFSQ;
	if ATOM#^) then ERROR(44) else RATOM;
	end
other	\illegal factor\ ERROR(26);
end;	\SPECFAC



proc	IDFAC;		\IDENTIFIER FACTOR
int	SLEV,SVAL,SID,SINDX;
begin	\IDFAC
LOOKUP;
SID:=IDTYPE;
case IDTYPE of
UNDEF:	ERROR(10);

INVAR, RLVAR, ADDRVAR:		\VARIABLE
	begin
	if SID=RLVAR then GENFLOD(LEV,VAL) else GENLOD(LEV,VAL);
	RATOM;
	if ATOM=^( then				\IT IS INDEXED
		begin
		loop	begin
			RATOM;
			BOOLEXP;
			if FACTYP#INTEGER then ERROR(47);
			if ATOM#^, then quit;
			if SID=RLVAR then GENTRI else GENDBX;
			end;
		if ATOM#^) then ERROR(44) else RATOM;
		case SID of
		 INVAR:   GENDBX;
		 RLVAR:   GENTRI;
		 ADDRVAR: GENLDX
		other	  [];
		end;
	end;

INSEG, RLSEG, ADSEG, SHSEG:	\SEGMENT VARIABLE
	begin
	SINDX:=0;			\WATCH INDEX LEVEL
	GENLOD(LEV,VAL);
	RATOM;
	if ATOM=^( then		\HANDLE FIRST INDEX
		begin
		RATOM;
		BOOLEXP;
		if FACTYP#INTEGER then ERROR(47);
		GENDBX;
		if ATOM=^, then	\HANDLE SECOND INDEX
			begin
			SINDX:=2;	\FLAG TWO INDEXES
			RATOM;
			BOOLEXP;
			if FACTYP#INTEGER then ERROR(47);
			case SID of
			  INSEG: GENLDSI;
			  ADSEG: GENLDSB;
			  RLSEG: GENLDSR;
			  SHSEG: GENLSHORT
			other	 [];
			end;
		if ATOM#^) then ERROR(44) else RATOM;
		end;
	\FORCE REAL TO BE INTEGER WHEN THERE ARE LESS THAN 2 INDEXES
	case SID of RLSEG,SHSEG: [if SINDX<2 then SID:=INSEG] other [];
	end;

INCON:	begin	\INTEGER CONSTANT IDENTIFIER
	GENIMM(^C,VAL);
	RATOM;
	end;

RLCON:	begin	\REAL CONSTANT IDENTIFIER
	RLATOM:=RLTBL(VAL); DOGENFIMM(^C,0);
	RATOM;
	end

other	begin	\PROCEDURES USED AS FUNCTIONS (BY DEFAULT)
	PROCAL(true);
	if SID<ININT then
		if SID&1 then GENLOD(0,0) else GENFLOD(0,0);
	end;
FACTYP:=if SID & 1 then INTEGER else REAL; \ODD IDS ARE INTEGER
end;	\IDFAC


begin	\FACTOR
while ATOM=^+ do RATOM;			\IGNORE UNARY "+"
if ATOM=^- then				\UNARY "-"
	begin
	RATOM;
	FACTOR;
	if FACTYP=INTEGER then GENNEG else GENFNEG;
	end
else	case ATYPE of
	SPECIAL:SPECFAC;
	INTCON:	[FACTYP:=INTEGER;	\INTEGER CONSTANT
		GENIMM(^C,IATOM);
		RATOM];
	REALCON:[FACTYP:=REAL;		\REAL CONSTANT
		DOGENFIMM(^C,0);
		RATOM]
	other	IDFAC;	\ATYPE = IDENTIFIER (BY DEFAULT)
end;	\FACTOR



proc	SHIFTEXP;

	proc SHIFTX;
	[if FACTYP # INTEGER then \INTEGER EXPECTED\ ERROR(47);
	RATOM; FACTOR;
	if FACTYP # INTEGER then ERROR(47)];

begin	\SHIFTEXP
FACTOR;
case ATOM of
  LSLSYM: [SHIFTX; GENLSL];	\<<
  LSRSYM: [SHIFTX; GENLSR];	\>>
  ASRSYM: [SHIFTX; GENASR]	\->>
other	  [];
end;	\SHIFTEXP



proc	TERM;
int	SFACTYP;

	proc TERMX; int INOP,RLOP;
	[RATOM; SHIFTEXP;
	if SFACTYP#FACTYP then \MIXED MODE\ ERROR(46);
	GENMTH(if FACTYP=INTEGER then INOP else RLOP)];

begin	\TERM
SHIFTEXP;
SFACTYP:=FACTYP;
loop	case ATOM of
	  ^*:	TERMX(\MUL\$0F,\MULF\$2F);
	  ^/:	TERMX(\DIV\$10,\DIVF\$30)
	other	quit;
end;	\TERM


proc	ALGEXP;		\ALGEBRIAC EXPRESSION
int	SFACTYP;

	proc ALGX; int INOP,RLOP;
	[RATOM; TERM;
	if SFACTYP#FACTYP then \MIXED MODE\ ERROR(46);
	GENMTH(if FACTYP=INTEGER then INOP else RLOP)];

begin	\ALGEXP
TERM;
SFACTYP:=FACTYP;
loop	case ATOM of
	  ^+:	ALGX(\ADD\$0D,\ADDF\$2D);
	  ^-:	ALGX(\SUB\$0E,\SUBF\$2E)
	other	quit;
end;	\ALGEXP


proc	LOGEXP;		\LOGICAL EXPRESSION
int	SFACTYP;

	proc	LOGX; int INOP,RLOP;
	[RATOM; ALGEXP;
	if SFACTYP#FACTYP then \MIXED MODE\ ERROR(46);
	GENCMP(if FACTYP=INTEGER then INOP else RLOP);
	FACTYP:=INTEGER];

begin	\LOGEXP
if ATOM=NOTSYM ! ATOM=^~ then		\UNARY 'NOT' OPERATOR
	[RATOM; LOGEXP;
	if FACTYP#INTEGER then ERROR(47);
	GENNOT]
else	begin
	ALGEXP;
	SFACTYP:=FACTYP;
	case ATOM of
	 ^=:	LOGX(\EQ\$12,\EQF\$32);
	 ^#:	LOGX(\NE\$13,\NEF\$33);
	 ^>:	LOGX(\GT\$15,\GTF\$35);
	 ^<:	LOGX(\LT\$17,\LTF\$37);
	 GESYM:	LOGX(\GE\$14,\GEF\$34);
	 LESYM:	LOGX(\LE\$16,\LEF\$36)
	other	[];
	end;
end;	\LOGEXP


proc	BOOLTERM;	\Boolean "&" expressions
begin
LOGEXP;
loop	begin
	if ATOM=^& ! ATOM=ANDSYM then
		[if FACTYP#INTEGER then ERROR(47);
		RATOM; LOGEXP; GENAND]
	else	quit;
	if FACTYP#INTEGER then ERROR(47);
	end;
end;	\BOOLTERM


proc BEXPX;
begin
if FACTYP # INTEGER then \INTEGER EXPECTED\ ERROR(47);
RATOM; BOOLTERM;
if FACTYP # INTEGER then ERROR(47);
end;	\BEXPX


begin	\BOOLEXP
if ATOM=IFSYM then			\'IF' EXPRESSION
	[RATOM; BOOLEXP;
	P1:=NEWLAB;
	GENJPC(P1);
	if ATOM#THENYM then ERROR(22);
	RATOM; BOOLEXP; SFACTYP:=FACTYP;
	if ATOM#ELSEYM then ERROR(30);
	P2:=NEWLAB;
	GENJMP(P2);
	CLABEL(P1);
	RATOM; BOOLEXP;
	if SFACTYP#FACTYP then \MIXED MODE\ ERROR(46);
	CLABEL(P2)]
else	begin			\BOOLEAN "!" (OR) EXPRESSIONS
	BOOLTERM;
	loop	case ATOM of
		  ^!, ORSYM:	[BEXPX; GENOR];
		  ^|, XORSYM:	[BEXPX; GENEOR]
		other		quit;
	end;
end;	\BOOLEXP

\-------------------------------------------------------------------------------

proc	SSTATEMENT; int SSTK;	\(FOR 'QUIT'S IN 'CASE' STMNTS)


proc	STATEMENT;
int	P2,P3,SFIXS,SLEV,SVAL,SFACTYP,I,DOWNTO;


proc	ASSIGN;		\ASSIGNMENT STATEMENT (ALSO INCLUDES PROCEDURE CALLS)
int	SID;

	proc	ASSX;
	[if ATOM#GETSYM then ERROR(21);
	RATOM;
	BOOLEXP];	\RIGHT-HAND SIDE OF ASSIGNMENT

begin	\ASSIGN
if ATOM=PORTYM then	\port($123):= boolexp
	begin
	RATOM;
	if ATOM#^( then ERROR(39) else RATOM;
	BOOLEXP;
	if FACTYP#INTEGER then ERROR(47);
	if ATOM#^) then ERROR(44) else RATOM;
	ASSX;
	GENPOUT;
	if FACTYP#INTEGER then ERROR(47);
	return;
	end;
if ATYPE#IDENTIFIER then
	\BAD START OF A STATEMENT\ [ERROR(20); SKIPIT; return];
LOOKUP; if IDTYPE=UNDEF then [ERROR(10); SKIPIT; return];
SLEV:=LEV; SVAL:=VAL;	\SAVE THESE FOR "GEN." AN INTERIM
SID:= IDTYPE;		\ BOOLEXP MAY CHANGE LEV & VAL

case of
IDTYPE>=INPROC & IDTYPE<=RLEXT: PROCAL(false);

IDTYPE=INVAR ! IDTYPE=RLVAR ! IDTYPE=ADDRVAR:
	begin
	SFACTYP:=if IDTYPE=RLVAR then REAL else INTEGER;
	RATOM;
	if ATOM=^( then			\INDEXED
		begin
		GENLOD(SLEV,SVAL);
		RATOM;
		BOOLEXP;		\1ST INDEX
		if FACTYP#INTEGER then ERROR(47);

		while ATOM=^, do	\MULTIPLE INDEXING
			begin
			if SFACTYP=INTEGER then GENDBX else GENTRX;
			RATOM;
			BOOLEXP;
			if FACTYP#INTEGER then ERROR(47);
			end;
		if ATOM#^) then ERROR(44) else RATOM;
		case SID of
		 INVAR:   GENDBA;
		 RLVAR:   GENTRA;
		 ADDRVAR: GENADD
		other	  [];

		ASSX;	\TOS NOW POINTS TO ARRAY ELEMENT
		case SID of
		 INVAR:   GENSTD;
		 RLVAR:   GENSTT;
		 ADDRVAR: GENSTX
		other	  [];
		end

	else	[ASSX;
		if SFACTYP=INTEGER then GENSTO(SLEV,SVAL)
		else GENFSTO(SLEV,SVAL)];

	if FACTYP#SFACTYP then \MIXED MODE\ ERROR(46);
	end;

IDTYPE=INSEG ! IDTYPE=RLSEG ! IDTYPE=ADSEG ! IDTYPE=SHSEG:
	begin				\SEGMENT VARIABLES
	SFACTYP:=INTEGER;
	RATOM;
	if ATOM#^( then	[ASSX;GENSTO(SLEV,SVAL)]
	else	begin			\1ST INDEX
		GENLOD(SLEV,SVAL);
		RATOM;
		BOOLEXP;
		if FACTYP#INTEGER then ERROR(47);
		if ATOM=^, then		\2ND INDEX
			begin
			GENDBX;
			RATOM; BOOLEXP;
			if FACTYP#INTEGER then ERROR(47);
			if ATOM#^) then ERROR(44) else RATOM;
			ASSX;	\GET RIGHT SIDE OF ASSIGNMENT
			case SID of
			  INSEG:GENSTSI;
			  RLSEG:[GENSTSR; SFACTYP:=REAL];
			  ADSEG:GENSTSB;
			  SHSEG:[GENSSHORT; SFACTYP:=REAL]
			other	[];
			end
		else	begin
			GENDBA;
			if ATOM#^) then ERROR(44) else RATOM;
			ASSX;	\GET RIGHT SIDE OF ASSIGNMENT
			GENSTD;
			end;
		end;
	if FACTYP#SFACTYP then \MIXED MODE\ ERROR(46);
	end

other \STATEMENT STARTING WITH A CONSTANT\ [ERROR(27); SKIPIT];
end;	\ASSIGN


proc	CASER; int TYPE;
int	SPC1,SPC2,SPC3;


proc	CASER2;
begin
RATOM;
BOOLEXP;
if FACTYP#INTEGER then ERROR(47);
SPC1:=NEWLAB;
if TYPE=$08 then  GENJPC(SPC1) else GENCJP(SPC1);
if ATOM=^, then				\MULTIPLE LABELS
	[SPC3:=NEWLAB;
	repeat	RATOM;	\FOR MORE THAN 2 LABELS
		GENJMP(SPC3);	\2 JUMPS TO STATEMENT
		CLABEL(SPC1);
		BOOLEXP;
		if FACTYP#INTEGER then ERROR(47);
		SPC1:=NEWLAB;
		if TYPE=$08 the GENJPC(SPC1) else GENCJP(SPC1);
	until ATOM#^,;
	CLABEL(SPC3)];
if ATOM#^: then [ERROR(49); SKIPIT; return];
RATOM;
STATEMENT;
end;	\CASER2


begin	\CASER
CASER2;
SPC2:=NEWLAB;
GENJMP(SPC2);		\JUMP OUT OF CASE STATEMENT
CLABEL(SPC1);
while ATOM=^; do
	[CASER2;
	GENJMP(SPC2);	\2 JUMP EXIT (BACK THEN OUT)
	CLABEL(SPC1)];
if ATOM#ELSEYM & ATOM#OTHSYM then ERROR(29);
RATOM;
STATEMENT;
CLABEL(SPC2);
end;	\CASER


begin	\STATEMENT
case ATOM of
BEGSYM,^[:
	begin
	RATOM;
	loop	begin
		if ATOM=ELSEYM then [ERROR(52); RATOM];
		if ATOM=OTHSYM then [ERROR(53); RATOM];
		STATEMENT;
		case ATOM of
		^;:	RATOM;
		ENDSYM:	quit;
		^]:	quit;
		EOF:	[ERROR(62); exit 1]
		other	\SEMI EXPECTED\ ERROR(41);
		end;
	RATOM;		\READ PAST THE 'END'
	end;
CASEYM:	begin				\CASE STATEMENT
	RATOM;
	if ATOM=OFSYM then CASER(\JPC\$08)
	else	begin
		BOOLEXP;
		if FACTYP#INTEGER then ERROR(47);
		if ATOM#OFSYM then [ERROR(48); SKIPIT; return];
		STKLOD:=STKLOD+1;
		CASER(\CJP\$25);
		GENDRP;
		STKLOD:=STKLOD-1;
		end;
	end;
QUITYM:	begin
	for I:=SSTK,STKLOD-1 do GENDRP;
	if FIXCNT>=QUITMAX then [ERROR(4); FIXCNT:=QUITMAX-1];
	FIXES(FIXCNT):=NEWLAB;		\QUIT STATEMENT
	GENJMP(FIXES(FIXCNT));   \(WILL BE "FIXED" AT END OF 'LOOP')
	FIXCNT:=FIXCNT+1;
	RATOM;
	end;
IFSYM:	begin				\IF STATEMENT
	RATOM;
	BOOLEXP;
	P3:=NEWLAB;
	GENJPC(P3);
	if ATOM#THENYM then [ERROR(22); SKIPIT; return];
	RATOM;
	STATEMENT;
	if ATOM=ELSEYM then
		[P2:=NEWLAB;
		GENJMP(P2);
		CLABEL(P3);
		P3:=P2;
		RATOM;
		STATEMENT];
	CLABEL(P3);
	end;
REPSYM:	[P2:=NEWLAB;			\REPEAT STATEMENT
	CLABEL(P2);
	repeat RATOM; STATEMENT until ATOM#^;;
	if ATOM#UNTSYM then [ERROR(28); SKIPIT; return];
	RATOM;
	BOOLEXP;
	GENJPC(P2)];
WHILYM:	[RATOM;				\WHILE STATEMENT
	P2:=NEWLAB;
	CLABEL(P2);
	BOOLEXP;
	P3:=NEWLAB;
	GENJPC(P3);
	if ATOM#DOSYM then [ERROR(23); SKIPIT; return];
	RATOM;
	STATEMENT;
	GENJMP(P2);
	CLABEL(P3)];
RETSYM:	begin				\RETURN STATEMENT
	RATOM;
	for I:=0,STKLOD-1 do GENDRP;
	if ATOM#ELSEYM & ATOM#OTHSYM & ATOM#^; & ATOM#^] &
	    ATOM#ENDSYM & ATOM#UNTSYM then	\STORE THE RETURNED
		[BOOLEXP;			\ VALUE IN GLOBAL #0
		if FACTYP=INTEGER then GENSTO(0,0)
			else GENFSTO(0,0)];
	if OPTPROC then GENRTS else GENRET(LEVEL);
	end;
LOOPYM:	begin				\LOOP STATEMENT
	SFIXS:=FIXCNT;
	RATOM;
	P2:=NEWLAB;
	CLABEL(P2);
	SSTATEMENT(STKLOD);
	GENJMP(P2);
	while FIXCNT>SFIXS do	\FIX THE JUMPS FOR THE 'QUIT'S
		[FIXCNT:=FIXCNT-1; CLABEL(FIXES(FIXCNT))];
	end;
FORSYM:	begin				\FOR STATEMENT
	RATOM;
	if ATYPE#IDENTIFIER then [ERROR(33); SKIPIT; return];
	LOOKUP;
	if IDTYPE=UNDEF then ERROR(10)
	else if IDTYPE#INVAR & IDTYPE#ADDRVAR then ERROR(33);
	SLEV:=LEV; SVAL:=VAL;
	RATOM;
	if ATOM#GETSYM then [ERROR(21); SKIPIT; return];
	RATOM;
	BOOLEXP;
	if FACTYP#INTEGER then ERROR(47);
	GENSTO(SLEV,SVAL);

	DOWNTO:= false;
	if ATOM=TOSYM ! ATOM=^, then []
	else if ATOM=DOWNYM then DOWNTO:= true
	else [ERROR(24); SKIPIT; return];

	RATOM;
	BOOLEXP;
	if FACTYP#INTEGER then ERROR(47);
	if ATOM#DOSYM then [ERROR(23); SKIPIT; return];
	\New For Loop with everything at the end
	P2:=NEWLAB;
	if SLEV#0 then NEWLEVEL(SLEV);
	GENJMP(P2);
	P3:=NEWLAB;
	CLABEL(P3);
	RATOM;
	STKLOD:=STKLOD+1;
	STATEMENT;
	STKLOD:=STKLOD-1;
	GENINP(SLEV,SVAL,DOWNTO);
	CLABEL(P2);
	GENFOR(SLEV,SVAL,P3,DOWNTO);
	end;
EXITYM:	[RATOM;				\EXIT STATEMENT
	if ATOM#ELSEYM & ATOM#OTHSYM & ATOM#^; & ATOM#^] &
	    ATOM#ENDSYM & ATOM#UNTSYM then	\STORE THE RETURNED
		[BOOLEXP;			\ VALUE IN GLOBAL #0
		if FACTYP=INTEGER then GENSTO(0, 0) else GENFSTO(0, 0)];
	GENEXIT];
ELSEYM,OTHSYM,^;,^],ENDSYM,UNTSYM:	[];
EOF:	[]		\(THIS IS MOSTLY AN ACADEMIC POINT)
other	ASSIGN;
end;	\STATEMENT


begin	\SSTATEMENT
\TRICK TO ADJUST STACK (WITH DRP'S) WHEN A 'QUIT' IS IN
\ A 'CASE' STATEMENT.
STATEMENT;
end;	\SSTATEMENT

\-------------------------------------------------------------------------------

proc	PROCEDURE; int SSNOX;	\"SSNO" FROM "PROCDEC" FOR OPTIMIZED PROCEDURES
int	SLEVEL,		\SAVE LEVEL (COMPLICATED BY OPTIMIZED PROCEDURES)
	K,I,J,P1,
	DX,		\HEAP SPACE REQUIREMENT COUNTER (BETWEEN ARRAYS) unsigned
	DXOFF,		\VARIABLE'S OFFSET FROM BASE (=DX IF NO ARRAYS) unsigned
	HAVEGENBASE,	\FLAG: HAVE GENERATED BASE CODE FOR PROCEDURE
	HAVESTART,	\FLAG: HAVE GENERATED START-UP CODE FOR PROGRAM
	OLDCNT,		\OLD SEGMENT COUNT
	FPBASE,		\PC AT END OF DECLARATIONS
	FPROCNT;	\COUNT OF PENDING FORWARD PROCEDURES


proc	EATARGS;	\SKIP ARGUMENTS IN PARENTHESIS
begin
if ATOM=^( then
	begin
	loop	[if CHAR=^) then quit;
		if CHAR=EOL then [ERROR(44); quit];
		GETCH];
	GETCH; RATOM;
	end;
end;	\EATARGS


proc	CODDEC;		\DECLARE INTRINSIC NAMES
int	SID;
begin
SID:=ININT;		\DEFAULT IS INTEGER INTRINSIC
RATOM;
if ATOM=REALYM then [SID:=RLINT; RATOM]
	else if ATOM=INTSYM then RATOM;
if ATYPE#IDENTIFIER then ERROR(45);
while ATYPE=IDENTIFIER do
	[RATOM;
	EATARGS;
	if ATOM#^= then ERROR(40);
	RATOM;
	GETCON; if FACTYP#INTEGER then ERROR(47);
	if IATOM<0 ! IATOM>127 then ERROR(7);
	INSERT(SID,LEVEL,IATOM);
	RATOM;
	if ATOM=^, then RATOM];
if ATOM#^; then ERROR(41) else RATOM;
end;	\CODDEC


proc	CONDEC;		\DECLARE CONSTANT NAMES
int	CNTR,SSNO;
begin
RATOM;
CNTR:=0;
if ATYPE#IDENTIFIER then ERROR(45);
while ATYPE=IDENTIFIER do
	begin
	RATOM;
	if ATOM#^= then [INSERT(INCON,LEVEL,CNTR); CNTR:=CNTR+1]
	else	begin
		SSNO:=NOSYM;
		INSERT(INCON,LEVEL,NORLSY);	\INSERT ID NOW
		RATOM;				\ FIX UP PARMS LATER
		CONEXPRESS;
		if FACTYP=INTEGER then [SYMVAL(SSNO):=IATOM; CNTR:= IATOM+1]
		else	[SYMTYP(SSNO):=RLCON;	\FACTYP=REAL
			if NORLSY>=RLMAX then
				[ERROR(2); NORLSY:=RLMAX-1];
			RLTBL(NORLSY):=RLATOM;
			NORLSY:=NORLSY+1];
		end;
	if ATOM=^, then RATOM;
	end;
if ATOM#^; then ERROR(41) else RATOM;
end;	\CONDEC


proc	CHECKBASE;	\GENERATE BASE INIT CODE FOR PROCEDURE, IF NOT DONE
begin
if not HAVEGENBASE then
	begin
	if not HAVESTART then
		if LEVEL=0 then [GENSTART; HAVESTART:=true];
			\BEWARE OF MORE THAN 64K OF EPROCs
	GENBASE(LEVEL,DX);
	OPTPROC:= false;
	HAVEGENBASE:= true;
	end;
end;	\CHECKBASE


proc	VARDEC; int TYPE;  \DECLARE VARIABLES--INT, REAL & ADDR
int	SDX,DIMS,T,ST;
begin
RATOM;
if ATYPE#IDENTIFIER then ERROR(45);
while ATYPE=IDENTIFIER do
	begin
	INSERT(TYPE,LEVEL,DXOFF);
	SDX:=DXOFF;
	DXOFF:=DXOFF +(if TYPE=RLVAR then RLSIZE else INTSIZE);
	DX:=   DX    +(if TYPE=RLVAR then RLSIZE else INTSIZE);
	T:=if TYPE=RLVAR then RLSIZE else INTSIZE;
	ST:= T;
	if TYPE=ADDRVAR then ST:=1;
	RATOM;
	if ATOM=^( then		\GET DIMENSIONS OF AN ARRAY
		begin
		CHECKBASE;
		if DX#0 then [GENHPI(LEVEL,DX); DX:=0];
		\WARNING: DX can appear negative because of array declarations)
		DIMS:= 0;
		loop	begin
			RATOM;
			CONEXPRESS;
			if FACTYP#INTEGER then ERROR(47);
			\PUSH SIZES OF EACH DIMENSION ON STACK
			GENIMM(^C,IATOM);
			DIMS:= DIMS +1;	\COUNT NUMBER OF DIMENSIONS
			if ATOM # ^, then	\last dimension of char array is
				begin		\ only a single byte per entry
				if ST=1 then T:=T/INTSIZE;
				T:=T*IATOM;
				if ST=1 then T:= T+3 & ~3;
				\keep aligned on 4-byte (word) boundary
				DXOFF:=DXOFF+T;
				quit;
				end;
			T:=T*IATOM;    \ACCUMULATE BYTE COUNT FOR EACH DIMENSION
			DXOFF:=DXOFF+T;	\OFFSET FROM BASE
			end;
		if ATOM#^) then ERROR(44) else RATOM;
		\GEN CODE TO SET UP ARRAY AT RUN TIME, FOR EXAMPLE:
		\GEN CALL MakeArray([3, 5, 7, 10], 4, 2, addr ArrayName);
		GENIMM(^C,DIMS);	\NUMBER OF DIMENSIONS
		GENIMM(^C,ST);		\NUMBER OF BYTES IN EACH ELEMENT
		GENADR(LEVEL, SDX);	\ADDRESS OF POINTER TO ARRAY
		GENARY;
		end;
	if ATOM=^, then RATOM;
	end;
if ATOM#^; then ERROR(41) else RATOM;
end;	\VARDEC



procedure SEGDEC;	\DECLARE SEGMENT VARIABLES
begin
RATOM;
case ATOM of
  INTSYM: VARDEC(INSEG);
  REALYM: VARDEC(RLSEG);
  ADRSYM,CHARYM: VARDEC(ADSEG);
  SHTSYM: VARDEC(SHSEG)
other	  ERROR(72);
end;	\SEGDEC


proc	EXTDEC;int XPL;
\DECLARE EXTERNAL PROCEDURES
\DEFERENTIATE XPL TYPE EXTERNALS FROM ASSEMBLY
int	SID,D;
begin
\HANDLE XPL VS. ASSEMBLY TYPE EXTERNAL
\THE DEFAULT IS INTEGER EXTERNAL
SID:=if XPL then INEPRO else INEXT;

\HANDLE REAL VS INTEGER PROCEDURE
RATOM;
if ATOM=REALYM then [SID:=if XPL then RLEPRO else RLEXT; RATOM]
	else if ATOM=INTSYM then RATOM;
if ATYPE#IDENTIFIER then ERROR(45);

D:=0;
while ATYPE=IDENTIFIER do
	begin
	INSERT(SID,LEVEL,IATOM);
	RATOM;
	EATARGS;
	if ATOM=^, then RATOM;
	if REM(D/4)=0 then TTXT("|	EXTRN	")
		else TTXT(",");

	if XPL then CHOUT(BINDEV, ^_);
	SYMOUT(NOSYM-1); TTXT(":FAR");
	D:=D+1;
	end;
CRLF(BINDEV);
if ATOM#^; then ERROR(41) else RATOM;
end;	\EXTDEC


proc	FPRDEC; int OPTYPE;	\DECLARE FORWARD REFERENCED PROCEDURES
int	SID,I;
begin
SID:=INFPROC;			\DEFAULT TYPE
RATOM;
if ATOM=REALYM then [SID:=RLFPROC;RATOM]
else if ATOM=INTSYM then RATOM;
if ATYPE#IDENTIFIER then ERROR(45);
while ATYPE=IDENTIFIER do
	begin
	RATOM;
	I:=NEWLAB;
	INSERT(SID,LEVEL,I);
	CLABEL(I);
	GENJMP(NEWLAB);
	FPROCNT:=FPROCNT+1;
	EATARGS;
	if ATOM=^, then RATOM;
	end;
if ATOM#^; then ERROR(41) else RATOM;
end;	\FPRDEC


proc	PROCDEC; int CANOPT,PUBLIC;	\DECLARE PROCEDURE NAMES
int	SNOSYM,HASH,I,K,SID,SSNO,SNORL;
begin
SID:=INPROC;			\TYPED PROCEDURE (FOR FUNCTIONS)
RATOM;
if ATOM=REALYM then [SID:=RLPROC; RATOM]
	else if ATOM=INTSYM then RATOM;
if ATYPE#IDENTIFIER then ERROR(45) else LOOKUP;

\SET UP NEXT NESTED SEGMENT
if DEEPER then OLDCNT:=SEGCNT;
DEEPER:=true;
SEGCNT:=SEGCNT+1;
STARTSEG(LEVEL+2,SEGCNT);

if IDTYPE=INFPROC ! IDTYPE=RLFPROC then
	\PROCEDURE HAS BEEN PREVIOUSLY DECLARED BY A 'FPROC' OR 'FFUNCT'
	begin
	if LEVEL#LEV then ERROR(65);	\('FPROC' & 'PROC' MUST BE SAME SCOPE)
	if PUBLIC then			\HANDLE PUBLIC PROCEDURES
		begin
		TTXT("	PUBLIC	_");
		SYMOUT(SYMNUM);
		TTXT("|_");
		SYMOUT(SYMNUM);
		TTXT(":|");
		end;
	CLABEL(VAL+1);
	SYMVAL(SYMNUM):=VAL+1;
	SYMTYP(SYMNUM):=if IDTYPE=INFPROC then INPROC else RLPROC;
	if SID#SYMTYP(SYMNUM) then \MIXED MODE\ ERROR(46);
	if VAL>=FPBASE then FPROCNT:=FPROCNT-1;
	OPTPROC:=false;
	end
else 	begin
	SSNO:=NOSYM;
	I:=NEWLAB;
	INSERT(SID,LEVEL,I);
	if PUBLIC then			\HANDLE PUBLIC PROCEDURES
		begin
		TTXT("	PUBLIC	_");
		SYMOUT(NOSYM-1);
		TTXT("|_");
		SYMOUT(NOSYM-1);
		TTXT(":|");
		end;
	CLABEL(I);
	OPTPROC:=CANOPT;
	end;

\ASSUME FOR NOW THAT PROCEDURE IS NOT OPTIMIZED AND MOVE DOWN A LEVEL
SLEVEL:=LEVEL;
LEVEL:=LEVEL+2; if LEVEL>14 then ERROR(5);
					\EAT THE ARGUMENT LIST AS A COMMENT
while CHAR#^; & CHAR#EOL do GETCH;	\SPECIAL COMMENT STOPS ON EOL
if CHAR#^; then ERROR(41);
GETCH; RATOM;

SNOSYM:=NOSYM; SNORL:=NORLSY;
PROCEDURE(SSNO);	\PASS LOCATION OF PROC NAME IN CASE IT'S OPTIMIZED
if ATOM#^; then ERROR(41) else RATOM;
while NOSYM>SNOSYM do	\RESTORE SYMBOL TABLE TO PREVIOUS LEVEL
\I.E. REMOVE THE IDENTIFIERS WHICH WERE LOCAL TO THIS PROCEDURE
	[NOSYM:=NOSYM-1;
	HASH:=0; K:=NOSYM;
	for I:=0,SIGCHAR-1 do
		[HASH:=HASH+SYMBOL(K); K:=K+SYMAX];
	BOX(HASH&HASHMSK):=SYMPNT(NOSYM)];
NORLSY:=SNORL;
LEVEL:=SLEVEL;

STARTSEG(LEVEL,OLDCNT);			\END CURRENT SEGMENT
DEEPER:=false;
end;	\PROCDEC


begin	\PROCEDURE
DXOFF:= if LEVEL<=0 then RLSIZE else 0;	\SAVE HEAP SPACE FOR RETURN
DX:= DXOFF;
HAVEGENBASE:= false;
HAVESTART:= false;

FPROCNT:=0;
FPBASE:=LABCNT;

repeat
  loop	case ATOM of
	INTSYM:	VARDEC(INVAR);
	ADRSYM:	VARDEC(ADDRVAR);
	CHARYM:	VARDEC(ADDRVAR);
	REALYM:	VARDEC(RLVAR);
	SEGSYM: SEGDEC;
	CODSYM:	CODDEC;
	EXTNYM:	EXTDEC(false);
	DEFSYM:	CONDEC;
	EPRSYM, EFUNYM:	if LEVEL=0 then EXTDEC(true)
			else [ERROR(68); SKIPIT]
	other	quit;

  if not HAVESTART then
	if LEVEL=0 then [GENSTART; HAVESTART:=true];
	\BEWARE OF MORE THAN 64K OF EPROCs

  \SEE IF THERE ARE ANY NESTED PROCEDURES WE NEED TO JUMP OVER
  case ATOM of PUBSYM,PROCYM,FUNSYM,FPRSYM,FFUNYM:
	begin
	P1:=NEWLAB;
	GENJMP(P1);
	end
  other P1:=0\NULL\;

  loop	case ATOM of
	PUBSYM: [if LEVEL#0 then ERROR(68);
		RATOM;
		case ATOM of PROCYM,FUNSYM:PROCDEC(false,true)
		other ERROR(67)];
	PROCYM,FUNSYM:	PROCDEC(true,false);
	FPRSYM,FFUNYM:	FPRDEC(7)
	other	quit;

  if P1#0 then CLABEL(P1);

until	ATOM#INTSYM & ATOM#ADRSYM & ATOM#CHARYM & ATOM#REALYM & ATOM#SEGSYM &
	ATOM#CODSYM & ATOM#EXTNYM & ATOM#DEFSYM & ATOM#EPRSYM & ATOM#EFUNYM;

if DXOFF#0 then OPTPROC:=false;
if OPTPROC then
	begin
	SYMTYP(SSNOX):= if SYMTYP(SSNOX)=INPROC then INOPT else RLOPT;
	LEVEL:=LEVEL-2;	\SAME LEVEL AS NESTING PROCEDURE (TRICKY!)
	GENOPTPROC;
	end
else	begin		\RESERVE SPACE FOR LOCAL VARIABLES (IF NOT ALREADY DONE)
	if not HAVEGENBASE then GENBASE(LEVEL,DX);
	if DX#0 then GENHPI(LEVEL,DX);
	end;

\PROCESS THE BODY
CODFLG:=false;		\IS THERE A MAIN PROCEDURE?
SSTATEMENT(STKLOD);	\(STKLOD WILL ALWAYS BE ZERO HERE)
HASMAIN:=CODFLG;

if OPTPROC then GENRTS else GENRET(LEVEL);

if FIXCNT#0 then \SOME 'QUIT'S NOT IN A 'LOOP'\ ERROR(60);
if FPROCNT#0 then \UNRESOLVED FWD REFERENCES\ ERROR(66);
end;	\PROCEDURE

\------------------------ ROUTINES TO OPEN DOS FILES ---------------------------

string 0;
proc	DOSOPEN;
int	T, I, Ch;
char	CMDTAIL;

	func	GETSWT;		\FIND, REMOVE AND RETURN A SWITCH
	int	P, T;
	begin
	for P:= 1, CMDTAIL(0) do
	    if CMDTAIL(P)=^- then
		begin
		CMDTAIL(P):= $20;
		if P < CMDTAIL(0) then
			begin
			T:= CMDTAIL(P+1);
			if T>=^a & T<=^z then T:= T-$20; \MAKE UPPERCASE
			CMDTAIL(P+1):= $20;
			return T;
			end;
		end;
	return 0;
	end;	\GETSWT


	proc	PARSE;		\PARSE COMMAND TAIL AND SET I/O HANDLES
	char	EXTIN, EXTOUT;
	int	P, P0, EXTFLG, I;
	begin
	EXTIN:= ".xpl";   EXTOUT:= ".s";
	P:= 1;
	EXTFLG:= false;
	loop	begin			\PARSE COMMAND TAIL FOR EXTENSION
		if CMDTAIL(P) = ^. then
			begin
			EXTFLG:= true;
			quit;
			end;
		if CMDTAIL(P) = ^; then quit;	\IGNORE SEMICOLON
		if P > CMDTAIL(0) then quit;	\if quit then P points to 0
		P:= P+1;
		end;

	\Back up over any trailing spaces (caused by switches)
	repeat P:= P-1 until CMDTAIL(P) # ^ ;
	P:= P+1;

	if not EXTFLG then		\SET EXTENSION AND INPUT HANDLE
		for I:= 0, 4 do CMDTAIL(P+I):= EXTIN(I);
	INHAND:= FOPEN(CMDTAIL+1, 0);

	for I:= 0, 2 do CMDTAIL(P+I):= EXTOUT(I);  \SET EXTENSION AND OUTPUT HANDLE
	OUTHAND:= FOPEN(CMDTAIL+1, 1);

	\COPY FILE NAME INTO FNAME
	P0:= P-1;			\P0 points to last character in file name
	repeat	P:= P-1;
	until	CMDTAIL(P)=^  ! CMDTAIL(P)=^\ ! CMDTAIL(P)=^:;

	P:= P+1;			\point to first character of file name
	for I:= 0, 7 do
		begin
		FNAME(I):= CMDTAIL(P+I);
		if P+I=P0 ! I=7 then
			begin
			FNAME(I):= FNAME(I) ! $80;
			I:= 7;
			end;
		end;
	end;	\PARSE


begin	\DOSOPEN
CMDTAIL:= RESERVE($80+4);	\GET COMMAND TAIL FROM PSP
I:= 1;				\get file name from command line
repeat	Ch:= CHIN(8);
	CMDTAIL(I):= Ch;
	I:= I+1;
until	Ch=0 ! Ch=EOF;
CMDTAIL(0):= I-2;	\number of characters in CmdTail not including last 0

loop	begin				\HANDLE SWITCHES
	T:= GETSWT;
	case T of
	  ^L:	LSTDEV:= 0;
	  ^C:	COMFLG:= true;
	  ^A:	[BINDEV:= 0; LSTDEV:= 7];
	   0:	quit			\NO MORE SWITCHES ON COMMAND LINE
	other	begin
		TEXT(TV,"UNRECOGNIZED SWITCH: -"); CHOUT(TV, T); CRLF(TV);
		TEXT(0, "Usage: xplr [options] source(.xpl)
   -a: Display Assembly code
   -c: Include i2l Comments in output file
   -l: Display source code Listing
");
		exit 1;
		end;
	end;
PARSE;					\PARSE COMMAND LINE AND SET HANDLES
FSET(INHAND, ^I);
FSET(OUTHAND, ^O);
end;	\DOSOPEN
string 1;

\-------------------------------------------------------------------------------

begin	\MAIN--DISPLAY TITLE AND INITIALIZE
IDENT:=RESERVE(SIGCHAR);
FIXES:=RESERVE(QUITMAX*INTSIZE);
SYMBOL:=RESERVE(SIGCHAR*SYMAX);	\SYMBOL TABLE
SYMTYP:=RESERVE(SYMAX);
SYMVAL:=RESERVE(SYMAX*INTSIZE);
SYMLEV:=RESERVE(SYMAX);
SYMPNT:=RESERVE(SYMAX*INTSIZE);
BOX:=RESERVE(BOXNUM*INTSIZE);	\HASH TABLE
RLTBL:=RLRES(RLMAX);
HEXDIGIT:="0123456789ABCDEF ";

\SET INCLUDE ARRAYS
OLDHAN:=RESERVE(HANMAX*INTSIZE);
HANPTR:=0;

\SET DEFAULT CHANNELS
COMFLG:=false;

TEXT(TV,"
-- XPL0 NATIVE COMPILER, VER R2.2 --
    COPYRIGHT 2018 P.J.R. BOYLE

XPL0 comes with ABSOLUTELY NO WARRANTY.
This is free software. You are welcome and encouraged to redistribute
it under certain conditions. For details see LICENSE.TXT.

");
\OPEN DOS FILES
FNAME:=RESERVE(8);
BINDEV:= 3; LSTDEV:= 8; SRCDEV:= 3;
DOSOPEN;
OPENO(BINDEV); OPENO(LSTDEV); OPENI(SRCDEV);

LABCNT:=0; \LEVEL:=0;\			\INITIALIZE SOME STUFF
STKLOD:=0; NOSYM:=0; NORLSY:=0; FIXCNT:=0;
for II:=0,BOXNUM do BOX(II):=EMTPNT;	\ZERO THE SYMBOL TABLE
ERRCNT:=0;
LASTOP:=$FFFF\NUL\;
HASLAB:=false;

\INTIALIZE SEGMENT STUFF
SEGCNT:=0;
DEEPER:=true;

LOCAL:=0;
OLDLEV:=-1;
TOS:=STACKED;		\THE DEFAULT CASE;
CONDITIONAL:= true;
STRTERM:=-1;		\default to MSB string termination (nonzero)
REMAINDER:= 0;

\GETCH; RATOM;

LEVEL:= -1;		\automatically include 'code' declarations
INCLCODE;		\ such that they can be overridden by
GETCH; RATOM;		\ by explicit 'code' declarations

OPTPROC:=false;		\(FOR 2 REASONS)
PROCEDURE(0);		\COMPILE MAIN PROCEDURE, I.E. THE PROGRAM
while ATOM=^; do RATOM;
if ATOM#EOF then \MORE CODE AFTER END\ [ERROR(61); PROCEDURE(0)];

\IF PROGRAM HAS CODE IN 'MAIN' THEN MAKE IT PUBLIC
\if HASMAIN then TTXT("	PUBLIC	PROGRM|");
GENEND;

CLOSE(BINDEV);

CRLF(LSTDEV);
TEXT(LSTDEV,"ERRORS DETECTED: "); INTOUT(LSTDEV,ERRCNT); CRLF(LSTDEV);
CLOSE(LSTDEV);
CRLF(TV);

FCLOSE(OUTHAND);	\CLOSE DOS FILES
FCLOSE(INHAND);
return if ERRCNT#0 then 1 else 0;
end;	\MAIN
