@ nat.s	27-Feb-2018
@
@	**********************************************************
@	**							**
@	**     32-BIT XPL0 NATIVE LANGUAGE RUNTIME PACKAGE	**
@	**		    FOR RASPBERRY PI			**
@	**							**
@	**	       COPYRIGHT LOREN BLANEY 2018		**
@	**     BASED ON CODE BY LARRY FISH AND P.J.R. BOYLE	**
@	**	     CONTRIBUTIONS BY RICHARD OTTOSEN		**
@	**							**
@	**********************************************************
@
@ This program is free software; you can redistribute it and/or modify it under
@  the terms of the GNU General Public License version 2 as published by the
@  Free Software Foundation.
@ This program is distributed in the hope that it will be useful, but WITHOUT
@  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
@  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
@  details.
@ You should have received a copy of the GNU General Public License along with
@  this program (in the file LICENSE.TXT); if not, write to the Free Software
@  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
@
@ You can reach the authors at: loren.blaney@gmail.com

natrpi	= 0	@ set to 1 to configure for unoptimized "x" version: natrpi.s

@ If the above "natrpi" is set 'true' then this file gets included at the
@  beginning of compiled xpl code, which is then assembled like this:
@  gcc -o xplname xplname.s -lm
@ Otherwise this file is first made into a library like this:
@  rm libnat.a
@  gcc -c nat.s
@  ar r libnat.a nat.o /usr/lib/libasound.so
@ And then a compiled xpl program is assembled and linked to it like this:
@  gcc -o xplprog xplprog.s -lnat -lm
@ Note that the sound intrinsics only work when the library configuration is
@  used. Also note that gcc is merely used as an assembler for xpl programs.
@
@ REVISIONS:
@ 14-Mar-2014: Add transparency feature in Point intrinsic for 32-bit color.
@  Add GetDateTime and InsertKey intrinsics. Change Mickeys to MouseMove.
@  Fix default attribute colors for device 6 in 16-bit graphics mode. Device 0
@  also outputs colored text as specified by the Attrib intrinsic.
@ 29-Mar-2014: Double display memory (scrheight2) to support page flipping.
@  Fix Restart so stack remains balanced.
@ 03-Oct-2015: GetKey intrinsic recognizes more special keys. Add Hilight,GetFB
@  OpenO(6) resets attribute to white on black, and homes the cursor. Close(3)
@  won't rewrite the large buffer to the output file if it's called a 2nd time.
@  Arguments from command line are separated by spaces not zeros. Added Font14.
@  Added Sound intrinsic, but it only provides a delay. Disable Attrib for
@  device 0. Add MAlloc and Release intrinsics. Change interface for GetMouse
@  and GetMouseMove so that "@" is not used to pass arguments. Fixed bug in
@  Line intrinsic where it wouldn't properly draw a line one pixel long. Change
@  last 8 palette colors to black (0) instead of white. Add WaitForVSync,
@  DisplayPage, CopyMem, and FillMem intrinsics. Intrinsic 71 is Text for
@  null-terminated strings (like intr12a). Use offset from base in mkarray.
@  Double-word align stack for (insane) localtime in GetDateTime intrinsic.
@  IntOut works for $8000_0000 because div routine is now 32, not 31, bits.
@  Save argc and argv in memory because Restart can alter r4 and r5. Speed up
@  Point intrinsic slightly. FOpen (29) ignores any leading space characters.
@ 13-Oct-2015: Change Point and ReadPix to be compatible with Raspbian's new
@  order of red and blue bytes in 24 & 32-bit color modes. Doubled speed of
@  Point intrinsic for 24-bit color.
@ 14-Oct-2015: Disable use of mailbox.
@ 12-Aug-2016: Convert to a linkable library enabling "asound" to be used.
@  Fix Sound intrinsic (39) and add PlaySound intrinsic (103). ChOut(0,7) rings
@  the bell. Add Pause key to GetKey. Conditional assembly produces original
@  "natrpi.s" version. SetVid recognizes all CGA, EGA, and VGA modes; and mode 3
@  restores the original screen resolution when program started. yres_virtual:=
@  yres at exit prevents startx having a blank screen. Flashing cursor for
@  ChIn(1) is consistent with PC. Set correct end point for dotted lines. Option
@  for transparent backgrounds for output device 6. ChIn(6) echoes to device 6.
@  SetWind intrinsic. Device 6 cursor (penx, peny) tracks device 0's position.
@  Faster divide routine. Distribute with version 1.0 xpl0.
@ 29-May-2017: Set bytesPerLine in setfbp when it's 0 because Raspbian & Jessie
@  fail to do it. Return 0 for Ran(0). Suppress error messages for PlaySoundFile
@  Stricter checking for illegal output device numbers. Handle non-ASCII keys
@  in chin1 (like PC version, which returns a zero prefix). Pause key works for
@  chin1. (GetKey intrinsic should be avoided because it's not in PC versions.)
@  chin0 ignores non-ASCII keys, such as arrows, but echoes Esc key (like EXPL).
@  Characters stop being echoed when line buffer is full. FOpen now terminates
@  a filename with space, as well as comma, and any char <$21 or >$7E. RlOut of
@  infinite values no longer hangs but instead displays "INF". Add more VESA
@  modes. 32-bit graphic modes draw alpha transparency lines not dashed lines.
@  SetFB (as well as SetVid) reset the attribute colors to white on black for
@  all graphic depths and reset the pen position to 0,0. Text is not displayed
@  beyond bottom of window set by SetWind in clip mode. Graphic pen does not
@  track device 0 except for form feed (and Cursor). Fix bug where horizontal
@  lines could draw off-screen causing segmentation fault. Mask characters sent
@  to chout6 to a single byte. OpenO($106) is legal. Hilight preserves pen posn.
@  The graphic pen position is accessible via GetFB. Illegal arguments for
@  Sqrt and Fix cause error traps. Added ShowMouse intrinsic (77). Device 0 no
@  longer inserts a CR before an LF; it's done by the Linux terminal handler.
@  Reap zombies spawned by PlaySoundFile so sounds don't stop. Implement RawText
@  intrinsic. Clear sets the cursor position to the upper-left corner for device
@  6 but not device 0. OpenMouse (85) occurs automatically. Add MoveMouse (78).
@ 30-Jun-2017: Add GetPalette (80). Fixed bug in ReadPix (44) when scrdepth=32.
@  Chout6 scrolls when writing to lower-right corner.
@ 27-Feb-2018: Chin0 echoes Ctrl chars as ^char (e.g:^A). Backspace deletes back
@  two spaces for ctrl chars. Linux's new-line (LF) and Ctrl+Z flush kb buffer.
@  Chin0 moves device 6's pen position to start of next line. Make sure flashing
@  cursor is on at exit. Add SetHexDigits intrinsic (104) and rewrite HexOut.
@  SetFB sets fontheight. Add intrinsics Floor (64), Ceil (65), and Pow (66).
@  When Attrib background = foreground then don't draw character's background.
@  Add Esc, arrow keys, etc. to GetShiftKeys (93) intrinsic. Fast horizontal
@  lines for all graphic depths (not just 8-bit). Removed code that changed
@  Attrib colors when in 16-bit depth to match PC's 4-bit depth colors. Improve
@  Hilight intrinsic (72).
@  
@ CODING CONVENTIONS:
@ Intrinsics assume that "the top of stack" (tos) is in r0, with any additional
@  arguments on the actual hardware stack. Intrinsics used as functions return
@  a value in r0. Intrinsics may alter registers r0..r3, but all other registers
@  must remain unchanged.
@
@ Register usage:
@	r0 =	top of XPL's stack
@	r1-r3 =	scratch (not preserved by calls to subroutines)
@	r8 =	base address of our (nat.s) variables
@	r9 =	heap pointer (hp)
@	r10 =	base of XPL's global variables (heaplo)
@	r11 =	base of XPL's local variables (aka fp, x86's esi)
@	r12 =	scratch (not preserved by calls to Linux kernel)
@	r13 =	stack pointer (sp)
@	r14 =	link register (lr)
@	r15 =	program counter (pc)
@
	.if	!natrpi

@ The optimized XPL0 compiler links to these (mostly) intrinsic routines:

	.global	intr0,	intr1,	intr2,	intr3,	intr4,	intr5,	intr6,	intr7
	.global	intr8,	intr9,	intr10, intr11, intr12, intr13, intr14, intr15
	.global	intr16, intr17, intr18, intr19, intr20, intr21, intr22, intr23
	.global	intr24, intr25, intr26, intr27, intr28, intr29, intr30, intr31
	.global	intr32, intr33, intr34, intr35, intr36, intr37, intr38, intr39
	.global	intr40, intr41, intr42, intr43, intr44, intr45, intr46, intr47
	.global	intr48, intr49, intr50, intr51, intr52, intr53, intr54, intr55
	.global	intr56, intr57, intr58, intr59, intr60, intr61, intr62, intr63
	.global	intr64, intr65, intr66, intr67, intr68, intr69, intr70, intr71
	.global	intr72, intr73, intr74, intr75, intr76, intr77, intr78, intr79
	.global	intr80, intr81, intr82, intr83, intr84, intr85, intr86, intr87
	.global	intr88, intr89, intr90, intr91,	intr92, intr93, intr94, intr95
	.global	intr96, intr97, intr98, intr99, intr100,intr101,intr102,intr103
	.global	intr104
	.global	intr12a, intr255, mkarray, sdivx

	.endif

	.include	"nat.inc"	@ common variable declarations

@ Define some ASCII characters
BEL	=	0x07			@ bell
CR	=	0x0D			@ carriage return
LF	=	0x0A			@ line feed
EOF	=	0x1A			@ end of file
ESC	=	0x1B			@ escape

	.text
	.align	2

@-------------------------------------------------------------------------------
@ Signal handler to clear any terminated child status info from process table.
@ If this is not done, the table can fill to a limit that prevents additional
@  processes from starting. This occurs with many calls to PlaySoundFile.
@
reapzombies:
	push	{lr}
reapzom:mov	r0, #-1			@ waitpid(-1, NULL, WNOHANG)
	mov	r1, #0
	mov	r2, #1
	bl	waitpid
	cmp	r0, #0			@ loop until no more process IDs
	bgt	reapzom
	pop	{pc}

@===============================================================================
@				ENTRY POINT
@===============================================================================

	.text
	.align	2
	.global main
main:	mov	r4, r0			@ save r0 = argc
	mov	r5, r1			@ save r1 = argv

	push	{fp, lr}
	add	fp, sp, #4		@ struct termios orig;
	sub	sp, sp, #64

	sub	r3, fp, #64		@ save original terminal configuration
	mov	r0, #0			@ ioctl(0, 0x5401, &orig); (tcgetattr)
	ldr	r1, srdata
	mov	r2, r3
	bl	ioctl

	bl	setraw			@ set terminal into raw mode

	push	{r11}			@ r11 = fp
	ldr	r8, dbase
	mov	r0, #0			@ no errors yet
	str	r0, [r8, #errnum-dseg]
	mov	r0, #-1			@ trap all errors
	str	r0, [r8, #traps-dseg]

	bl	setfbp			@ set frame buffer pointer for device 6
	bl	intr85			@ OpenMouse
@	bl	openmailbox		@ open mailbox

	str	r4, [r8, #argc-dseg]	@ save because restart can bomb r4 & r5
	str	r5, [r8, #argv-dseg]

	mov	r0, #17			@ set up zombie reaper
	ldr	r1, reapzp		@ signal(SIGCHLD, reapzombies)
	bl	signal

	ldr	r0, pkd20		@ get file descriptor for GetShiftKeys
	mov	r1, #0			@ kdfd = open("/dev/input/event0",
	bl	open			@        O_RDONLY);
	str	r0, [r8, #kdfd-dseg]

restart:adr	r0, hlo			@ get address of item in code segment
	ldr	r10, [r0]		@ r10 = base address of global variables
	mov	r11, r10		@ r11 = base address of local variables
	mov	r9, r10			@ r9 = heap pointer
	mov	r0, r9			@ set up 64 MB of heap space
	add	r0, #0x4000000		@ this must match size set in xplr.xpl
	str	r0, [r8, #heaphi-dseg]

	bl	setint			@ initialize intrinsics

@ copy anything on the command line into pseudo device 8
	ldr	r4, [r8, #argc-dseg]
	ldr	r5, [r8, #argv-dseg]
main20:	cmp	r4, #2			@ is there something after prog name?
	blt	main30			@ skip if not
	ldr	r6, [r5, #4]		@ get pointer to argument
main25:	ldrb	r0, [r6]		@ get character
	cmp	r0, #0			@ was it the terminating zero?
	beq	main27			@ skip if so
	 add	r6, #1			@ bump pointer
	 bl	psout			@ write char in r0 to device 8
	 b	main25			@ loop
main27:	mov	r0, #0x20		@ separate arguments with spaces
	bl	psout			@ (this actually terminates with spaces)
	sub	r4, #1			@ decrement argc counter
	add	r5, #4			@ point to next argument
	b	main20			@ loop
main30:
	adr	r0, noblank		@ disable screen saver blanking
	bl	text

	bl	startxpl		@ run the compiled XPL code
	str	r0, [r10, #0]		@ save exit value, if any, in global 0

	mov	r0, #0			@ make sure mouse pointer is off
	bl	intr77			@ ShowMouse(false)

	mov	r0, #-1			@ make sure flashing cursor is on
	bl	intr88			@ ShowCursor(true)

	adr	r0, doblank		@ re-enable screen saver blanking
	bl	text

	ldr	r0, [r8, #errnum-dseg]	@ did a runtime error occur?
	cmp	r0, #0
	beq	main90			@ skip if not

	adr	r0, em0			@ else display runtime error message
	bl	text			@ e.g: RUN-TIME ERROR 5: BAD INTRINSIC
	mov	r0, #0
	str	r0, [r8, #nowdev-dseg]
	ldr	r0, [r8, #errnum-dseg]	@ display the number
	bl	intout
	ldr	r1, [r8, #errnum-dseg]
	sub	r1, #1
	adr	r0, errmsg
	add	r0, r1, lsl #2
	ldr	r0, [r0]
	bl	text
	bl	crlf
main90:	bl	crlf			@ make Linux prompt at start of line

	bl	nopage2
	pop	{r11}
	sub	r3, fp, #64		@ restore original terminal config.
	mov	r0, #0			@ ioctl(0, 0x5404, &orig); (tcsetattr)
	ldr	r1, srdata+4
	mov	r2, r3
	bl	ioctl

	ldr	r0, [r10, #0]		@ get exit value, if any, from global 0
	sub	sp, fp, #4
	pop	{fp, pc}

reapzp:	.word	reapzombies
dbase:	.word	dseg
hlo:	.word	heaplo

noblank:.byte	ESC			@ disable screen saver blanking
	.asciz	"[9;0]"
doblank:.byte	ESC			@ disable screen saver blanking
	.asciz	"[9;30]"		@ enable 30-minute screen saver blanking
	.align	2

pkd20:	.word	kd20			@ key down for GetShiftKeys
kd20:	.ascii	"/dev/input/event0\000"
	.align	2

@-------------------------------------------------------------------------------
@ Make sure that yres_virtual = yres else startx will display a blank screen.
@
nopage2:stmfd	sp!, {r3, r4, r5, lr}
	mov	r1, #2			@ fbfd = open("/dev/fb0", O_RDWR);
	ldr	r0, .L302
	bl	open
	ldr	r4, .L302+4		@ ioctl(fbfd, FBIOGET_VSCREENINFO,
	mov	r1, #0x4600		@	&vinfo2);
	mov	r2, r4
	mov	r5, r0
	bl	ioctl
	ldr	r3, [r4, #4]		@ vinfo.yres_virtual = vinfo.yres;
	mov	r0, r5			@ ioctl(fbfd, FBIOPUT_VSCREENINFO,
	ldr	r1, .L302+8		@ 	&vinfo2);
	mov	r2, r4
	str	r3, [r4, #12]
	ldmfd	sp!, {r3, r4, r5, lr}
	b	ioctl			@ call and return
.L302:	.word	.L30C0
	.word	vinfo2
	.word	0x4601
	.comm	vinfo2,160,4
.L30C0:	.ascii	"/dev/fb0\000"
	.align	2

@-------------------------------------------------------------------------------
@ Set terminal into raw mode: noncanonical mode with all input and output
@  processing disabled.
@
setraw:	push	{fp, lr}
	add	fp, sp, #4		@ struct termios t;
	sub	sp, sp, #64

	sub	r3, fp, #64		@ ioctl(0, 0x5401, &t);
	mov	r0, #0
	ldr	r1, srdata
	mov	r2, r3
	bl	ioctl

@ Noncanonical mode, disable signals, extended input processing, and echoing
@ t.c_lflag &= ~(ICANON | ISIG | IEXTEN | ECHO);
	ldr	r3, [fp, #-52]
	bic	r3, r3, #0x8000
	bic	r3, r3, #0x0B
	str	r3, [fp, #-52]

@ Disable special handling of CR, NL, and BREAK.
@ No 8th-bit stripping or parity error handling.
@ Disable START/STOP output flow control. */
@ t.c_iflag &= ~(BRKINT | ICRNL | IGNBRK | IGNCR | INLCR | INPCK | ISTRIP |
@ 		IXON | PARMRK);
	ldr	r3, [fp, #-64]
	bic	r3, r3, #0x05F0
	bic	r3, r3, #0x0B
	str	r3, [fp, #-64]
@	ldr	r3, [fp, #-60]		@ Disable all output processing
@	bic	r3, r3, #1		@ t.c_oflag &= ~OPOST;
	mov	r3, #5			@ enable ONLCR and OPOST
	str	r3, [fp, #-60]		@ t.c_oflag = ONLCR | OPOST

	mov	r3, #1			@ Character-at-a-time input
	strb	r3, [fp, #-41]		@ t.c_cc[VMIN] = 1;

	mov	r3, #0			@ with blocking
	strb	r3, [fp, #-42]		@ t.c_cc[VTIME] = 0;

	sub	r3, fp, #64		@ ioctl(0, 0x5404, &t);
	mov	r0, #0
	ldr	r1, srdata+4
	mov	r2, r3
	bl	ioctl

	sub	sp, fp, #4
	pop	{fp, pc}

srdata:	.word	0x5401
	.word	0x5404

@-------------------------------------------------------------------------------
@ Start the compiled XPL program.
@
startxpl:
	push	{lr}
	str	sp, [r8, #stkptr-dseg]	@ save stack pointer for 'exit' command
	pop	{lr}			@ (weird! -- beware of interrupts?)
	b	progrm			@ go to compiled XPL code

@-------------------------------------------------------------------------------
@ Runtime error handler. This does not alter any registers.
@  r0 contains the error number
@
error:	push	{r0-r1, lr}		@ don't alter any registers
	str	r0, [r8, #errnum-dseg]	@ save the error number
	ldr	r1, [r8, #traps-dseg]	@ get trap flags
	lsrs	r1, r1, r0		@ (Larry's way)
	bcc	err90			@ if bit was clear then no trap--return
	ldr	sp, [r8, #stkptr-dseg]	@ trap error, i.e. abort program
	pop	{pc}			@ returns from anywhere
err90:
	pop	{r0-r1, pc}		@ restore registers and return

@-------------------------------------------------------------------------------
@ Display the zero-terminated string pointed to by r0 on the terminal.
@
text:	push	{r4, lr}
	mov	r4, r0			@ put pointer into preserved register
	b	txt20			@ enter loop
txt10:	 bl	putcharnow
txt20:	ldrb	r0, [r4]		@ get char
	add	r4, #1			@ increment pointer
	cmp	r0, #0			@ last char?
	bne	txt10
	pop	{r4, pc}

@-------------------------------------------------------------------------------
@ Output a carriage return and a line feed to the console.
@
crlf:	push	{lr}
	mov	r0, #CR
	bl	putcharnow
	mov	r0, #LF
	bl	putcharnow
	pop	{pc}

@-------------------------------------------------------------------------------
@ Error messages:
@
em0:	.byte	BEL, CR, LF
	.asciz	"RUN-TIME ERROR "
	.align	2
errmsg:	.word	em1, em2, em3, em4, em5, em6, em7, em8, em9, em10, em11, em12
	.word	em13, em14, em15
em1:	.asciz	": DIV BY 0"
em2:	.asciz	": OUT OF MEMORY"
em3:	.asciz	": I/O"
em4:	.asciz	": BAD OPCODE"
em5:	.asciz	": BAD INTRINSIC"
em6:	.asciz	": DIV BY 0.0"
em7:	.asciz	": OVERFLOW"
em8:	.asciz	": UNDERFLOW"
em9:	.asciz	": FIX OVERFLOW"
em10:	.asciz	": SQRT < 0.0"
em11:	.asciz	": LOG <= 0.0"
em12:	.asciz	": EXP OVERFLOW"
em13:	.asciz	": OVERFLOW"
em14:	.asciz	": ATAN2(0.0,0.0)"
em15:	.asciz	": VIDEO MODE"
	.align	2

@-------------------------------------------------------------------------------
@ Point framebuf to current display memory, and set up screen parameters.
@ Get "Segmentation fault" when attempting to write to the framebuf
@ provided by the system. So what this actually does is to set up another
@ chunk of memory the same size as the system framebuffer. This enables
@ writing to device 6 without getting a Segmentation fault. Device 6 and
@ device 0 use seprate chunks of memory that are both displayed
@ simultaneously. The Clear intrinsic only clears characters written to
@ device 6.
@
setfbp:	stmfd	sp!, {fp, lr}
	add	fp, sp, #4
	sub	sp, sp, #32

	ldr	r0, .LL4		@ fbfd = open("/dev/fb0", O_RDWR);
	mov	r1, #2
	bl	open
	str	r0, [fp, #-12]
	str	r0, [r8, #fbfd-dseg]

	ldr	r0, [fp, #-12]		@ ioctl(fbfd, FBIOGET_VSCREENINFO,
	mov	r1, #0x4600		@				&vinfo);
	ldr	r2, .LL4+4
	bl	ioctl
	ldr	r3, .LL4+12
	ldr	r3, [r3, #44]
	str	r3, [r8, #bytesPerLine-dseg]
	ldr	r0, [fp, #-12]		@ ioctl(fbfd, FBIOGET_FSCREENINFO,
	ldr	r1, .LL4+8		@				&finfo);
	ldr	r2, .LL4+12
	bl	ioctl

@ map fb to user memory
	ldr	r3, .LL4+4		@ screensize = vinfo.xres * vinfo.yres *
	ldr	r3, [r3]		@		(vinfo.bits_per_pixel/8)
	str	r3, [r8, #scrwidth-dseg]
	str	r3, [r8, #scrwidth0-dseg]
	ldr	r2, .LL4+4
	ldr	r2, [r2, #4]
	str	r2, [r8, #scrheight-dseg]
	str	r2, [r8, #scrheight0-dseg]
	str	r2, [r8, #scrheight2-dseg]
	mul	r3, r2, r3
	ldr	r2, .LL4+4
	ldr	r2, [r2, #24]
	str	r2, [r8, #scrdepth-dseg]
	str	r2, [r8, #scrdepth0-dseg]
	mov	r2, r2, lsr #3
	mul	r3, r2, r3
	str	r3, [fp, #-16]
	ldr	r2, [fp, #-12]
	str	r2, [sp]
	mov	r2, #0
	str	r2, [sp, #4]
	mov	r0, #0			@ fbp = (char*)mmap(0,
	mov	r1, r3			@	screensize,
	mov	r2, #3			@	PROT_READ | PROT_WRITE,
	mov	r3, #1			@	MAP_SHARED,
	bl	mmap			@	fbfd, 0);
	str	r0, [r8, #framebuf-dseg]

@ if bytesPerLine is not set then set it to width * 2, which assumes depth = 16
	ldr	r0, [r8, #scrwidth-dseg]
	lsl	r0, #1
	ldr	r1, [r8, #bytesPerLine-dseg]
	cmp	r1, #0
	streq	r0, [r8, #bytesPerLine-dseg]

	sub	sp, fp, #4
	ldmfd	sp!, {fp, pc}

.LL4:	.word	.LLC0			@ 0
	.word	vinfo			@ 4
	.word	0x4602			@ 8
	.word	finfo			@12
.LLC0:	.asciz	"/dev/fb0"
	.align	2

@-------------------------------------------------------------------------------
@ Open mailbox file descriptor. Used to communicate with the GPU.
@
openmailbox:
	push	{lr}
	adr	r0, omb00		@ mbfd = open("/bin/mailbox", 0);
	mov	r1, #0
	bl	open
	str	r0, [r8, #mbfd-dseg]
	cmp	r0, #0			@ if (mbfd < 0)
	bge	omb90
	 adr	r0, omb10		@   printf("Cannot open /bin/mailbox");
	 bl	text
omb90:	pop	{pc}

@ The following two strings must remain in this order:
omb10:	.ascii	"Cannot open "
omb00:	.asciz	"/bin/mailbox"
	.align	2

	.data
	.align	2
mbfd:	.word	0			@ mailbox file descriptor
	.text

@===============================================================================
@ Routine to reserve an array with any number of dimensions. Returns base addr.
@ For example: mkarray( (3, 5, 7, 11), 4, 8, addr);
@ The stack contains (from bottom to top):
@  The number of elements in each dimension (3, 5, 7, 11)
@  The number of dimensions (4)
@  The number of bytes in each element of the array (1=char, 4=int, 8=real) = r0
@  (Only the last dimension of a char array has a single byte in it; the other
@  dimensions in a multidimensional char array all have 4-byte pointers.)
@  The address of where the pointer to the array and the array are to be placed.
@ WARNING: This increases r9.
@
	.data
	.align	2
r4save:	.word	0		@ register saves
r5save:	.word	0
r6save:	.word	0
r7save:	.word	0
retadr:	.word	0		@ return address
baseadr:.word	0		@ address of (first dimension of) heap space
levmax:	.word	0		@ maximum level of recursion (last dimension)
	.text

mkarray:str	r4, [r8, #r4save-dseg]
	str	r5, [r8, #r5save-dseg]
	str	r6, [r8, #r6save-dseg]
	str	r7, [r8, #r7save-dseg]
	str	lr, [r8, #retadr-dseg]	@ save return address (stack not avail.)
	pop	{r1}			@ get offset to pointer to array
	ldr	r0, [r8, #baseary-dseg]	@ add base address to form a pointer
	add	r1, r0			@  to the pointer to the array
	str	r9, [r8, #baseadr-dseg]

	pop	{r4}		@ get number of bytes in each element
	pop	{r0}		@ get number of dimensions
	sub	r0, #1		@ convert to maximum index off of r6 (levmax)
	lsl	r0, #2		@ qudruple for word entries on the stack
	mov	r6, r0
	add	r6, sp		@ point to number of elements in first dimension
	rsb	r0, #0
	str	r0, [r8, #levmax-dseg]

	mov	r5, #0		@ set index to point to size of first dimension
	bl	genary		@ reserve first dimension of array and store its
				@  base address into location pointed to by r1
	add	r6, #4		@ clean up stack--discard arguments
	mov	sp, r6

	add	r9, #3		@ round heap pointer up to next word boundary
	bic	r9, #3		@  (char arrays can misalign it)
	ldr	r0, [r8, #baseadr-dseg]	@ return base address of array space
	ldr	r4, [r8, #r4save-dseg]
	ldr	r5, [r8, #r5save-dseg]
	ldr	r6, [r8, #r6save-dseg]
	ldr	r7, [r8, #r7save-dseg]
	ldr	pc, [r8, #retadr-dseg]

@ Recursively called subroutine for setting up a multi-dimensional array.
@ Reserve an array and store its base address into the location pointed to
@  by the r1 register.
@ Register usage: (* indicates input values):
@	r0 - number of elements for current dimension
@   *	r1 - address of pointer to (next dimension of) array
@	r2 - loop counter and index
@	r3 - base address of current dimension of array
@   *	r4 - number of bytes in an element (1, 4 or 8)
@   *	r5 - index to size of each dimension; also = level of recursion (* -4)
@   *	r6 - points to first argument on the stack (size of first dimension)
@	r7 - scratch

genary:	push	{lr}
	ldr	r0, [r6, r5]	@ get number of elements for current dimension
	mov	r2, r0		@ save for possible use by loop counter below

@ Reserve space on heap
	ldr	r7, [r8, #levmax-dseg]	@ is this the last dimension?
	cmp	r5, r7
	bgt	ga02		@ skip if not
	 cmp	r4, #1		@ is this a char array?
	 beq	ga07		@ jump if so (last dimension is a single byte)
ga02:	cmp	r4, #8		@ is this a real array?
	bne	ga05		@ skip if not
	 mov	r7, #0x3F000000	@ address goes into low end of 1.0
	 orr	r7, #0x00F00000
	 str	r7, [r1, #4]
	 lsls	r0, #1		@ multiply elements by 8 bytes per real number
	 bcs	ga10		@ check for memory overflow: error if > FFFFFFFF
ga05:	lsls	r0, #1		@ multiply elements by 4 bytes per integer
	bcs	ga10		@ check for memory overflow: error if > FFFFFFFF
	lsls	r0, #1		@ (this was more reasonable when 64K was limit)
	bcs	ga10
ga07:	mov	r3, r9		@ r3 = base of reserved heap space
	adds	r9, r0		@ add number of bytes to reserve to heap pointer
	bcs	ga10		@ check for memory overflow: error if > FFFFFFFF
	 ldr	r0, [r8, #heaphi-dseg]	@ check for memory overflow if >= heaphi
	 cmp	r9, r0
	 blo	ga20		@ skip if ok
ga10:	mov	r0, #2		@ error 2: memory overflow
	bl	error
ga20:	str	r3, [r1]	@ store base address of reserved space into ptr

	ldr	r7, [r8, #levmax-dseg]	@ is this the last dimension?
	cmp	r5, r7
	ble	ga90		@ jump if so--don't recurse any lower

	mov	r0, #0		@ initialize index to first element
ga30:	cmp	r0, r2		@ is index beyond limit?
	bhs	ga90		@ jump if so--exit loop

	sub	r5, #4		@ advance index to next dimension size on stack
	mov	r1, r0		@ get index for element in current dimension
	cmp	r4, #8		@ is this a real array?
	bne	ga40		@ skip if not
	 lsl	r1, #1		@ reserve space for 8 bytes
ga40:	lsl	r1, #2		@ multiply by 4 to index by words
	add	r1, r3		@ add base address of reserved space

	push	{r0, r2, r3}	@ save some things (index, limit, base addr)
	bl	genary		@ reserve next dimension of array and store its
	pop	{r0, r2, r3}	@  base address into location pointed to by r1
	add	r5, #4		@ restore index to previous dimen. size on stack

	add	r0, #1		@ next element
	b	ga30		@ loop for each element in this dimension
ga90:	pop	{pc}

@===============================================================================
@				INTRINSICS
@===============================================================================
@ Start-up initialization for intrinsics.
@
setint:	push	{lr}
	mov	r0, #8			@ default number of digits for HexOut
	str	r0, [r8, #hexdigits-dseg] @ number of digits to output
	bl	ranint			@ set random number generator seed
	bl	fullwind		@ set device 6 window to full screen
	bl	openi3			@ reset disk buffer pointers: input,
	bl	openo3			@  and output
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to return the absolute value of top of stack.
@  VARIABLE:= ABS(VALUE)
@
intr0:	tst	r0, r0
	movpl	pc, lr
	rsb	r0, #0
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to generate a random number between 0 and (top of stack)-1.
@  VARIABLE:= RAN(VALUE)
@  If tos = 0 then this initializes a seed for a repeatable sequence.
@  If tos < 0 then this randomizes the seed and returns RAN(-tos).
@   VAL:= RAN(10)	\e.g: random number from 0 through 9
@  Algorithm is from Knuth's "Art of Computer Programming" Vol 2. 3rd ed. p.185.
@  This passes all tests in the Diehard suite (i.e. it's good and random).
@
	.data
	.align	2
seed:	.word	33049			@ standard seed (prime number)
	.text

intr1:	tst	r0, r0			@ range is in r0
	bne	ran01			@ skip if not zero
	 ldr	r1, sd33049		@ else reinitialize with standard seed
	 str	r1, [r8, #seed-dseg]
	 mov	pc, lr			@ return 0
ran01:
	push	{lr}
	bpl	ran05
	 rsb	r0, #0			@ make range positive
	 push	{r0}			@ save range
	 bl	ranint			@ randomize seed
	 pop	{r0}
ran05:
	push	{r0}			@ save range
@ Seed:= A*rem(Seed/Q) - R*(Seed/Q)
@  M = 7FFFFFFFh			@ Mersenne prime
@  A = 48271				@ "does well in spectral test"
@  Q = 44488 = M/A
@  R =	3399 = rem(M/A)
	ldr	r0, [r8, #seed-dseg]	@ get Seed
	ldr	r1, sd44488		@ /Q
	bl	sdiv			@ r0:= r0/r1; r1:= remainder
	ldr	r2, sd3399
	mul	r0, r2
	ldr	r2, sd48271
	mul	r1, r2
	rsbs	r0, r1			@ r0:= r1-r0
	addmi	r0, #0x7FFFFFFF		@ if Seed < 0 then Seed:= Seed + M
	str	r0, [r8, #seed-dseg]

	pop	{r1}			@ get range
	bl	sdiv			@ r0:= r0/r1; r1 = remainder
	mov	r0, r1			@ return rem(Seed/range)
	pop	{pc}

sd33049:.word	33049
sd44488:.word	44488
sd3399:	.word	3399
sd48271:.word	48271

@-------------------------------------------------------------------------------
@ Signed divide for compiled code: r0:= tos/r0; remain:= remainder
@
sdivx:	mov	r1, r0
	pop	{r0}

@ Signed divide: r0:= r0/r1; remain:= remainder
sdiv:	push	{r4, r5, lr}
	eor	r4, r0, r1		@ set msb if quotient will be negative
	movs	r5, r0			@ set msb if remainder will be negative
	rsbmi	r0, #0			@ abs(numerator)
	movs	r1, r1
	rsbmi	r1, #0			@ abs(denominator)
	bl	udiv
	movs	r4, r4			@ negate quotient if msb was set
	rsbmi	r0, #0
	movs	r5, r5			@ negate remainder if msb was set
	rsbmi	r1, #0
	str	r1, [r8, #remain-dseg]
	pop	{r4, r5, pc}

@-------------------------------------------------------------------------------
@ Unsigned 32-bit divide: r0:= r0/r1; r1:= remainder 
@  Based on C's helper function __aeabi_uidiv.
@
udiv:	subs	r2, r1, #1		@ branch if dividing by 0 or 1
	bls	udiv80
	cmp	r0, r1			@ branch if numerator <= denominator
	bls	udiv60
	tst	r1, r2			@ branch if dividing by a power of 2
	beq	udiv70

	clz	r3, r0			@ count leading zeros in numerator
	clz	r2, r1			@  and denominator
	sub	r3, r2, r3		@ D - N
	rsbs	r3, r3, #31		@ 31 - (D-N)
	addne	r3, r3, r3, lsl #1	@ * 3
	mov	r2, #0			@ initialize quotient
	addne	pc, pc, r3, lsl #2	@ * 4 => 12 bytes per jump table entry
	nop
	cmp	r0, r1, lsl #31		@ compare N to D<<31
	adc	r2, r2, r2		@ if D<<31 <= N then shift carry into Q
	subcs	r0, r0, r1, lsl #31	@                    N:= N - D<<31
	cmp	r0, r1, lsl #30		@ etc.
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #30
	cmp	r0, r1, lsl #29
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #29
	cmp	r0, r1, lsl #28
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #28
	cmp	r0, r1, lsl #27
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #27
	cmp	r0, r1, lsl #26
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #26
	cmp	r0, r1, lsl #25
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #25
	cmp	r0, r1, lsl #24
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #24
	cmp	r0, r1, lsl #23
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #23
	cmp	r0, r1, lsl #22
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #22
	cmp	r0, r1, lsl #21
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #21
	cmp	r0, r1, lsl #20
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #20
	cmp	r0, r1, lsl #19
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #19
	cmp	r0, r1, lsl #18
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #18
	cmp	r0, r1, lsl #17
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #17
	cmp	r0, r1, lsl #16
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #16
	cmp	r0, r1, lsl #15
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #15
	cmp	r0, r1, lsl #14
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #14
	cmp	r0, r1, lsl #13
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #13
	cmp	r0, r1, lsl #12
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #12
	cmp	r0, r1, lsl #11
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #11
	cmp	r0, r1, lsl #10
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #10
	cmp	r0, r1, lsl #9
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #9
	cmp	r0, r1, lsl #8
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #8
	cmp	r0, r1, lsl #7
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #7
	cmp	r0, r1, lsl #6
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #6
	cmp	r0, r1, lsl #5
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #5
	cmp	r0, r1, lsl #4
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #4
	cmp	r0, r1, lsl #3
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #3
	cmp	r0, r1, lsl #2
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #2
	cmp	r0, r1, lsl #1
	adc	r2, r2, r2
	subcs	r0, r0, r1, lsl #1
	cmp	r0, r1			@ lsl #0
	adc	r2, r2, r2
	subcs	r0, r0, r1		@ lsl #0

	mov	r1, r0			@ r1:= remainder
	mov	r0, r2			@ r0:= quotient
	mov	pc, lr			@ return

udiv60:	moveq	r1, #0			@ if N = D then rem:= 0; Q:= 1
	moveq	r0, #1
	movne	r1, r0			@ if N < D then rem:= N; Q:= 0
	movne	r0, #0
	mov	pc, lr			@ return

udiv70:	clz	r3, r1			@ divide (positive N) by power of 2
	rsb	r3, r3, #31		@ 31 - number of leading zeros in D
	and	r1, r0, r2		@ mask to get remainder bits
	lsr	r0, r0, r3		@ shift to get Q
	mov	pc, lr			@ return

udiv80:	movhs	r1, #0			@ remainder gets 0 if dividing by 1
	movhs	pc, lr			@  and r0 = numerator (=r0)
	mov	r0, #1			@ else flag divide-by-zero error
	push	{lr}
	bl	error
	mov	r0, #0x7fffffff		@ return best answer in case error
	mov	r1, #0			@  routine doesn't trap
	pop	{pc}

@-------------------------------------------------------------------------------
@ Initialize random number generator.
@
ranint:	push	{lr}
	bl	intr82			@ get random seed from GETTIME intrinsic
	orr	r0, #4			@ make sure it's not zero
	and	r0, #0x7FFFFFFF		@  and not negative
	str	r0, [r8, #seed-dseg]
	pop	{pc}			@ return

@-------------------------------------------------------------------------------
@ Intrinsic to set the seed for the random number generator, for a repeatable
@  sequence.
@  RANSEED(INTEGER)
@
intr79:	cmp	r0, #0			@ don't set seed to 0 because the RNG
	moveq	r0, #0x40000000		@  will then return 0's forever
	str	r0, [r8, #seed-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the remainder of the most recent division. The argument
@  is an expression whose result is thrown away. This expression can contain
@  a division, or simply be a zero to get the result of the last division.
@  VARIABLE:= REM(EXPRESSION)
@
intr2:	ldr	r0, [r8, #remain-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to reserve bytes on the heap and return the address of the
@  reserved space. This is the method by which XPL handles dynamic storage
@  for arrays. Since the space is reserved in the heap allocation of a
@  procedure, the array will disappear when the procedure is exited.
@  ADDRESS:= RESERVE(VALUE)
@
intr3:	mov	r2, r9			@ save heap pointer
	add	r9, r0			@ add in number of bytes to reserve
	add	r9, #3			@ round up to next word boundary
	and	r9, #0xFFFFFFFC
	ldr	r1, [r8, #heaphi-dseg]	@ test for overlfow
	cmp	r9, r1
	blo	res90			@ skip if ok
	 mov	r0, #2			@ runtime error # 2: out of memory
	 push	{lr}
	 bl	error
	 pop	{lr}
res90:	mov	r0, r2			@ return address of reserved space
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to swap the two low bytes (and two high bytes) of top of stack.
@  VARIABLE:= SWAP(VALUE)
@
intr4:	rev16	r0, r0
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to extend the sign of the low byte into the high bytes.
@  VARIABLE:= EXTEND(VALUE)
@
intr5:	sxtb	r0, r0
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to restart the current program.
@  RESTART
@
intr6:	mov	r0, #-1			@ set rerun flag
	str	r0, [r8, #rerunf-dseg]
	ldr	sp, [r8, #stkptr-dseg]
	pop	{r0}			@ discard address pushed by startxpl
	b	restart

@-------------------------------------------------------------------------------
@ Intrinsic to input a byte from the specified device and return it in r0.
@  VARIABLE:= CHIN(DEVICE)
@
intr7:	str	r0, [r8, #nowdev-dseg]
	b	chin			@ branch and return

@-------------------------------------------------------------------------------
@ Intrinsic to output the byte in r0 to the specified device.
@  CHOUT(DEVICE, BYTE)
@
intr8:	pop	{r1}
	str	r1, [r8, #nowdev-dseg]
	b	chout			@ branch and return

@-------------------------------------------------------------------------------
@ Intrinsic to output a carriage return and a line feed to the specified device.
@  CRLF(DEVICE)
@
intr9:	str	r0, [r8, #nowdev-dseg]	@ set the device number
	push	{lr}
	mov	r0, #CR
	bl	chout
	mov	r0, #LF
	bl	chout
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to input ASCII digits from the specified device and convert them to
@  a signed, decimal, 32-bit value that is returned in r0.
@  VAL:= INTIN(DEV)
@	r0 = character input
@	r4 = working register (contains number to be converted)
@	r5 = flag: a numeric character has been entered
@	r6 = flag: a minus sign was entered, i.e. the number is negative
@
intr10:	push	{r4, r5, r6, lr}	@ save registers
	str	r0, [r8, #nowdev-dseg]	@ set the device number
ii00:	mov	r4, #0			@ NUM:=0
	mov	r5, #0			@ NUMFLG:=false
	mov	r6, #0			@ SIGN:=false

	bl	chin			@ get character
	cmp	r0, #'-'		@ if r0 = ^- then SIGN:=true
	bne	ii30
	mov	r6, #-1
					@ loop begin
ii20:	bl	chin			@   get digit in r0
	cmp	r0, #'_'		@   ignore underlines
	beq	ii20
ii30:	cmp	r0, #EOF		@   if EOF then exit
	beq	ii55
	subs	r0, #'0'		@   if al<'0' ! al>'9' then quit
	blo	ii50
	 cmp	r0, #9
	 bhi	ii50
	 mov	r5, #-1			@   NUMFLG:=true
	 mov	r1, #10
	 mla	r4, r4, r1, r0		@   NUM:= NUM*10 + (digit-^0)
	 b	ii20			@   end
ii50:
	cmp	r5, #0			@ if NUMFLG then quit
	beq	ii00			@ else loop
ii55:	cmp	r6, #0			@ if SIGN then
	rsbne	r4, #0			@   NUM:= -NUM
	mov	r0, r4			@ return NUM
	pop	{r4, r5, r6, pc}	@ restore registers and return

@-------------------------------------------------------------------------------
@ Intrinsic to output the signed integer in r0 to the specified device. The
@  integer is converted to an ASCII string.
@  INTOUT(DEVICE, VALUE)
@
intr11:	pop	{r1}
	str	r1, [r8, #nowdev-dseg]	@ set the device number
					@ fall into intout routine
@-------------------------------------------------------------------------------
@ Convert signed integer in r0 to decimal ASCII and output the characters to
@  "nowdev"
@
intout:	cmp	r0, #0
	bge	io90
	 rsb	r0, #0			@ make integer positive
	 push	{r0,lr}
	 mov	r0, #'-'
	 bl	chout
	 pop	{r0,lr}
io90:
@ Convert unsigned integer in r0 to ASCII and display it
numout:	push	{lr}
	mov	r1, #10
	bl	udiv
	push	{r1}
	cmp	r0, #0
	blne	numout
	pop	{r0}
	add	r0, #'0'
	bl	chout
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to output a text string to the specified device. The starting
@  address is in r0, and is terminated by a character with its bit 7 set.
@  TEXT(DEVICE, "HELLO WORLD!")
@
intr12:	pop	{r1}			@ get device number
	str	r1, [r8, #nowdev-dseg]	@ set the device number
	push	{r4, lr}		@ save registers
	mov	r4, r0			@ put pointer into preserved register
	b	txtl20			@ enter loop
txtl10:	 bl	chout			@ output character
txtl20:	ldrb	r0, [r4]		@ get char
	add	r4, #1			@ increment pointer
	cmp	r0, #0x80		@ last char?
	blo	txtl10
	and	r0, #0x7F		@ clear MSB of char
	bl	chout			@ output last char
	pop	{r4, pc}		@ restore registers and return

@-------------------------------------------------------------------------------
@ Intrinsic to output a text string to the specified device. The starting
@  address is in r0, and is terminated by a binary 0.
@  TEXT(DEV, "HELLO WORLD")
@
intr255:				@ (nat.s and xpl0.xpl use this label)
intr12a:pop	{r1}			@ get device number (natrpi.s uses this)
	str	r1, [r8, #nowdev-dseg]	@ set the device number
intr12b:push	{r4, lr}		@ entry point from flout routine
	mov	r4, r0			@ put pointer into preserved register
	b	txta20			@ enter loop
txta10:	 bl	chout			@ output character
txta20:	ldrb	r0, [r4]		@ get char
	add	r4, #1			@ increment pointer
	cmp	r0, #0			@ last char?
	bne	txta10			@ loop back if not
	pop	{r4, pc}		@ restore registers and return

@-------------------------------------------------------------------------------
@ Intrinsic to open an input device.
@  OPENI(DEVICE)
@
intr13:	str	r0, [r8, #nowdev-dseg]	@ set device number
	b	openi			@ (pjmp) go do it

@-------------------------------------------------------------------------------
@ Intrinsic to open an output device.
@  OPENO(DEVICE)
@
intr14:	str	r0, [r8, #nowdev-dseg]	@ set device number
	b	openo			@ (pjmp) go do it

@-------------------------------------------------------------------------------
@ Intrinsic to close an output device.
@  CLOSE(DEVICE)
@
intr15:	str	r0, [r8, #nowdev-dseg]	@ set device number
	b	close			@ (pjmp) go do it

@-------------------------------------------------------------------------------
@ Intrinsic to abort the program.
@  ABORT
@
intr16:	ldr	sp, [r8, #stkptr-dseg]
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to set the error trap flags.
@  TRAP(BITS)
@
intr17:	str	r0, [r8, #traps-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to tell a user how many bytes of heap space he has left. Of
@  course he may not reserve all of it, since he must leave a working
@  heap for subsequent procedure calls etc. Since only the user knows
@  how much this might be, it is left to him to decide how much he
@  actually has free to play with.
@  VARIABLE:= FREE
@
intr18:	ldr	r0, [r8, #heaphi-dseg]	@ r0 = heaphi - hp
	sub	r0, r9
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the rerun flag.
@  BOOLEAN:= RERUN
@
intr19:	ldr	r0, [r8, #rerunf-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the current value of the heap pointer. This intrinsic
@  is used for saving the heap pointer for later restoration with SETHP.
@  The user had better have a good idea of the funtioning of XPL before
@  dinging with the heap pointer or he will surely bomb himself!
@  ADDRESS:= GETHP
@
intr20:	mov	r0, r9
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to set the heap pointer--a very dangerous thing to do! See
@  GETHP's comments.
@  SETHP(ADDRESS)
@
intr21:	mov	r9, r0
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the runtime error number and then clear it.
@  ERROR:= GETERR
@
intr22:	ldr	r0, [r8, #errnum-dseg]
	mov	r1, #0
	str	r1, [r8, #errnum-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to set the text cursor to a specified position (for devices 6 and 0)
@  CURSOR(X, Y)
@
intr23:	pop	{r1}			@ X
	push	{lr}
	push	{r0, r1}

@ ANSI command to position cursor on devices 0 and 1; outputs: Esc[<Y+1>;<X>H
	mov	r0, #ESC		@ ESC [
	bl	putcharnow
	mov	r0, #'['
	bl	putcharnow
	mov	r0, #0			@ nowdev = 0
	str	r0, [r8, #nowdev-dseg]
	pop	{r0, r1}		@ get coordinates
	push	{r0, r1}
	push	{r1}			@ save X
	add	r0, #1			@ topmost line is 1
	bl	intout
	mov	r0, #';'
	bl	putcharnow
	pop	{r0}			@ get X
	add	r0, #1			@ leftmost column is 1
	bl	intout
	mov	r0, #'H'
	bl	putcharnow

@ Position cursor on device 6 display:
	pop	{r0, r1}		@ get coordinates
	lsl	r1, #3			@ penx:= X*8
	str	r1, [r8, #penx-dseg]
	ldr	r1, [r8, #fontheight-dseg] @ peny:= Y*fontheight
	mul	r0, r1
	str	r0, [r8, #peny-dseg]
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to associate the file descriptor with device 3. This also determines
@  input and output, and the size of the buffer.
@  FSET(DESCRIPTOR, MODE)
@		Input	Output
@	Large:	  I	  O
@	Small:	  i	  o
@
intr24:	pop	{r1}			@ file descriptor
	push	{lr}

	mov	r2, r0			@ save copy of buffer size flag
	orr	r0, #0x20		@ convert flag to lowercase
	cmp	r0, #'i'		@ if IO = 'i' then
	bne	fs20
	 str	r1, [r8, #infd-dseg]	@   infd = fd
	 str	r2, [r8, #inbufflg-dseg]@   inbufflg = 'I' (or 'i')
	 b	fs90
fs20:
	cmp	r0, #'o'		@ else if IO = 'o' then
	bne	fs40
	 str	r1, [r8, #outfd-dseg]	@   outfd = fd
	 str	r2, [r8, #outbufflg-dseg]@  outbufflg = 'O' (or 'o')
	 b	fs90
fs40:
	mov	r0, #3			@ else error 3
	bl	error
fs90:	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to set the rerun flag directly.
@  SETRUN(BOOLEAN)
@
intr25:	str	r0, [r8, #rerunf-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to input ASCII digits and convert them to a 32-bit hex value that
@  is returned in r0.
@  VARIABLE:= HEXIN(DEVICE)
@	r0 = character input
@	r4 = accumulated value
@	r5 = digit counter
@
intr26:	push	{r4, r5, lr}		@ preserve critical registers
	str	r0, [r8, #nowdev-dseg]	@ set the device number
	mov	r4, #0			@ clear result register
	mov	r5, #8			@ initialize digit counter
hexi00:	bl	chin			@ get character
	cmp	r0, #'_'		@ ignore underlines
	beq	hexi00
	cmp	r0, #EOF		@ if it's an EOF then exit
	beq	hexi90
	cmp	r0, #'0'		@ is character in range 0 thru 9?
	blo	hexi40			@ branch if not
	cmp	r0, #'9'
	bhi	hexi20			@ (might be A-F)
	 sub	r0, #'0'		@ convert ASCII to binary value
	 b	hexi30			@ go combine with other digits
hexi20:
	and	r0, #0xDF		@ force to uppercase
	cmp	r0, #'A'		@ is character in range A thru F?
	blo	hexi40			@ branch if not
	cmp	r0, #'F'
	bhi	hexi40
	sub	r0, #'A'-10		@ convert ASCII to binary value
hexi30:
	lsl	r4, #4			@ multiply current value by 16
	add	r4, r0			@ add new digit
	subs	r5, #1
	bne	hexi00			@ exit if there are 8 digits
hexi40:
	cmp	r5, #8			@ was there a valid hex digit?
	beq	hexi00			@ branch if not -- keep trying
hexi90:	mov	r0, r4			@ return the hex value in r0
	pop	{r4, r5, pc}		@ restore critical registers and return

@-------------------------------------------------------------------------------
@ Intrinsic to output the top of stack (r0) in hex ASCII.
@  HEXOUT(DEVICE, VALUE)
@
intr27:	pop	{r1}			@ get device number
	str	r1, [r8, #nowdev-dseg]	@ set device number

	push	{r4, lr}
	ldr	r2, [r8, #hexdigits-dseg] @ number of digits to output
ho10:	mov	r1, r0			@ push least significant hex digit
	and	r1, #0x0F
	push	{r1}
	lsr	r0, #4
	subs	r2, #1			@ loop for specified number of digits
	bgt	ho10
	cmp	r0, #0			@ also keep looping if more digits
	bne	ho10

	ldr	r4, [r8, #hexdigits-dseg]
	sub	r4, r2			@ number pushed = hexdigits - r2
ho20:	pop	{r0}			@ pop most significant digit first
	cmp	r0, #10			@ output digit in r0 as ASCII hex
	addge	r0, #7
	add	r0, #'0'		@ convert to ASCII
	bl	chout			@ output it
	subs	r4, #1			@ loop for number pushed
	bgt	ho20
	pop	{r4, pc}		@ return

	.data
	.align	2
hexdigits: .word	8		@ number of hex digits to output
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to open a file by name and return its file descriptor.
@  DESCRIPTOR:= FOPEN("/path/filename.ext", 0=r/1=w)
@
fnsize	=	256			@ maximum length of filename and path+1

intr29:	pop	{r2}			@ get pointer to filename
	push	{r4, lr}
	mov	r4, r0			@ save I/O mode (0 or 1)

fo02:	ldrb	r0, [r2]		@ skip any leading space characters
	cmp	r0, #0x20		@ while P(0) = Space do P++
	addeq	r2, #1
	beq	fo02

	sub	sp, #fnsize		@ make temporary space on stack for name
	mov	r3, sp			@ point to base of temporary stack space
fo05:	ldrb	r0, [r2]		@ loop	C:= P(0) & $7F
	and	r0, #0x7F
	cmp	r0, #0x21		@ 	if C<! or C>~ or C=, then quit
	blt	fo10
	cmp	r0, #0x7E
	bgt	fo10
	cmp	r0, #0x2C		@ 	comma
	beq	fo10

	sub	r1, r3, sp		@       if S >= fnsize then too long
	cmp	r1, #fnsize
	bge	fo40
	strb	r0, [r3]		@       S(0):= C
	add	r3, #1			@	S++

	ldrb	r0, [r2]		@       if MSB terminated name then quit
	cmp	r0, #0x80
	bge	fo10
	add	r2, #1			@	P++
	b	fo05
fo10:
	mov	r0, #0			@ S(0):= 0; terminate string
	strb	r0, [r3]

	mov	r0, sp			@ point to filename
	cmp	r4, #0			@ if I/O = 0 then read mode
	bne	fo20
	 mov	r1, #0			@ fd = open(filename, O_RDONLY, 0)
	 mov	r2, #0
	 bl	open
	 b	fo30
fo20:
	cmp	r4, #1			@ else if I/O = 1 then write mode
	bne	fo40			@ else error(3)
	mov	r1, #0x0B6		@ fd = creat(filename, 0666)    (octal)
	orr	r1, #0x100
	bl	creat			@ read & write for owner, group, others
fo30:	cmp	r0, #-1			@ if fd = -1 then error(3)
	bne	fo90
fo40:	 mov	r0, #3
	 bl	error
	 mov	r0, #-1			@ return an illegal file descriptor
fo90:
	add	sp, #fnsize		@ discard temporary stack space
	pop	{r4, pc}		@ return with file descriptor in r0

@-------------------------------------------------------------------------------
@ Intrinsic to close a disk file by its file descriptor.
@  FCLOSE(DESCRIPTOR)
@
intr32:	push	{lr}
	bl	close			@ r0 = fd
	cmp	r0, #-1			@ if = -1 then error(3)
	bne	fc90
	 mov	r0, #3
	 bl	error
fc90:	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to return 'true' if a key has been struck.
@  BOOLEAN:= CHKKEY
@
intr33:	push	{lr}
	ldr	r0, [r8, #scancode-dseg]@ if a non-ASCII key was just hit then
	cmp	r0, #0			@  indicate that the 2nd byte is pending
	bne	kh30

	mov	r0, #0			@ ioctl(STDIN, FIONREAD, &bytesWaiting)
	mov	r1, #0x541B&0x00FF
	orr	r1, #0x541B&0xFF00
	push	{r0}			@ make temporary location on stack
	mov	r2, sp			@ point to it
	bl	ioctl
	pop	{r0}			@ get number of bytes waiting (0 or 1)
	cmp	r0, #0			@ return 'false' if no bytes waiting
	popeq	{pc}
kh30:	mov	r0, #-1			@ else return 'true'
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to emit a sound for a specified duration and wavelength period.
@  SOUND(VOLUME, DURATION, PERIOD)
@ This works approximately the same as the Sound intrinsic in the PC version.
@ What's trivial on other computers is unbelieveably complicated on the RPi!
@ Volume: 0 = off, 1 = full, other = intensity 2..255 (unimplemented).
@ Duration is in 1/20 seconds (but varies somewhat depending on the Period).
@ Period of the output square wave is in microseconds (very approximately).
@
intr39:	mov	r2, r0			@ period
	pop	{r1}			@ duration
	pop	{r0}			@ volume
	.if	!natrpi
	cmp	r0, #0			@ branch if volume # 0
	bne	snd10
	.endif

	mov	r0, #54925&0x00FF	@ DelayUS(54925)
	orr	r0, #54925&0xFF00
	mul	r0, r1
	b	intr94

	.if	!natrpi
snd10:
	stmfd	sp!, {r4, r5, r6, r7, r8, lr}
	sub	sp, sp, #24

	add	r1, #1			@ kludge fix

@ Shift = -7;
@ i = Period;
@ while (i) {i = i>>1;  Shift++;}
@ Toggle = 1 << Shift;
	mov	r6, r1			@ save duration
	cmp	r2, #0
	mvneq	r1, #6
	beq	snd2
	mov	r3, r2			@ period
	mvn	r1, #6			@ -7
snd3:
	add	r1, r1, #1		@ <<
	movs	r3, r3, asr #1
	bne	snd3
snd2:
@ Shift = Shift+6;
@ SampFreq = 37237 - (((Period - (1<<Shift))*18619) >> Shift);
	mov	r7, #1
	mov	r5, r7, asl r1
	add	r1, r1, #6

	sub	r2, r2, r7, asl r1
	ldr	r3, snd16
	mul	r3, r3, r2
	mov	r1, r3, asr r1
	rsb	r8, r1, #37120
	add	r8, r8, #117

@ BufSize = 56*(SampFreq>>10);
@ BufSize = BufSize & ~((Toggle<<1) - 1);
	mov	r4, r8, asr #10
	rsb	r4, r4, r4, asl #3
	mov	r3, r5, asl r7
	rsb	r3, r3, #0
	and	r4, r3, r4, asl #3

@ snd_pcm_open(&handle, "default", SND_PCM_STREAM_PLAYBACK, 0);
	add	r0, sp, #20
	ldr	r1, snd16+4
	mov	r2, #0
	mov	r3, r2
	bl	snd_pcm_open

@ snd_pcm_set_params(
@ r0 =	handle,				// PCM handle
@ r1 =	SND_PCM_FORMAT_U8,		// unsigned, 8-bit data
@ r2 =	SND_PCM_ACCESS_RW_INTERLEAVED,	// access
@ r3 =	1,				// number of channels
@ sp(0)	SampFreq,			// samples per second
@ sp(4)	0,				// flag: allow alsa-lib resample stream
@ sp(8)	150000);			// latency in microseconds
	str	r8, [sp, #0]		@ SampFreq
	mov	r3, #0			@ disable resample stream
	str	r3, [sp, #4]
@	ldr	r3, snd16+8		@ latency

	mov	r3, #4992		@ latency = 4992*duration + 43008
	mul	r3, r3, r6		@ imperically determined values rounded
	add	r3, #43008		@  to single byte values

@	add	r3, #2496
@	mov	r3, #51968+2048

	str	r3, [sp, #8]
	ldr	r0, [sp, #20]
	mov	r1, r7			@ FORMAT (r7=1)
	mov	r2, #3			@ ACCESS
	mov	r3, r7			@ one channel
	bl	snd_pcm_set_params

	sub	sp, #2048		@ make buffer space on stack

@ for (i=0; i<BufSize; i++)		// fill buffer with square waves
@ 	{s = 0xFF;
@ 	if (i & Toggle) s = 0;
@ 	buffer[i] = s;}       
	cmp	r4, #0
	bgt	snd11
	b	snd5
snd7:
	and	r2, r3, r5
	cmp	r2, #0
	moveq	r2, #255
	movne	r2, #0
	b	snd15
snd11:
	mov	r3, #0
	mov	r2, #255
@	ldr	r1, snd16+12
	mov	r1, sp
snd15:
	strb	r2, [r3, r1]
	add	r3, r3, #1
	cmp	r3, r4
	bne	snd7
snd5:
@ for (i=0; i<Duration; i++)		// play buffer Duration times
@ 	snd_pcm_writei(handle, buffer, BufSize);
	cmp	r6, #0
	ble	snd8
	mov	r5, #0
@	ldr	r7, snd16+12
	mov	r7, sp
snd9:
@	ldr	r0, [sp, #20]
	ldr	r0, [sp, #20+2048]

	mov	r1, r7
	mov	r2, r4
	bl	snd_pcm_writei
	add	r5, r5, #1
	cmp	r5, r6
	bne	snd9
snd8:
	add	sp, #2048		@ discard buffer space

@ snd_pcm_close(handle);
	ldr	r0, [sp, #20]
	bl	snd_pcm_close
	add	sp, sp, #24
	ldmfd	sp!, {r4, r5, r6, r7, r8, pc}

snd16:	.word	18619
	.word	sndC0
	.word	150000
sndC0:	.ascii	"default\000"
	.align	2
	.endif				@ !natrpi

@-------------------------------------------------------------------------------
@ Intrinsic to quickly clear the displayed portion of the graphic screen, and
@  set the graphic pen position to the upper-left corner.
@  CLEAR
@
intr40:	push	{r4-r12, lr}
	ldr	r2, [r8, #framebuf-dseg]
	ldr	r3, [r8, #bytesPerLine-dseg]
	ldr	r4, [r8, #scrwidth-dseg]
	ldr	r5, [r8, #scrheight-dseg]
	ldr	r0, [r8, #scrdepth-dseg]
	cmp	r0, #16			@ if scrdepth=16 then use 16
	movhi	r0, #32			@ if scrdepth>16 then use 32
	movlo	r0, #8			@ if scrdepth<16 then use 8
	mul	r4, r4, r0		@ bytesPerLine - bytesPerWidth
	lsr	r4, #3			@ = bytesPerLine - X*Z/8
	sub	r3, r4			@ = amount to skip (modulo, normally 0)
	lsr	r4, #5			@ = bytesPerWidth/32
					@ bytesPerWidth = X*Z/8
	mov	r0, #0			@ set up black = 0 in 8 registers for a
	mov	r6, r0			@  total of 8*4 = 32 bytes
	mov	r7, r0
	mov	r8, r0
	mov	r9, r0
	mov	r10, r0
	mov	r11, r0
	mov	r12, r0

clr10:	mov	r1, r4			@ bytesPerWidth/32
clr20:	stmia	r2!, {r0, r6-r12}	@ store 32 zero (black) bytes
	subs	r1, #1			@ loop
	bne	clr20

	add	r2, r3			@ amount to skip (modulo)
	subs	r5, #1			@ loop for height scan lines
	bne	clr10

	pop	{r4-r12, lr}		@ restore registers
	str	r0, [r8, #penx-dseg]	@ set cursor position to upper-left
	str	r0, [r8, #peny-dseg]	@ corner for device 6
	mov	pc, lr

	.data
	.align	2
fbfd:	.word	0			@ frame buffer file descriptor: /dev/fb0
@ GETFB requires this order: scrwidth, scrheight, scrdepth, framebuf, penx, peny
scrwidth: .word	1280			@ graphic screen width & height (pixels)
scrheight:.word	720
scrdepth: .word	16			@ bits per pixel (0=>8, 16, 24=>32)
framebuf:.word	0			@ graphic memory frame buffer address
penx:	.word	0			@ pen coordinates (pixels) for drawing
peny:	.word	0

scrwidth0:.word	1280			@ initial screen dimensions that are
scrheight0:.word 720			@  restored by SetVid(3)
scrdepth0:.word	16
scrheight2:.word 720			@ screen height including second page
bytesPerLine: .word 1280*2		@ number of bytes per graphic scan line
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to plot a pixel on the graphic screen at X,Y in color C.
@  POINT(X, Y, COLOR)
@
intr41:	pop	{r1, r2}		@ r1=Y, r2=X
intr41a:ldr	r3, [r8, #scrheight2-dseg] @ clip to frame buffer dimensions
	str	r1, [r8, #peny-dseg]
	str	r2, [r8, #penx-dseg]	@ set pen position
	cmp	r1, r3
	ldr	r3, [r8, #scrwidth-dseg]
	movhs	pc, lr			@ return if Y >= scrheight2 or Y < 0
	cmp	r2, r3
	movhs	pc, lr			@ return if X >= scrwidth or X < 0

	mla	r1, r1, r3, r2		@ offset in pixels = Y*scrwidth + X
	ldr	r2, [r8, #scrdepth-dseg] @ bits per pixel (8, 16, 24, 32)
	ldr	r3, [r8, #framebuf-dseg]

	cmp	r2, #8
	beq	pnt8
	blt	pnt0
	cmp	r2, #24
	beq	pnt24
	bgt	pnt32

pnt16:	lsl	r1, #1			@ 16 bits per pixel
	strh	r0, [r3, r1]
	mov	pc, lr			@ return

pnt0:	tst	r0, #0x80		@ if MSB set then
	ldrneb	r2, [r3, r1]		@   get existing pixel's color
	eorne	r0, r2			@   exclusive-or it with new color
	andne	r0, #0x7F		@   make sure high bit is clear
pnt8:	strb	r0, [r3, r1]		@ 8 bits per pixel
	mov	pc, lr			@ return

pnt24:	orr	r0, #0xFF000000		@ full brightness
	str	r0, [r3, r1, lsl #2]	@ display pixel
	mov	pc, lr			@ return

@ Register usage:
@  r0 = foreground color ABGR (A = alpha)
@  r1 = background color
@  r2 = foreground product = F * A
@  r3 = background product = B * (1-A) (approximately)
@  r4 = pointer to pixel in frame buffer
@  r5 = 0 (used to do a lsr simultaneously on the 3 BGR bytes)
pnt32:	push	{r4, r5}
	add	r4, r3, r1, lsl #2
	ldr	r1, [r4]		@ get background pixel
	mov	r2, #0			@ initialize products
	mov	r3, #0
	mov	r5, #0

@ vector multiply:
@ r2 = F * A
@ r3 = B * (1-A)
	tst	 r0, #0x01000000	@ alpha's bit 0
	uhadd8ne r2, r2, r0		@ r2B=(r2B+r0B)/2, r2G=(r2G+r0G)/2,
	uhadd8eq r2, r2, r5		@ r2R=(r2R+r0R)/2
	uhadd8eq r3, r3, r1		@ does the 3 bytes of BGR all at once!
	uhadd8ne r3, r3, r5		@ r2B=r2B>>1, r2G=r2G>>1, r2R=r2R>>1

	tst	 r0, #0x02000000
	uhadd8ne r2, r2, r0
	uhadd8eq r2, r2, r5
	uhadd8eq r3, r3, r1
	uhadd8ne r3, r3, r5

	tst	 r0, #0x04000000
	uhadd8ne r2, r2, r0
	uhadd8eq r2, r2, r5
	uhadd8eq r3, r3, r1
	uhadd8ne r3, r3, r5

	tst	 r0, #0x08000000
	uhadd8ne r2, r2, r0
	uhadd8eq r2, r2, r5
	uhadd8eq r3, r3, r1
	uhadd8ne r3, r3, r5

	tst	 r0, #0x10000000
	uhadd8ne r2, r2, r0
	uhadd8eq r2, r2, r5
	uhadd8eq r3, r3, r1
	uhadd8ne r3, r3, r5

	tst	 r0, #0x20000000
	uhadd8ne r2, r2, r0
	uhadd8eq r2, r2, r5
	uhadd8eq r3, r3, r1
	uhadd8ne r3, r3, r5

	tst	 r0, #0x40000000
	uhadd8ne r2, r2, r0
	uhadd8eq r2, r2, r5
	uhadd8eq r3, r3, r1
	uhadd8ne r3, r3, r5

	tst	 r0, #0x80000000
	uhadd8ne r2, r2, r0
	uhadd8eq r2, r2, r5
	uhadd8eq r3, r3, r1
	uhadd8ne r3, r3, r5

	uqadd8	r0, r2, r3		@ F*A + B*(1-A)
	orr	r0, #0xFF000000		@ full (h/w) brightness
	str	r0, [r4]		@ display blended pixel
	pop	{r4, r5}
	mov	pc, lr			@ return

@-------------------------------------------------------------------------------
@ Draw line from PenX, PenY to X, Y in COLOR.
@  LINE(X, Y, COLOR)
@
intr42:	pop	{r2}			@ get Y
	pop	{r1}			@ get X

@ See if a fast horizontal line can be drawn
	ldr	r3, [r8, #peny-dseg]	@ if Y = PenY (i.e: horizontal line)
	cmp	r2, r3
	bne	lin40			@ skip if not

	ldr	r3, [r8, #scrheight-dseg]@ if X >= height*2 then
	cmp	r2, r3, lsl #1		@   (unsigned also tests X < 0)
	strhs	r1, [r8, #penx-dseg]	@   set PenX to new end of line
	movhs	pc, lr			@   and return

	ldr	r3, [r8, #scrdepth-dseg]@ if depth <= 8 then
	cmp	r3, #8
	bhi	lin10
	 tst	r0, #0x0000FF00		@   if (color & $FF00) # 0 then
	 bne	lin40			@     dot/dash pattern, skip
	 cmp	r3, #8			@   if depth < 8 then
	 tstne	r0, #0x80		@     if (color & $80) # 0 then
	 bne	lin40			@       xor feature, skip
	 b	lin25			@   else go draw fast line
lin10:
	cmp	r3, #16			@ else if depth=16 or depth=24 then
	beq	lin15
	 cmp	r3, #24
	 bne	lin20
lin15:	tst	r0, #0xFF000000		@   if (color & $FF00_0000) # 0 then
	bne	lin40			@     dot/dash pattern, skip
	b	lin25			@   else go draw fast line
lin20:
	cmp	r3, #32			@ else if depth = 32 then
	bne	lin40			@ skip if not (for safety)
	cmp	r0, #0xFF000000		@   if color < $FF00_0000 then
	blo	lin40			@   transparent line, skip
lin25:
@ draw a fast horizontal line from left to right
	push	{r4, lr}
	ldr	r3, [r8, #penx-dseg]	@ r1:= min(r1, penx)
	str	r1, [r8, #penx-dseg]	@ first set pen to end of line
	mov	r4, r3			@ assume penx is at right end of line
	cmp	r1, r3			@ if x > penx then
	movgt	r4, r1			@  set x to the right end of line and
	movgt	r1, r3			@  start with the original penx position
@ r0=color, r1=X (start/left), r2=Y, r3=scratch, r4=X (end/right)
	cmp	r1, #0			@ clip to screen dimensions
	movlt	r1, #0			@ if startX < 0 then startX:= 0
	cmp	r4, #0			@ if endX < 0 then exit
	blt	lin39
	ldr	r3, [r8, #scrwidth-dseg]
	cmp	r4, r3			@ if endX >= width then
	movge	r4, r3			@   endX:= width-1
	subge	r4, #1
	cmp	r1, r3			@ if startX >= width then exit
	bge	lin39

	mla	r3, r3, r2, r1		@ tentative Addr:= W*Y + X

	ldr	r2, [r8, #scrdepth-dseg]@ if depth <= 8 then
	cmp	r2, #8
	bgt	lin33
	 ldr	r2, [r8, #framebuf-dseg]
	 add	r3, r2			@   plus framebuf
	 push	{r3}
	 push	{r0}			@ Value = Color (only low byte is used)
	 sub	r0, r4, r1		@ Size = end - start + 1
	 add	r0, #1
	 bl	intr101			@ FillMem(Addr, Value, Size)
	 b	lin39
lin33:
	cmp	r2, #16			@ if depth <= 16
	bgt	lin35
	 ldr	r2, [r8, #framebuf-dseg]
	 lsl	r3, #1			@   double for half-words
	 add	r3, r2			@   plus framebuf
	 push	{r3}
	 push	{r0}			@ Value = Color (only low word is used)
	 sub	r0, r4, r1		@ Size = end - start + 1
	 add	r0, #1
	 bl	fillhalfwords		@ FillHalfWords(Addr, Value, Size)
	 b	lin39
lin35:					@ assume depth = 24 or 32
	orr	r0, #0xFF000000		@ set full brightness for depth = 24
					@ (doesn't change value for depth = 32)
	ldr	r2, [r8, #framebuf-dseg]
	lsl	r3, #2			@ times 4 for full-words
	add	r3, r2			@ plus framebuf
	push	{r3}
	push	{r0}			@ Value = Color (only low word is used)
	sub	r0, r4, r1		@ Size = end - start + 1
	add	r0, #1
	bl	fillwords		@ FillWords(Addr, Value, Size)
lin39:
	pop	{r4, pc}		@ restore r4 and return
lin40:
@ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@ Draw all other types of lines
@	DX, DY	(r10, r11)	delta X, delta Y
@	IX, IY	(r3, r4)	plotting increments along X and Y axes
@	SY	(r5)		flag: step along Y direction
@	D	(r6)		decision variable: move straight or diagonally
@	E, NE	(r7, r9)	East, Northeast (straight or diagonal)
@
	push	{r4-r11, lr}		@ save critical registers
	mov	r10, r1			@ X
	mov	r11, r2			@ Y
	ldr	r1, [r8, #penx-dseg]
	ldr	r2, [r8, #peny-dseg]

	subs	r10, r10, r1		@ DX:= X - PenX;	delta X
					@ if DX >= 0 then
	movge	r3, #1			@	IX:= 1;		increment
	rsblt	r10, #0			@ else	[DX:= -DX;
	movlt	r3, #-1			@	IX:= -1];

@ repeat for Y direction
	subs	r11, r11, r2		@ DY:= Y - PenY
					@ if DY >= 0 then
	movge	r4, #1			@	IY:= 1
	rsblt	r11, #0			@ else	[DY:= -DY;
	movlt	r4, #-1			@	IY:= -1];

	cmp	r10, r11		@ if DX >= DY then
	movge	r5, #0			@	SY:= false; step along X dir
	movlt	r5, #-1			@ else  SY:= true;  step along Y dir
	movlt	r6, r10			@	swap DX and DY
	movlt	r10, r11
	movlt	r11, r6

@ set up decision variables
	mov	r6, r11, lsl #1		@ D:= 2*DY - DX;
	sub	r6, r10
	mov	r7, r11, lsl #1		@ E:= 2*DY;
	sub	r9, r11, r10		@ NE:= 2*(DY-DX);
	lsl	r9, #1

@ set r11 for dotted and dashed lines
	ldr	r11, [r8, #scrdepth-dseg]
	bic	r11, #8
	cmp	r11, #16
	movlt	r11, r0, lsl #16	@ depth 8 and below use bits 15..8
	moveq	r11, r0			@ depth 16 and 24 use bits 31..24
	movgt	r11, #0			@ depth 32 does not have dotted lines
	and	r11, #0xFF000000
	b	lin70			@ for I:= 1, DX do

lin50:	cmp	r6, #0			@	[if D < 0 then
	bpl	lin60
	 add	r6, r7			@		[D:= D + E;
	 cmp	r5, #0			@		if SY then
	 addne	r2, r4			@			PenY:= PenY + IY
	 addeq	r1, r3			@		else	PenX:= PenX + IX
	 b	lin70
lin60:					@	else
	 add	r6, r9			@		[D:= D + NE; move diag.
	 add	r1, r3			@		PenX:= PenX + IX;
	 add	r2, r4			@		PenY:= PenY + IY;
lin70:					@		];
	lsls	r11, #1
	orrcs	r11, #0x01000000	@ rotate high byte left
	bcs	lin75			@ skip plotting point if carry id set
	 push	{r0-r3}			@ preserve: color, PenX, PenY, IX
	 push	{r1}			@ Point(PenX, PenY, C); plot initial pt.
	 push	{r2}
	 bl	intr41			@ (r0 = color)
	 pop	{r0-r3}
lin75:
	subs	r10, #1			@ for loop; I-- (same as DX--)
	bge	lin50
	str	r1, [r8, #penx-dseg]	@ make sure end point is set if
	str	r2, [r8, #peny-dseg]	@  dotted line
	pop	{r4-r11, pc}		@ restore registers and return

@-------------------------------------------------------------------------------
@ Quickly fill an array of words. This is logically equivalent to:
@  for I:= 0 to SIZE-1 do ARRAY(I):= Value; \ but is many times faster.
@  FILLWORDS(ADDRESS, VALUE, SIZE)
@   regs:    r2       r0     r1
@
fillwords:
	mov	r1, r0			@ get the number of words to fill
	pop	{r0}			@ get value to set the words to
	pop	{r2}			@ get starting address of the array

fmw10:	tst	r2, #0xF		@ fill array up to 1st 16-byte boundary
	beq	fmw20
	cmp	r1, #0			@ if remaining Size = 0 then all done
	beq	fmw90
	str	r0, [r2]
	add	r2, #4
	sub	r1, #1
	b	fmw10			@ loop
fmw20:
	lsrs	r3, r1, #2		@ r3 = Size/4 = number of 4-word
	beq	fmw30			@  chunks to store; skip if none
	push	{r4-r6}			@ save registers
	mov	r4, r0			@ copy Value into 3 more registers
	mov	r5, r0			@  giving a total of 4 words of Value
	mov	r6, r0
fmw25:	stmia	r2!, {r0, r4-r6}	@ store 4 words and bump up r2
	subs	r3, #1			@ count down number of 4-word chunks
	bne	fmw25			@ tight loop only takes 8 cycles
	pop	{r4-r6}			@ (unrolled loop goes no faster)
fmw30:
	ands	r1, #0x3		@ fill any remaining amount that's less
	beq	fmw90			@  than 4 words
fmw35:	str	r0, [r2]
	add	r2, #4
	subs	r1, #1
	bne	fmw35			@ loop
fmw90:	mov	pc, lr			@ return

@-------------------------------------------------------------------------------
@ Quickly fill an array of half-words.
@  FILLHALFWORDS(ADDRESS, VALUE, SIZE)
@       regs:    r2       r0     r1
@
fillhalfwords:
	mov	r1, r0			@ get the number of bytes to fill
	pop	{r0}			@ get value to set the bytes to
	pop	{r2}			@ get starting address of the array

fmh10:	tst	r2, #0xF		@ fill array up to 1st 16-byte boundary
	beq	fmh20
	cmp	r1, #0			@ if remaining Size = 0 then all done
	beq	fmh90
	strh	r0, [r2]
	add	r2, #2
	sub	r1, #1
	b	fmh10			@ loop
fmh20:
	lsrs	r3, r1, #3		@ r3 = Size/8 = number of 8 half-word
	beq	fmh30			@  chunks to store; skip if none
	push	{r4-r6}			@ save registers
	bic	r0, #0x00FF0000		@ copy Value into high word of r0
	bic	r0, #0xFF000000
	orr	r0, r0, lsl #16
	mov	r4, r0			@ copy Value into 3 more registers
	mov	r5, r0			@  giving a total of 16 bytes of Value
	mov	r6, r0
fmh25:	stmia	r2!, {r0, r4-r6}	@ store 8 half-words and bump up r2
	subs	r3, #1			@ count down num of 8 half-word chunks
	bne	fmh25			@ tight loop only takes 8 cycles
	pop	{r4-r6}			@ (unrolled loop goes no faster)
fmh30:
	ands	r1, #0x7		@ fill any remaining amount that's less
	beq	fmh90			@  than 8 half-words
fmh35:	strh	r0, [r2]
	add	r2, #2
	subs	r1, #1
	bne	fmh35			@ loop
fmh90:	mov	pc, lr			@ return

@-------------------------------------------------------------------------------
@ Move pen position for the start of a line or a device 6 character.
@  MOVE(X, Y)
@
intr43:	str	r0, [r8, #peny-dseg]
	pop	{r0}
	str	r0, [r8, #penx-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the color of a pixel on the screen at X, Y.
@  COLOR:= READPIX(X, Y)
@
intr44:	pop	{r2}			@ r0=Y, r2=X

intr44a:ldr	r3, [r8, #scrheight2-dseg] @ clip to frame buffer dimensions
	cmp	r0, r3
	bhs	rdp3			@ return if Y >= scrheight2 or Y < 0
	ldr	r3, [r8, #scrwidth-dseg]
	cmp	r2, r3
	bhs	rdp3			@ return if X >= scrwidth or X < 0

	mla	r1, r0, r3, r2		@ offset in pixels = Y*scrwidth + X
	ldr	r2, [r8, #scrdepth-dseg] @ bits per pixel (8, 16, 24, 32)
	ldr	r3, [r8, #framebuf-dseg]
	add	pc, r2, lsl #1		@ 8*2/4 = 4 instructions per case
	nop				@ NOTE: jump table  entry points follow:

rdp0:	ldrb	r0, [r3, r1]		@ < 8 bits per pixel (for eor mode)
	mov	pc, lr			@ return
rdp3:	mov	r0, #0			@ return 0 for read beyond screen limit
	mov	pc, lr			@ return

rdp8:	ldrb	r0, [r3, r1]		@ 8 bits per pixel
	mov	pc, lr			@ return
	nop
	nop

rdp16:	lsl	r1, #1			@ 16 bits per pixel
	ldrh	r0, [r3, r1]
	mov	pc, lr			@ return
	nop

rdp24:	ldr	r0, [r3, r1, lsl #2]	@ 32 bits per pixel
	mov	pc, lr			@ return
	nop
	nop

rdp32:	ldr	r0, [r3, r1, lsl #2]	@ 32 bits per pixel
	mov	pc, lr			@ return

@-------------------------------------------------------------------------------
@ Intrinsic to set up a video mode. This sets up the frame buffer for modes used
@  by the IBM-PC.
@  SETVID(MODE)
@  Outputs: framebuf, scrwidth, scrheight, scrheight2, scrdepth, bytesPerLine
@	fontheight, fontheight6, scrdepth0
@
intr45:	push	{lr}
	push	{r0}			@ save video mode

	adr	r2, svtbl		@ point to mode table
sv10:	ldr	r1, [r2]		@ get mode number from table
	cmp	r0, r1			@ is this the specified mode?
	beq	sv20			@ branch if so
	 cmp	r1, #-1			@ default entry?
	 beq	sv20			@ branch if so
	  add	r2, #4*4		@ point to next table entry
	  b	sv10			@ loop
sv20:
	push	{r2}
	mov	r0, #0x07		@ Attrib($07): white letters on black
	bl	intr69			@  background for device 0
	mov	r0, #0x0C		@ must clear the device 0 screen even
	bl	conout			@  when setting a graphic mode (cursor)
	pop	{r2}

	ldr	r0, [r2]		@ if mode = 3 then restore the original
	cmp	r0, #3			@  width, height, and depth
	ldreq	r0, [r8, #scrwidth0-dseg]
	ldrne	r0, [r2, #4]		@ get width
	push	{r0}
	ldreq	r0, [r8, #scrheight0-dseg]
	ldrne	r0, [r2, #8]		@ get height
	push	{r0}
	ldreq	r0, [r8, #scrdepth0-dseg]
	ldrne	r0, [r2, #12]		@ get bits per pixel

	bl	intr84			@ set up frame buffer

	pop	{r0}			@ get video mode
	cmp	r0, #4			@ if mode < 4 then
	movlo	r0, #-1			@ turn on flashing cursor 
	bllo	intr88			@ ShowCursor(true)
	pop	{pc}

@		 mode width height bits/pixel
svtbl:	.word	 0x00,  320, 200,  0	@ for 40x25 text
	.word	 0x01,  320, 200,  0
	.word	 0x02, 1280, 720, 16
	.word	 0x03, 1280, 720, 16
	.word	 0x04,  320, 200,  0	@ for CGA compatibility
	.word	 0x05,  320, 200,  0
	.word	 0x06,  640, 200,  0
	.word	 0x07, 1280, 720,  0
	.word	 0x0D,  320, 200,  0	@ actually depth = 8; 0 used for intr41
	.word	 0x0E,  640, 200,  0	@  which enables pixels to be 'eor-ed'
	.word	 0x0F,  640, 350,  0	@ for EGA compatibility
	.word	 0x10,  640, 350,  0
	.word	 0x11,  640, 480,  0
	.word	 0x12,  640, 480,  0	@ for VGA compatibility
	.word	 0x13,  320, 200,  8
	.word	 0x6A,  800, 600,  0	@ for VESA compatibility
	.word	0x100,  640, 400,  8
	.word	0x101,  640, 480,  8
	.word	0x102,  800, 600,  0
	.word	0x103,  800, 600,  8
	.word	0x104, 1024, 768,  0
	.word	0x105, 1024, 768,  8
	.word	0x106, 1280,1024,  0
	.word	0x107, 1280,1024,  8
	.word	0x10D,  320, 200, 16	@ actually depth = 15
	.word	0x10E,  320, 200, 16
	.word	0x10F,  320, 200, 24
	.word	0x110,  640, 480, 16	@ 15
	.word	0x111,  640, 480, 16
	.word	0x112,  640, 480, 24
	.word	0x113,  800, 600, 16	@ 15
	.word	0x114,  800, 600, 16
	.word	0x115,  800, 600, 24
	.word	0x116, 1024, 768, 16	@ 15
	.word	0x117, 1024, 768, 16
	.word	0x118, 1024, 768, 24
	.word	0x119, 1280,1024, 16	@ 15
	.word	0x11A, 1280,1024, 16
	.word	0x11B, 1280,1024, 24
	.word	   -1,  640, 480,  8	@ default

@-------------------------------------------------------------------------------
@ Intrinsic to reserve heap space for real numbers. This reserves tos*8
@  bytes of heap space and returns the starting 32-bit address converted
@  to a real value on the stack.
@  REAL VARIABLE:= RLRES(INTEGER)
@
intr46:	pop	{r0}			@ real intrinsic, so tos is on the stack
	mov	r1, r9			@ get base address of reserved space
	add	r9, r0, lsl #3		@ increase heap pointer by 8 * argument
	sub	sp, #8			@ push base address on the stack
	str	r1, [sp]		@  in the low half of a real
	mov	r0, #0x3F000000		@ the high half is set to 1.0
	orr	r0, #0x00F00000		@ thus tos = 0x3FF0_0000_base_address
	str	r0, [sp, #4]
	ldr	r0, [r8, #heaphi-dseg]	@ test for heap memory overlfow
	cmp	r9, r0
	blo	flres1
	 mov	r0, #2			@ flag OUT OF MEMORY error
	 push	{lr}
	 bl	error
	 pop	{lr}
flres1:	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to input a floating-point number from the specified device.
@  REAL VARIABLE:= RLIN(DEVICE)
@
intr47:	pop	{r0}			@ get device number (from stack)
	str	r0, [r8, #nowdev-dseg]
	push	{lr}
	bl	flin			@ input a real into d0
	pop	{lr}
	vpush	{d0}
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to output a floating point number to the specified device.
@  RLOUT(DEVICE, REAL)
@
intr48:	vpop	{d0}
	pop	{r0}			@ get device number (from stack)
	str	r0, [r8, #nowdev-dseg]
	b	flout			@ branch and return

@-------------------------------------------------------------------------------
@ Intrinsic to convert the integer on tos to a floating point number.
@  Note that this is compiled as a real intrinsic with tos on the stack.
@  REAL VARIABLE:= FLOAT(INTEGER)
@
intr49:	pop	{r0}
	fmsr	s2, r0			@ move integer in r0 to s2
	fsitod	d0, s2			@ convert signed integer in s2 to d0
	sub	sp, #8			@ push real
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to convert real on tos to nearest integer.
@  Note that this is compiled as an integer intrinsic with tos in r0.
@  INTEGER:= FIX(REAL)
@
intr50:	fmrx	r1, fpscr		@ clear exception bits in status reg.
	and	r1, #0xFFFFFFF0
	fmxr	fpscr, r1

	push	{r0}			@ balance stack
	fldd	d0, [sp]		@ pop d0
	add	sp, #8
	ftosid	s2, d0			@ convert d0 to signed integer in s2
	fmrs	r0, s2			@ move s2 to r0

	fmrx	r1, fpscr		@ get exception bits from status reg.
	tst	r1, #0x0F
	moveq	pc, lr			@ return if none are set

	push	{r0, lr}
	mov	r0, #9			@ else report error 9: fix overflow
	bl	error
	pop	{r0, pc}

@-------------------------------------------------------------------------------
@ Intrinsic to take the absolute value of the real number on tos.
@  REAL VARIABLE:= RLABS(REAL)
@
intr51:	fldd	d0, [sp]
	fabsd	d0, d0
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to set the format of floating point numbers output by RLOUT.
@  FORMAT(F1, F2)
@   F1 > 0:  decimal notation, with F1 digits in front of decimal point
@   F1 = 0:  scientific notation	   1.23E+023
@   F1 < 0:  engineering notation	 123.0E+021
@   F2:  number of digits after the decimal point
@
intr52:	pop	{r0}
	str	r0, [r8, #form2-dseg]
	pop	{r0}
	str	r0, [r8, #form1-dseg]
	mov	pc, lr

	.data
	.align	2
form1:	.word	5
form2:	.word	5
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to return the square root of top of stack.
@  TOS:= SQRT(TOS)
@
intr53:	fldd	d0, [sp]
	fsqrtd	d0, d0			@ do square root
	fstd	d0, [sp]		@ return result in TOS

	fcmpd	d0, d0			@ test for not a number (NAN)
	fmstat				@ get status
	movvc	pc, lr			@ return if ok (V flag not set)

	mov	r0, #10			@ else report error 10: sqrt < 0.0
	push	{lr}
	bl	error
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to return the natural logarithm (base e) of top of stack.
@  TOS:= LN(TOS)
@
intr54:	fldd	d0, [sp]
	push	{lr}
	bl	log			@ log = ln
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the exponential (e^TOS) of top of stack.
@  TOS:= EXP(TOS)
@
intr55:	fldd	d0, [sp]
	push	{lr}
	bl	exp
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the sine of top of stack.
@  TOS:= SIN(TOS)
@
intr56:	fldd	d0, [sp]
	push	{lr}
	bl	sin
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the angle (+/-pi) whose tangent is Y/X.
@  TOS:= ATAN2(Y, X)
@
intr57:	vpop	{d1}			@ TOS = X = d1
	fldd	d0, [sp]		@ NOS = Y = d0
	push	{lr}
	bl	atan2
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return A modulo B = A - INT(A/B)*B.
@  TOS:= MOD(A, B)
@
intr58:	vpop	{d1}			@ TOS = X = d1
	fldd	d0, [sp]		@ NOS = Y = d0
	push	{lr}
	bl	fmod
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the logarithm (base 10) of top of stack.
@  TOS:= LOG(TOS)
@
intr59:	fldd	d0, [sp]
	push	{lr}
	bl	log10
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the cosine of top of stack.
@  TOS:= COS(TOS)
@
intr60:	fldd	d0, [sp]
	push	{lr}
	bl	cos
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the tangent of top of stack.
@  TOS:= TAN(TOS)
@
intr61:	fldd	d0, [sp]
	push	{lr}
	bl	tan
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the arc-sine of top of stack.
@  TOS:= ASIN(TOS)
@
intr62:	fldd	d0, [sp]
	push	{lr}
	bl	asin
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return the arc-cosine of top of stack.
@  TOS:= ACOS(TOS)
@
intr63:	fldd	d0, [sp]
	push	{lr}
	bl	acos
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to round down to a (real) integer value.
@  TOS:= FLOOR(TOS)
@
intr64:	fldd	d0, [sp]
	push	{lr}
	bl	floor
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to round up to a (real) integer value.
@  TOS:= CEIL(TOS)
@
intr65:	fldd	d0, [sp]
	push	{lr}
	bl	ceil
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to return X raised to the Y power.
@  TOS:= POW(X, Y)
@
intr66:	vpop	{d1}			@ TOS = Y = d1
	fldd	d0, [sp]		@ NOS = X = d0
	push	{lr}
	bl	pow
	pop	{lr}
	fstd	d0, [sp]		@ return result in TOS
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to set the color attribute used by output device 6.
@  ATTRIB(COLORS)
@
intr69:	str	r0, [r8, #attrib-dseg]	@ for device 6
	mov	pc, lr

	.data
	.align	2
attrib:	.word	0x07			@ white on black for 8 & 16-bit modes
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to set up a window for output device 6.
@  SETWIND(X0, Y0, X1, Y1, MODE, FILL)
@          r2  r1  r3  r2  r1    r0
@
intr70:	pop	{r1-r3}			@ set window parameters
	str	r1, [r8, #windmode-dseg]
	str	r2, [r8, #windy1-dseg]
	str	r3, [r8, #windx1-dseg]
	pop	{r1, r2}
	str	r1, [r8, #windy0-dseg]
	str	r2, [r8, #windx0-dseg]
	cmp	r0, #0			@ skip if window is not filled
	beq	wind30

@ Erase window and fill with background color specified by Attrib intrinsic (69)
	push	{r4-r7, r10, r11, lr}
	ldr	r0, [r8, #attrib-dseg]	@ set background color
	ldr	r1, [r8, #scrdepth-dseg]
	cmp	r1, #8			@ depth	bgnd
	andlt	r0, #0xF0		@  4	a>>4&F
	lsrlt	r0, #4			@  8	a>>8
	lsreq	r0, #8			@ 16	a>>16 WARNING: conflicts
	lsrgt	r0, #16			@ 24	0
	cmp	r1, #24			@ 32	0
	movge	r0, #0

@ for y:= windy0*fontheight to (windy1+1)*fontheight-1 do
	ldr	r11, [r8, #fontheight-dseg]
	ldr	r7, [r8, #windy1-dseg]	@ limity:= (windy1+1)*fontheight
	add	r7, #1
	mul	r7, r7, r11
	ldr	r6, [r8, #windy0-dseg]	@ y:= windy0*fontheight
	mul	r6, r6, r11

@   for x:= windx0*8 to (windx1+1)*8-1 do
wind10:	ldr	r4, [r8, #windx1-dseg]	@ limitx:= (windx1+1)*8
	add	r4, #1
	lsl	r4, #3
	ldr	r5, [r8, #windx0-dseg]	@ x:= windx0*8
	lsl	r5, #3

@     Point(x, y, bkgndcolor);
wind20:	mov	r2, r5			@ x
	mov	r1, r6			@ y
	bl	intr41a			@ Point

	add	r5, #1			@ next x
	cmp	r5, r4
	blt	wind20

	add	r6, #1			@ next y
	cmp	r6, r7
	blt	wind10
	pop	{r4-r7, r10, r11, lr}
wind30:
	ldr	r0, [r8, #windx0-dseg]	@ set cursor to upper-left corner
	lsl	r0, #3			@  of window
	str	r0, [r8, #penx-dseg]	@ penx:= windx0*8
	ldr	r0, [r8, #windy0-dseg]	@ peny:= windy0*fontheight
	ldr	r1, [r8, #fontheight-dseg]
	mul	r0, r0, r1
	str	r0, [r8, #peny-dseg]
	mov	pc, lr			@ return

	.data
	.align	2
windx0:	.word	0			@ character coordinates of upper-left
windy0:	.word	0			@  and lower-right corners of window
windx1:	.word	1280/8-1		@ default scrwidth/8 - 1
windy1:	.word	720/16-1		@ default scrheight/fontheight - 1
windmode:.word	0			@ default to scroll mode
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to output a text string to the specified device. The starting
@  address is in r0, and is terminated by a space character with its bit 7 set.
@  RAWTEXT(6, "ͻ ")
@
intr71:	pop	{r1}			@ get device number
	str	r1, [r8, #nowdev-dseg]	@ set the device number
	push	{r4, lr}		@ save registers
	mov	r4, r0			@ put pointer into preserved register
	b	rtxt20			@ enter loop
rtxt10:	bl	chout			@ output character
rtxt20:	ldrb	r0, [r4]		@ get char
	add	r4, #1			@ increment pointer
	cmp	r0, #0			@ (in case string 0 is active)
	beq	rtxt90
	cmp	r0, #0xA0		@ loop until terminating space?
	bne	rtxt10
rtxt90:	pop	{r4, pc}		@ restore registers and return

@-------------------------------------------------------------------------------
@ Intrinsic to change the color of an area of text.
@  HILIGHT(X0, Y0, X1, Y1, ATTRIBUTE)
@
intr72:	mov	r12, r0			@ save attribute into r12 (scratch)
	pop	{r0-r3}			@ r0=Y1, r1=X1, r2=Y0, r3=X0
	cmp	r3, r1			@ if X0 > X1 then
	eorgt	r3, r1			@   swap X0, X1
	eorgt	r1, r3
	eorgt	r3, r1
	cmp	r2, r0			@ if Y0 > Y1 then
	eorgt	r2, r0			@   swap Y0, Y1
	eorgt	r0, r2
	eorgt	r2, r0
@ X0	Y0	X1	Y1
@ r3	r2	r1	r0

	push	{r4-r11, lr}
	ldr	r4, [r8, #penx-dseg]	@ save current pen position
	ldr	r5, [r8, #peny-dseg]
	push	{r4, r5}
	mov	r10, r12		@ save color attribute (again)

	mov	r9, r10			@ get new foreground & background colors
	ldr	r7, [r8, #scrdepth-dseg]@ r7	r9	r10
	cmp	r7, #8			@ depth	fgnd	bgnd
	andlt	r9, #0x0F		@  0	a&F	a>>4&F
	andlt	r10, #0xF0		@  4	a&F	a>>4&F
	lsrlt	r10, #4			@  8	a	a>>8
	lsreq	r10, #8			@ 16	a	a>>16
	lsrgt	r10, #16		@ 24	a	0
	cmp	r7, #24			@ 32	a	0
	movge	r10, #0
@ X	Y	I	J	NewFg	NewBg	OldBg
@ r4	r7	r5	r6	r9	r10	r11

	mov	r7, r2			@ for Y:= Y0 to Y1 do
hil10:	push	{r0, r1, r3}

	mov	r4, r3			@   for X:= X0 to X1 do
hil20:	push	{r1}
	mov	r0, r4, lsl #3		@ OldBg:= ReadPix(X*8+7, (Y+1)*FontHeight-2)
	add	r0, #7
	push	{r0}
	ldr	r0, [r8, #fontheight-dseg]
	add	r1, r7, #1
	mul	r0, r1
	sub	r0, #2
	bl	intr44
	mov	r11, r0

	ldr	r0, [r8, #fontheight-dseg] @ for J:= Y*FontHeight to (Y+1)*FontHeight-1 do
	mul	r6, r7, r0
	add	r1, r7, #1
	mul	r0, r1
	sub	r0, #1
hil30:	push	{r0}

	mov	r5, r4, lsl #3		@ for I:= X*8 to X*8+7 do
	add	r0, r5, #7
hil40:	push	{r0}

	push	{r5}			@   I
	mov	r0, r6			@   J
	bl	intr44			@   if ReadPix(I, J) = OldBg then
	cmp	r0, r11
	moveq	r0, r10			@        Point(I, J, NewBg)
	movne	r0, r9			@   else Point(I, J, NewFg)
	push	{r5}			@   I
	push	{r6}			@   J
	bl	intr41			@   plot r0

	pop	{r0}			@ next I
	add	r5, #1
	cmp	r5, r0
	ble	hil40

	pop	{r0}			@ next J
	add	r6, #1
	cmp	r6, r0
	ble	hil30

	pop	{r1}			@ next X
	add	r4, #1
	cmp	r4, r1
	ble	hil20

	pop	{r0, r1, r3}		@ next Y
	add	r7, #1
	cmp	r7, r0
	ble	hil10

	pop	{r4, r5}		@ restore original pen position
	str	r4, [r8, #penx-dseg]
	str	r5, [r8, #peny-dseg]
	pop	{r4-r11, pc}		@ return

@-------------------------------------------------------------------------------
@ Intrinsic to allocate a block of memory and return its starting address in r0.
@  The number of bytes to allocate is passed in r0. The allocated memory remains
@  available until the Release intrinsic is called.
@  ADDRESS:= MALLOC(BYTES)
@
intr73:	push	{lr}
	bl	malloc
	cmp	r0, #0			@ if 0 then out of memory error
	bne	mal90			@ skip if ok
	 mov	r0, #2			@ runtime error # 2: out of memory
	 bl	error
	 mov	r0, #0
mal90:	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to release memory allocated by MAlloc. The address of the block to
@  release is passed in r0. (Any allocated memory is automatically released when
@  a program terminates.)
@  RELEASE(ADDRESS)
@
intr74:	b	free

@-------------------------------------------------------------------------------
@ Intrinsic to turn on or off control-c trapping. Passing 'true' prevents
@  Ctrl-C and Ctrl-Break keys from aborting a program.
@  TRAPC(BOOLEAN)
@
intr75:	cmp	r0, #0			@ is flag false?
	beq	ctc10			@ branch if so
	 mov	r0, #-1			@ else set flag to 'true'
ctc10:	str	r0, [r8, #ctcflag-dseg]
	mov	pc, lr

	.data
	.align	2
ctcflag:.word	0			@ default is 'false' (abort on Ctrl-C)
cinflag:.word	0			@ default is 'false' no Ctrl-C
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to return 'true' if Ctrl-C or Ctrl-Break has been struck.
@  BOOLEAN:= TESTC
@
intr76:	ldr	r0, [r8, #cinflag-dseg]	@ get flag
	mov	r1, #0			@ clear flag
	str	r1, [r8, #cinflag-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to turn the display of the mouse pointer on or off.
@  SHOWMOUSE(BOOL)
@
intr77:	push	{lr}
	cmp	r0, #0
	beq	sm50			@ branch if turning off pointer

	ldr	r1, [r8, #mouseptron-dseg] @ if mouseptron is already on then
	cmp	r1, #0			@  return
	bne	sm90

	str	r0, [r8, #mouseptron-dseg] @ mouseptron:= true

	ldr	r0, [r8, #micex-dseg]	@ record background's coordinates
	str	r0, [r8, #mousex-dseg]
	ldr	r0, [r8, #micey-dseg]
	str	r0, [r8, #mousey-dseg]
	bl	savembg			@ save background
	bl	showmptr		@ display mouse pointer
	b	sm90
sm50:
	ldr	r1, [r8, #mouseptron-dseg] @ if mouseptron is already off then
	cmp	r1, #0			@  return
	beq	sm90

	mov	r0, #0			@ mouseptron:= false
	str	r0, [r8, #mouseptron-dseg]
	bl	restmbg			@ restore background to remove pointer

sm90:	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to move the mouse pointer as needed to track position of the mouse.
@  MOVEMOUSE
@
intr78:	push	{r4, lr}
	bl	intr86			@ get memory pointer to mouse state data
	mov	r4, r0			@ save it in r4
	ldr	r0, [r4, #0]		@ get current mouse X position
	ldr	r1, [r8, #mousex-dseg]	@ get old mouse X position
	cmp	r0, r1			@ branch if mouse has moved
	bne	mm10

	ldr	r0, [r4, #4]		@ get current mouse Y position
	ldr	r1, [r8, #mousey-dseg]	@ get old mouse Y position
	cmp	r0, r1			@ branch if mouse hasn't moved
	beq	mm90
mm10:
	bl	restmbg			@ restore background at old location
	ldr	r0, [r4, #0]		@ record mouse's new location
	str	r0, [r8, #mousex-dseg]
	ldr	r0, [r4, #4]
	str	r0, [r8, #mousey-dseg]
	bl	savembg			@ save background at new location
	bl	showmptr		@ show mouse pointer at new location
mm90:	pop	{r4, pc}

@-------------------------------------------------------------------------------
@ Display mouse pointer at mousex, mousey.
@
showmptr:
	push	{r5-r6, lr}
	mov	r6, #mouseheight-1	@ for r6:= mouseheight-1 downto 0 do
smp10:	mov	r5, #mousewidth-1	@   for r5:= mousewidth-1 downto 0 do
smp20:	ldr	r2, [r8, #mousex-dseg]	@     Point(r5+X, r6+Y, Color
	add	r2, r5
	ldr	r1, [r8, #mousey-dseg]
	add	r1, r6

	mov	r0, #mousewidth		@     get Color from Ptr(r6, r5)
	mla	r0, r0, r6, r5		@     r6*MouseW + r5 + Ptr
	adr	r3, smptr
	ldrb	r0, [r0, r3]

	cmp	r0, #'#'
	bne	smp30
	 mov	r0, #0			@     Color = black
	 b	smp35
smp30:
	cmp	r0, #'.'
	bne	smp40			@     skip if transparent
	 mov	r0, #0xFFFFFFFF		@     Color = bright white
	 ldr	r3, [r8, #scrdepth-dseg]
	 cmp	r3, #8
	 movle	r0, #0x1F
smp35:	 bl	intr41a
smp40:
	subs	r5, #1			@   next I
	bpl	smp20
	subs	r6, #1			@ next J
	bpl	smp10
	pop	{r5-r6, pc}

mousewidth	= 12			@ dimensions (in pixels) of rectangle
mouseheight	= 21			@  encompassing mouse pointer image

smptr:	.ascii	"#           "
	.ascii	"##          "
	.ascii	"#.#         "
	.ascii	"#..#        "
	.ascii	"#...#       "
	.ascii	"#....#      "
	.ascii	"#.....#     "
	.ascii	"#......#    "
	.ascii	"#.......#   "
	.ascii	"#........#  "
	.ascii	"#.........# "
	.ascii	"#......#####"
	.ascii	"#...#..#    "
	.ascii	"#..##..#    "
	.ascii	"#.#  #..#   "
	.ascii	"##   #..#   "
	.ascii	"#     #..#  "
	.ascii	"      #..#  "
	.ascii	"       #..# "
	.ascii	"       #..# "
	.ascii	"        ##  "
	.align	2

	.data
	.align	2
mousex:	.word	0			@ mouse pointer position on screen
mousey:	.word	0
pmousebg:.word	mousebg			@ pointer to saved background image
mouseptron: .word	0		@ boolean: mouse pointer is being shown
	.text

	.bss
	.align	2
mousebg:.skip	mousewidth * mouseheight * 4
	.text

@-------------------------------------------------------------------------------
@ Save image area on screen that's about to be overwritten by the mouse pointer.
@ In other words, copy the image on the screen at mousex, mousey to mousebg.
@
savembg:push	{r5-r6, lr}

	mov	r6, #mouseheight-1	@ for r6:= mouseheight-1 downto 0 do
smb10:	mov	r5, #mousewidth-1	@   for r5:= mousewidth-1 downto 1 do
smb20:	ldr	r2, [r8, #mousex-dseg]	@     r0:= ReadPix(r5+X, r6+Y)
	add	r2, r5
	ldr	r0, [r8, #mousey-dseg]
	add	r0, r6

	bl	intr44a			@     r0=y, r2=x
	mov	r2, #mousewidth		@     mousebg(r6, r5):= r0
	mla	r2, r2, r6, r5		@     (r6*MouseW + r5)*4 + Ptr
	ldr	r3, [r8, #pmousebg-dseg]
	str	r0, [r3, r2, lsl #2]

	subs	r5, #1			@   next I
	bpl	smb20
	subs	r6, #1			@ next J
	bpl	smb10
	pop	{r5-r6, pc}

@-------------------------------------------------------------------------------
@ Restore back to the screen the image that was overwritten by the mouse pointer
@  In other words, copy the image saved in mousebg to screen at mousex, mousey.
@
restmbg:push	{r5-r6, lr}

	mov	r6, #mouseheight-1	@ for r6:= mouseheight-1 downto 0 do
rmb10:	mov	r5, #mousewidth-1	@   for r5:= mousewidth-1 downto 0 do
rmb20:	ldr	r2, [r8, #mousex-dseg]	@     r0:= ReadPix(r5+X, r6+Y)
	add	r2, r5
	ldr	r1, [r8, #mousey-dseg]
	add	r1, r6

	mov	r0, #mousewidth		@     r0:= mousebg(r6, r5)
	mla	r0, r0, r6, r5		@     (r6*MouseW + r5)*4 + Ptr
	ldr	r3, [r8, #pmousebg-dseg]
	ldr	r0, [r3, r0, lsl #2]
	bl	intr41a			@ r1=y r2=x

	subs	r5, #1			@   next I
	bpl	rmb20
	subs	r6, #1			@ next J
	bpl	rmb10
	pop	{r5-r6, pc}

@-------------------------------------------------------------------------------
@ Intrinsic to return the RGB value for the given palette register.
@  RGB:= GETPALETTE(REG)
@
intr80:	ldr	r3, ppalette
	ldr	r0, [r3, r0, lsl #2]
	bic	r0, #0xFF000000
	mov	pc, lr

ppalette: .word	palette			@ pointer to palette array

@-------------------------------------------------------------------------------
@ Intrinsic to quickly copy an image in an array to video display memory.
@    PAINT(X,  Y,  W,  H,  IMAGE,  W2)
@	   4   3   2   1   0       r0
@  X,Y are the pixel coordinates of the upper-left corner of the image on screen
@  W,H are the width and height (in pixels) of the image as displayed.
@  Image is the address of the image array data.
@  W2 is the width (in pixels) of the image array data, which can be greater
@  than or equal to W. If it is greater than W then the right side of Image will
@  be clipped off. If an offset is added to Image (such as Image + Y2 + X2*W2)
@  then Image can be panned within the W-H window.
@
@ In 24-bit color modes, the Image array and frame buffer use 4 bytes per pixel.
@
@ Modulo is the amount to add to get to the start of the next scan line. It's
@  (usually) equal to 0 when the entire width of the screen is used.
@
intr81:					@ W2 (ignored for now)
	pop	{r0}			@ Image
	str	r0, [r8, #image-dseg]
	pop	{r0}			@ H
	str	r0, [r8, #imagehi-dseg]

	ldr	r2, [r8, #scrdepth-dseg]@ bits per pixel, use only 8, 16 or 32
	cmp	r2, #16
	movlt	r2, #0			@ amount to shift left to convert bits
	moveq	r2, #1			@  per pixel to bytes per pixel
	movgt	r2, #2

	pop	{r0}			@ W
	lsl	r3, r0, r2		@ image width in bytes

	pop	{r0}			@ Y, offset from upper-left corner of
	ldr	r1, [r8, #scrwidth-dseg]@    screen = Y*scrwidth + X
	mul	r0, r0, r1
	pop	{r1}			@ X
	add	r0, r1
	lsl	r0, r0, r2		@ convert pixel offset to byte offset
	ldr	r1, [r8, #framebuf-dseg]
	push	{r4-r12, lr}
	add	r4, r0, r1

	ldr	r0, [r8, #scrwidth-dseg]@ modulo = scrwidth - image_width
	lsl	r5, r0, r2		@ in bytes
	sub	r5, r3
	ldr	r2, [r8, #imagehi-dseg]	@ set up scan line counter
	ldr	r1, [r8, #image-dseg]	@ point to image data

pnt10:	mov	r0, r3, lsr #5		@ set loop counter for width/32 bytes
pnt20:	ldmia	r1!, {r6-r12, lr}	@ get 32 bytes and bump r1
	subs	r0, #1			@ loop width/32 times
	stmia	r4!, {r6-r12, lr}	@ put 32 bytes and bump r4
	bne	pnt20
	add	r4, r5			@ skip screen bytes (modulo)
	subs	r2, #1			@ loop for height scan lines
	bne	pnt10
	pop	{r4-r12, pc}

	.data
	.align	2
image:	.word	0
imagehi:.word	0
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to get the current time in microseconds. Range: 0..2^32 (1.19 hours)
@  TIME:= GETTIME
@
intr82:	push	{lr}
	push	{r0, r1}		@ make "struct" space on stack
	mov	r0, sp			@ point to it
	mov	r1, #0			@ error = gettimeofday(&struct, NULL)
	bl	gettimeofday
	pop	{r1}			@ r1 = seconds since 1-Jan-1970
	pop	{r0}			@ r0 = additional microseconds
	mov	r2, #0xF4000		@ 1,000,000 = 0xF4240
	orr	r2, #0x00240
	mla	r0, r1, r2, r0		@ r0:= r1*1,000,000 + r0
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to set the backup flag so the last input character gets re-read.
@  BACKUP
@
intr83:	mov	r0, #1
	str	r0, [r8, #backfl-dseg]
	mov	pc, lr

@-------------------------------------------------------------------------------
@ Intrinsic to set up a graphic frame buffer's width, height, and depth.
@  SETFB(WIDTH, HEIGHT, DEPTH)
@  Outputs: framebuf, scrwidth, scrheight, scrheight2, scrdepth, bytesPerLine,
@	mouseptron, fontheight, fontheight6, windx0, windy0, windx1, windy1
@
@  char *fbp = 0;
@  struct fb_var_screeninfo vinfo;
@  struct fb_fix_screeninfo finfo;
@
intr84:	and	r0, #0x38		@ restrict depth to legal values,
	cmp	r0, #0x20		@  which are:
	movhi	r0, #0x20		@  0, 8, 16, 24, 32
	str	r0, [r8, #scrdepth-dseg]@ get arguments
	pop	{r0}
	str	r0, [r8, #scrheight-dseg]
	lsl	r0, #1			@ double for page 2
	str	r0, [r8, #scrheight2-dseg]
	pop	{r0}
	str	r0, [r8, #scrwidth-dseg]

	ldr	r0, [r8, #scrheight-dseg]
	mov	r1, #16			@ assume fontheight = 16
	cmp	r0, #352		@ if screenheight <= 350, use 8x14 font
	movle	r1, #14
	cmp	r0, #200		@ if screenheight <= 200, use 8x8 font
	movle	r1, #8			@ (this makes it agree with the PC)
	str	r1, [r8, #fontheight-dseg]
	str	r1, [r8, #fontheight6-dseg]

	push	{lr}
	bl	openo6			@ reset device 6's colors, window, etc.
	mov	r0, #0			@ make sure flashing cursor is off
	bl	intr88			@ ShowCursor(false)
	mov	r0, #0			@ mouseptron:= false
	str	r0, [r8, #mouseptron-dseg]
	pop	{lr}

	stmfd	sp!, {fp, lr}
	add	fp, sp, #4
	sub	sp, sp, #24
	str	r0, [fp, #-16]
	str	r1, [fp, #-20]
	mov	r3, #0			@ int fbfd = 0;
	str	r3, [fp, #-8]
	ldr	r0, .L10		@ fbfd = open("/dev/fb0", O_RDWR);
	mov	r1, #2
	bl	open
	str	r0, [fp, #-8]
	str	r0, [r8, #fbfd-dseg]

	str	r0, [r8, #palfbfd-dseg]	@ (for SETPALETTE)
	ldr	r0, [fp, #-8]		@ ioctl(fbfd, FBIOGET_VSCREENINFO,
	mov	r1, #0x4600		@       &vinfo);
	ldr	r2, .L10+4
	bl	ioctl
	ldr	r3, .L10+4

@ Change variable information
	ldr	r3, .L10+4		@ &vinfo
	ldr	r2, [r8, #scrwidth-dseg]
	str	r2, [r3, #0]		@ xres
	str	r2, [r3, #8]		@ xres_virtual
	ldr	r2, [r8, #scrheight-dseg]
	str	r2, [r3, #4]		@ yres
	ldr	r2, [r8, #scrheight2-dseg]
	str	r2, [r3, #12]		@ yres_virtual
	ldr	r2, [r8, #scrdepth-dseg]
	cmp	r2, #16			@ if scrdepth=24 then use 32, but
	movhi	r2, #32			@ if scrdepth>16 ! scrdepth<0 then 32
	movlo	r2, #8			@ if scrdepth<16 then 8
	str	r2, [r3, #24]		@ bits_per_pixel
	ldr	r0, [fp, #-8]		@ if (ioctl(fbfd, FBIOPUT_VSCREENINFO,
	ldr	r1, .L10+8		@     &vinfo))
	ldr	r2, .L10+4
	bl	ioctl
	movs	r3, r0
	beq	.L5
	 ldr	r0, .L10+12		@ printf("Error setting variable info...
	 bl	puts
.L5:
	ldr	r0, [r8, #scrwidth-dseg]
	str	r0, [r8, #xres-dseg]	@ set up vinfo struc (for intr99)
	str	r0, [r8, #xvres-dseg]
	ldr	r0, [r8, #scrheight-dseg]
	str	r0, [r8, #yres-dseg]
	str	r0, [r8, #yvres-dseg]
	ldr	r0, [r8, #scrdepth-dseg]
	cmp	r2, #16			@ if scrdepth=24 then use 32, but
	movhi	r2, #32			@ if scrdepth>16 ! scrdepth<0 then 32
	movlo	r2, #8			@ if scrdepth<16 then 8
	str	r0, [r8, #pixbits-dseg]
	mov	r0, #0
	str	r0, [r8, #xoff-dseg]
	str	r0, [r8, #yoff-dseg]
	str	r0, [r8, #grays-dseg]

@ Get fixed screen information
	ldr	r0, [fp, #-8]		@ if (ioctl(fbfd, FBIOGET_FSCREENINFO,
	ldr	r1, .L10+16
	ldr	r2, .L10+20		@     &finfo))
	bl	ioctl
	movs	r3, r0
	beq	.L6
	 ldr	r0, .L10+24		@ printf("Error reading fixed info....
	 bl	puts
.L6:
@ Map frame buffer (fb) to user memory
	ldr	r3, .L10+4		@ screensize = vinfo.xres*vinfo.yres*2
	ldr	r3, [r3, #0]		@  xres
	ldr	r2, .L10+4		@  vinfo
	ldr	r2, [r2, #4]		@  yres
	mul	r3, r2, r3
	lsl	r3, #1			@  *2 = screensize in pixels
	ldr	r2, [r8, #scrdepth-dseg]
	cmp	r2, #16			@ if scrdepth=24 then use 32, but
	movhi	r2, #32			@ if scrdepth>16 ! scrdepth<0 then 32
	movlo	r2, #8			@ if scrdepth<16 then 8
	lsr	r2, #3
	mul	r3, r3, r2
	str	r3, [fp, #-12]		@ screensize in bytes

	ldr	r2, [fp, #-8]		@ fbfd
	str	r2, [sp, #0]
	mov	r2, #0
	str	r2, [sp, #4]
	mov	r0, #0			@ fbp = (char*)mmap(0,
	mov	r1, r3			@        screensize,
	mov	r2, #3			@        PROT_READ | PROT_WRITE,
	mov	r3, #1			@        MAP_SHARED,
	bl	mmap			@        fbfd,
	mov	r2, r0			@        0);
	ldr	r3, .L10+28
	str	r2, [r3, #0]
	ldr	r3, .L10+28		@ if ((int)fbp == -1)
	ldr	r3, [r3, #0]
	cmn	r3, #1
	bne	.L7
	 ldr	r0, .L10+32		@ printf("Failed to mmap.\n");
	 bl	puts
.L7:
	str	r2, [r8, #framebuf-dseg]

	ldr	r0, .L10+20		@ finfo
	ldr	r0, [r0, #44]		@ line_length in bytes
	str	r0, [r8, #bytesPerLine-dseg]
	bl	intr40			@ clear memory
	bl	intr85			@ set mouse pointer to center of screen

@ Set up color palette if 8 bits per pixel
	ldr	r0, [r8, #scrdepth-dseg]@ bits per pixel
	cmp	r0, #8
	bgt	not8
	 ldr	r0, [fp, #-8]		@ ioctl(fbfd, FBIOPUTCMAP, &palstruc);
	 mov	r1, #0x4605&0x00FF
	 orr	r1, #0x4605&0xFF00
	 adr	r2, palstruc
	 bl	ioctl

@ Set up our palette array
	 ldr	r1, palstruc+16		@ blue
	 ldr	r2, ppalette
	 bl	setpal
	 ldr	r1, palstruc+12		@ green
	 ldr	r2, ppalette
	 add	r2, #1
	 bl	setpal
	 ldr	r1, palstruc+8		@ red
	 ldr	r2, ppalette
	 add	r2, #2
	 bl	setpal

	 mov	r0, #0x07		@ also set color to what's expected
	 str	r0, [r8, #attrib-dseg]	@ white on black
not8:
	sub	sp, fp, #4
	ldmfd	sp!, {fp, pc}		@ restore registers and return


@ Routine to set our palette array with a primary color (RGB)
setpal:	mov	r3, #256
	add	r1, #1			@ skip leading zero
spal10:	ldrb	r0, [r1]		@ get color
	add	r1, #2			@ next color
	strb	r0, [r2]
	add	r2, #4
	subs	r3, #1
	bne	spal10
	mov	pc, lr			@ return

	.align	2
.L10:	.word	.LC0		@  0
	.word	vinfo		@  4
	.word	0x4601		@  8
	.word	.LC1		@ 12
	.word	0x4602		@ 16
	.word	finfo		@ 20
	.word	.LC2		@ 24
	.word	fbp		@ 28
	.word	.LC3		@ 32

	.align	2
.LC0:	.ascii	"/dev/fb0\000"
	.align	2
.LC1:	.ascii	"Error setting variable information.\000"
	.align	2
.LC2:	.ascii	"Error reading fixed information.\000"
	.align	2
.LC3:	.ascii	"Failed to mmap.\000"

	.align	2
palstruc:.word	0			@ 0: first palette register to set
	.word	256			@ 4: number to set
	.word	palred			@ 8: red values
	.word	palgrn			@ 12: green values
	.word	palblu			@ 16: blue values
	.word	0			@ 20: transparency, can be null

@ Standard VGA colors
palred:	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0xA8,0,0xA8,0,0xA8,0,0xA8
	.byte	0,0x54,0,0x54,0,0x54,0,0x54,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0x00,0,0x14,0,0x20,0,0x2C,0,0x38,0,0x44,0,0x50,0,0x60
	.byte	0,0x70,0,0x80,0,0x90,0,0xA0,0,0xB4,0,0xC8,0,0xE0,0,0xFC
	.byte	0,0x00,0,0x40,0,0x7C,0,0xBC,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xBC,0,0x7C,0,0x40
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x7C,0,0x9C,0,0xBC,0,0xDC,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xDC,0,0xBC,0,0x9C
	.byte	0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C
	.byte	0,0xB4,0,0xC4,0,0xD8,0,0xE8,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xE8,0,0xD8,0,0xC4
	.byte	0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4
	.byte	0,0x00,0,0x1C,0,0x38,0,0x54,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x54,0,0x38,0,0x1C
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x38,0,0x44,0,0x54,0,0x60,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x60,0,0x54,0,0x44
	.byte	0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38
	.byte	0,0x50,0,0x58,0,0x60,0,0x68,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x68,0,0x60,0,0x58
	.byte	0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50
	.byte	0,0x00,0,0x10,0,0x20,0,0x30,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x30,0,0x20,0,0x10
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x20,0,0x28,0,0x30,0,0x38,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x38,0,0x30,0,0x28
	.byte	0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20
	.byte	0,0x2C,0,0x30,0,0x34,0,0x3C,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x3C,0,0x34,0,0x30
	.byte	0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00

palgrn:	.byte	0,0x00,0,0x00,0,0xA8,0,0xA8,0,0x00,0,0x00,0,0x54,0,0xA8
	.byte	0,0x54,0,0x54,0,0xFC,0,0xFC,0,0x54,0,0x54,0,0xFC,0,0xFC
	.byte	0,0x00,0,0x14,0,0x20,0,0x2C,0,0x38,0,0x44,0,0x50,0,0x60
	.byte	0,0x70,0,0x80,0,0x90,0,0xA0,0,0xB4,0,0xC8,0,0xE0,0,0xFC
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x00,0,0x40,0,0x7C,0,0xBC,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xBC,0,0x7C,0,0x40
	.byte	0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C
	.byte	0,0x7C,0,0x9C,0,0xBC,0,0xDC,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xDC,0,0xBC,0,0x9C
	.byte	0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4
	.byte	0,0xB4,0,0xC4,0,0xD8,0,0xE8,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xE8,0,0xD8,0,0xC4
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x00,0,0x1C,0,0x38,0,0x54,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x54,0,0x38,0,0x1C
	.byte	0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38
	.byte	0,0x38,0,0x44,0,0x54,0,0x60,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x60,0,0x54,0,0x44
	.byte	0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50
	.byte	0,0x50,0,0x58,0,0x60,0,0x68,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x68,0,0x60,0,0x58
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x00,0,0x10,0,0x20,0,0x30,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x30,0,0x20,0,0x10
	.byte	0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20
	.byte	0,0x20,0,0x28,0,0x30,0,0x38,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x38,0,0x30,0,0x28
	.byte	0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C
	.byte	0,0x2C,0,0x30,0,0x34,0,0x3C,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x3C,0,0x34,0,0x30
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00

palblu:	.byte	0,0x00,0,0xA8,0,0x00,0,0xA8,0,0x00,0,0xA8,0,0x00,0,0xA8
	.byte	0,0x54,0,0xFC,0,0x54,0,0xFC,0,0x54,0,0xFC,0,0x54,0,0xFC
	.byte	0,0x00,0,0x14,0,0x20,0,0x2C,0,0x38,0,0x44,0,0x50,0,0x60
	.byte	0,0x70,0,0x80,0,0x90,0,0xA0,0,0xB4,0,0xC8,0,0xE0,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xBC,0,0x7C,0,0x40
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x00,0,0x40,0,0x7C,0,0xBC,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xDC,0,0xBC,0,0x9C
	.byte	0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C,0,0x7C
	.byte	0,0x7C,0,0x9C,0,0xBC,0,0xDC,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xFC,0,0xE8,0,0xD8,0,0xC4
	.byte	0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4,0,0xB4
	.byte	0,0xB4,0,0xC4,0,0xD8,0,0xE8,0,0xFC,0,0xFC,0,0xFC,0,0xFC
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x54,0,0x38,0,0x1C
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x00,0,0x1C,0,0x38,0,0x54,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x60,0,0x54,0,0x44
	.byte	0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38,0,0x38
	.byte	0,0x38,0,0x44,0,0x54,0,0x60,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x70,0,0x70,0,0x70,0,0x70,0,0x70,0,0x68,0,0x60,0,0x58
	.byte	0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50,0,0x50
	.byte	0,0x50,0,0x58,0,0x60,0,0x68,0,0x70,0,0x70,0,0x70,0,0x70
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x30,0,0x20,0,0x10
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00
	.byte	0,0x00,0,0x10,0,0x20,0,0x30,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x38,0,0x30,0,0x28
	.byte	0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20,0,0x20
	.byte	0,0x20,0,0x28,0,0x30,0,0x38,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x40,0,0x40,0,0x40,0,0x40,0,0x40,0,0x3C,0,0x34,0,0x30
	.byte	0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C,0,0x2C
	.byte	0,0x2C,0,0x30,0,0x34,0,0x3C,0,0x40,0,0x40,0,0x40,0,0x40
	.byte	0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00,0,0x00

	.bss
	.align	2
	.type	fbp, %object
	.size	fbp, 4
fbp:	.space	4
	.comm	vinfo,160,4
	.comm	finfo,68,4

	.align	2
palette:.skip	256*4			@ copy of palette register values

	.text
	.align	2

@-------------------------------------------------------------------------------
@ Intrinsic to open (initialize) any mouse (or mice) connected to the computer.
@  Its position is set to center of screen, but its pointer is left hidden.
@  Returns 'false' if unsuccessful (not implemented).
@  BOOLEAN:= OPENMOUSE
@
intr85:	push	{lr}
	adr	r0, micemsg		@ fd = open("/dev/input/mice",
	mov	r1, #0x800		@	ORDONLY | O_NONBLOCK)
	bl	open
	str	r0, [r8, #micefd-dseg]
	cmp	r0, #0
	mov	r0, #0			@ assume 'false'
	bmi	om90			@ skip if unsuccessful

	str	r0, [r8, #mickx-dseg]	@ initialize for GETMOUSEMOVE
	str	r0, [r8, #micky-dseg]
	str	r0, [r8, #mickx0-dseg]
	str	r0, [r8, #micky0-dseg]
	str	r0, [r8, #micebut-dseg]	@ r0:= 0; all buttons released
	str	r0, [r8, #mouseptron-dseg] @ mouseptron:= false

	ldr	r0, [r8, #scrwidth-dseg]@ initialize to center of screen
	lsr	r0, #1
	str	r0, [r8, #micex-dseg]
	str	r0, [r8, #mousex-dseg]

	ldr	r0, [r8, #scrheight-dseg]
	lsr	r0, #1
	str	r0, [r8, #micey-dseg]
	str	r0, [r8, #mousey-dseg]
	mov	r0, #-1			@ return 'true'
om90:	pop	{pc}

micemsg:.asciz	"/dev/input/mice"
	.align	2

@-------------------------------------------------------------------------------
@ Get mouse state.
@  ADDR:= GETMOUSE
@  ADDR points to an integer array with: X=ADDR(0), Y=ADDR(1), BUTTONS=ADDR(2).
@  X,Y = Mouse position in pixels.
@  BUTTONS bits: 0 = left down, 1 = right down, 2 = middle down.
@  e.g: $03 = both left and right buttons down.
@
intr86:	push	{lr}
	bl	readmice

	ldr	r0, [r8, #micex-dseg]	@ x:= x + dx
	ldr	r1, [r8, #micedx-dseg]
	adds	r0, r1			@ if x<0 then x:= 0
	movlt	r0, #0			@ restrict to screen area
	ldr	r1, [r8, #scrwidth-dseg]@ if x>width then x:= width
	cmp	r0, r1
	movgt	r0, r1
	str	r0, [r8, #micex-dseg]

	ldr	r0, [r8, #micey-dseg]	@ y:= y + dy
	ldr	r1, [r8, #micedy-dseg]
	adds	r0, r1			@ if y<0 then y:= 0
	movlt	r0, #0			@ restrict to screen area
	ldr	r1, [r8, #scrheight-dseg] @ if y>height then y:= height
	cmp	r0, r1
	movgt	r0, r1
	str	r0, [r8, #micey-dseg]

	ldr	r0, [r8, #scrheight-dseg] @ return height - micey
	ldr	r1, [r8, #micey-dseg]
	sub	r0, r1
	str	r0, [r8, #micey_-dseg]

	ldr	r0, pmicex		@ return address of our array
gm90:	pop	{pc}

pmicex:	.word	micex			@ point to micex array

	.data
	.align	2
@ The following three items must remain in order: micex, micey_, micebut.
micex:	.word	1280/2			@ mouse coordinates (pixels)
micey_:	.word	720 - (720/2)		@ height - micey
micebut:.word	0			@ 0=left, 1=right, 2=middle down
micey:	.word	720/2
	.text

@-------------------------------------------------------------------------------
@ Get the amount the mouse moved since the previous call to this intrinsic.
@  ADDR:= GETMOUSEMOVE
@  ADDR points to an integer array with: DX=ADDR(0), DY=ADDR(1).
@
intr87:	push	{lr}
	bl	readmice

	ldr	r0, [r8, #mickx-dseg]
	ldr	r1, [r8, #mickx0-dseg]
	str	r0, [r8, #mickx0-dseg]
	sub	r0, r1
	str	r0, [r8, #mickdx-dseg]

	ldr	r0, [r8, #micky-dseg]
	ldr	r1, [r8, #micky0-dseg]
	str	r0, [r8, #micky0-dseg]
	sub	r0, r1
	str	r0, [r8, #mickdy-dseg]

	ldr	r0, pmickdx		@ return address of our array
	pop	{pc}

pmickdx:.word	mickdx			@ point to mickdx array

	.data
	.align	2
@ The following two items must remain in order: mickdx, mickdy.
mickdx:	.word	0			@ mouse movement (mickeys)
mickdy:	.word	0
	.text

@-------------------------------------------------------------------------------
@ Low-level read mice (mouse) routine
@ Outputs: micebut, micedx, micedy: amount mouse moved since last call
@ Outputs: mickx, micky: accumulated position for GETMOUSEMOVE
@
readmice:
	push	{lr}
	ldr	r0, [r8, #micefd-dseg]	@ r0:= read(fd, @mem, 3)
	ldr	r1, [r8, #micemem-dseg]
	mov	r2, #3
	bl	read			@ (returns either 3 or -1)
	pop	{lr}
	cmp	r0, #3			@ if no new data then go return the
	blt	mice30			@  previous values

	ldrb	r0, [r8, #miceb0-dseg]	@ else update values with new data
	mov	r2, r0			@ save copy of button data byte
	and	r0, #0x07		@ return only the state of 3 buttons
	str	r0, [r8, #micebut-dseg]

	ldrb	r0, [r8, #micedx0-dseg]
	tst	r2, #0x10		@ if miceb0 & $10 then dx:= dx - 256
	subne	r0, #0x100		@  (deal with sign bit)
	str	r0, [r8, #micedx-dseg]

	ldr	r1, [r8, #mickx-dseg]	@ mickx:= mickx + micedx
	add	r1, r0
	str	r1, [r8, #mickx-dseg]

	ldrb	r0, [r8, #micedy0-dseg]
	tst	r2, #0x20		@ if miceb0 & $20 then dy:= dy - 256
	subne	r0, #0x100		@  (deal with sign bit)
	str	r0, [r8, #micedy-dseg]

	ldr	r1, [r8, #micky-dseg]	@ micky:= micky + micedy
	add	r1, r0
	str	r1, [r8, #micky-dseg]
	mov	pc, lr			@ return
mice30:
	mov	r0, #0
	str	r0, [r8, #micedx-dseg]
	str	r0, [r8, #micedy-dseg]
	mov	pc, lr			@ return

	.data
	.align	2
micefd:	.word	0			@ mice file descriptor
micedx:	.word	0			@ amount mouse moved since last call to
micedy:	.word	0			@  readmice
mickx:	.word	0			@ position for GETMOUSEMOVE
micky:	.word	0
mickx0:	.word	0			@ old position
micky0:	.word	0

micemem:.word	miceb0			@ pointer to "read" struct
miceb0:	.byte	0			@ raw data from "read"
micedx0:.byte	0
micedy0:.byte	0
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to set a color register (0..255) for 8-bit (palettized) graphics.
@  The R,G,B values range 0..255, but only the high 6 bits are actually used.
@  An 8-bit graphic mode must have been already set up with either SetFB (84)
@  or SETVID (45).
@  SETPALETTE(REGISTER, RED, GREEN, BLUE)
@

 .if 0
@ sp+
@  4	p[0] = 4*9;		// size of message in bytes
@  8	p[1] = 0;		// process request
@ 12	p[2] = 0x4800B;	// set palette
@ 16	p[3] = 3*4;		// buffer size in bytes
@ 20	p[4] = 0;		// request size
@ 24	p[5] = Reg;		// offset to palette register
@ 28	p[6] = 1;		// number of colors to change
@ 32	p[7] = (B<<16) + (G<<8) + R;
@ 36	p[8] = 0;		// end tag
@
intr90:	mov	r3, r0			@ B
	pop	{r2}			@ G
	pop	{r1}			@ R
	pop	{r0}			@ Reg=r0, R=r1, G=r2, B=r3

	push	{lr}

	ldr	lr, ppalette		@ save RGB values in 'palette' array
	strb	r3, [lr, r0, lsl #2]	@ B
	add	lr, #1
	strb	r2, [lr, r0, lsl #2]	@ G
	add	lr, #1
	strb	r1, [lr, r0, lsl #2]	@ R

	sub	sp, sp, #44
	mov	r2, r2, asl #8		@ green
	mov	lr, #4*9
	add	r3, r2, r3, asl #16	@ + blue
	str	lr, [sp, #4]
	ldr	r2, [r8, #mbfd-dseg]	@ mailbox file descriptor (mbfd)
	ldr	lr, spal90		@ 0x4800B = set palette
	mov	ip, #0
	str	lr, [sp, #12]		@ array(2)
	mov	lr, #3*4
	add	r3, r3, r1		@ + red
	str	r0, [sp, #24]		@ Reg
	str	lr, [sp, #16]		@ 3*4
	ldr	r0, [r2, #0]		@ mbfd
	mov	lr, #1
	ldr	r1, spal94		@ 0xC004_64_00
	add	r2, sp, #4		@ address of p array
	str	ip, [sp, #8]		@ 0
	str	ip, [sp, #20]		@ 0
	str	lr, [sp, #28]		@ 1
	str	r3, [sp, #32]		@ G B R
	str	ip, [sp, #36]		@ 0
	bl	ioctl			@ ioctl(mbfd, _IOWR(100, 0, char *), p);
	add	sp, sp, #44
	pop	{pc}

spal90:	.word	0x4800B			@ set palette register (0..255)
spal94:	.word	0xC0046400		@ _IOWR(100, 0, char *)
 .endif


@ Because of a pecularity of the kernel code Register 15 or
@  255 must be set to actually make any change. Unfortunately this code is slow
@  because it waits for vertical blank and has the unbelievable side effect of
@  blanking the screen in Raspbian versions prior to Sep 9, 2014. (Accessing the
@  mailbox directly has its own set of bizarre problems.)

intr90:	strb	r0, [r8, #bluval+1-dseg]@ copy arguments into pal structure
	pop	{r0}
	strb	r0, [r8, #grnval+1-dseg]
	pop	{r0}
	strb	r0, [r8, #redval+1-dseg]
	pop	{r0}
	and	r0, #0xFF		@ mask index, for safety
	str	r0, [r8, #regstruc-dseg]

	push	{lr}

	ldr	lr, ppalette		@ save RGB values in 'palette' array
	ldrb	r1, [r8, #bluval+1-dseg]
	strb	r1, [lr, r0, lsl #2]	@ B
	add	lr, #1
	ldrb	r1, [r8, #grnval+1-dseg]
	strb	r1, [lr, r0, lsl #2]	@ G
	add	lr, #1
	ldrb	r1, [r8, #redval+1-dseg]
	strb	r1, [lr, r0, lsl #2]	@ R

	ldr	r0, [r8, #palfbfd-dseg]	@ ioctl(fbfd, FBIOPUTCMAP, &regstruc)
	mov	r1, #0x4605&0x00FF
	orr	r1, #0x4605&0xFF00
	ldr	r2, regptr
	bl	ioctl
	pop	{pc}

regptr:	.word	regstruc		@ pointer to (address of) pal structure

	.data
	.align	2
palfbfd:.word	-1			@ palette frame buffer file descriptor
regstruc:.word	0			@ color register number to set (0..255)
	.word	1			@ number of color registers to set
	.word	redval
	.word	grnval
	.word	bluval
	.word	0			@ transparency (not used)

redval:	.byte	0
	.byte	0
grnval:	.byte	0
	.byte	0
bluval:	.byte	0
	.byte	0
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to turn the flashing cursor off and on.
@  SHOWCURSOR(false)	\turn cursor off
@
intr88:	push	{lr}
	push	{r0}
	adr	r0, scmsg
	bl	text
	pop	{r1}
	cmp	r1, #0			@ is flag false?
	moveq	r0, #'l'		@ turn cursor off
	movne	r0, #'h'		@ else turn it on
	bl	putcharnow
	pop	{pc}

scmsg:	.byte	ESC			@ escape
	.asciz	"[?25"
	.align	2

@-------------------------------------------------------------------------------
@ Get character from keyboard (wait if necessary). There is no echo, no abort
@  on Ctrl+C, and no flashing cursor. Function keys, arrow keys, etc. return the
@  negative value of their (IBM-PC) scan codes.
@  CHARACTER:= GETKEY
@
intr89:	push	{r4, lr}
	bl	readstd			@ Key:= STDIN
	cmp	r0, #0x7F		@ if keystroke is a Delete then
	moveq	r0, #0x08		@  change it back to the Backspace
	cmp	r0, #ESC		@ if Key # Esc then return Key
	bne	gk99
	 mov	r4, r0			@ r4:= Esc
	 bl	intr33			@ if KeyHit then
	 cmp	r0, #0			@   (branch if not)
	 beq	gk90
	  bl	getscan			@ else get scan code and
	  rsb	r0, #0			@ return negative value
	  b	gk99
gk90:	 mov	r0, r4
gk99:
	cmp	r0, #-0x45		@ if Pause key then
	bne	gk100			@ branch if not
	 bl	intr89			@  wait for another key
	 mov	r0, #0			@  ignore this key by returning 0
gk100:
	pop	{r4, pc}

@-------------------------------------------------------------------------------
@ Intrinsic to return the state of the Shift, Ctrl, and Alt keys.
@  BITS:= GETSHIFTKEYS \BITS set if key is down: 0=Shift, 2=Ctrl, 3=Alt. Also
@  5=NumLock. Left and right Shift keys return the same BITS, i.e. they're not
@  distinguished. The following keys also return BITS set:
@	$0800 Up Arrow		$8000 I		$8_0000 W
@	$0400 Left Arrow	$4000 J		$4_0000 A
@	$0200 Down Arrow	$2000 K		$2_0000 S
@	$0100 Right Arrow	$1000 L		$1_0000 D
@	$8000_0000 Esc
@
intr93:	push	{lr}
	mov	r0, #0			@ ioctl(0, TIOCLINUX, &shift_state)
	mov	r1, #0x541C&0x00FF
	orr	r1, #0x541C&0xFF00
	mov	r2, #6			@ point to subcode 6
	push	{r2}
	mov	r2, sp
	bl	ioctl
					@ Shift, Ctrl and Alt keys are in TOS
	mov	r0, #0			@ ioctl(0, KDGKBLED, &kb_flags)
	mov	r1, #0x4B64&0x00FF
	orr	r1, #0x4B64&0xFF00
	push	{r0}
	mov	r2, sp
	bl	ioctl
	pop	{r1}			@ r1 bit 1 = NumLock state
	pop	{r0}			@ get Shift, Ctrl and Alt state bits
	orr	r0, r1, lsl #4		@ r0 bit 5 = NumLock state
	and	r0, #0xFF		@ make sure high bits are clear

	push	{r0}
	ldr	r0, [r8, #kdfd-dseg]	@ ioctl(fd, EVIOCGKEY(sizeof keys),
	ldr	r1, kd15		@       &keys);
	ldr	r2, pkeys
	bl	ioctl
	pop	{r0}

@	3322 2222 2222 1111  1111 1100 0000 0000
@	1098 7654 3210 9876  5432 1098 7654 3210
@	$0800 Up Arrow		$8000 I		$8_0000 W
@	$0400 Left Arrow	$4000 J		$4_0000 A
@	$0200 Down Arrow	$2000 K		$2_0000 S
@	$0100 Right Arrow	$1000 L		$1_0000 D

	ldr	r1, [r8, #keys+3*4-dseg]
	mov	r2, r1			@ up arrow ($67)
	and	r2, #1<<7
	orr	r0, r2, lsl #11-7
	mov	r2, r1			@ left arrow ($69)
	and	r2, #1<<9
	orr	r0, r2, lsl #10-9
	mov	r2, r1			@ down arrow ($6C)
	and	r2, #1<<0xC
	orr	r0, r2, lsr #0xC-9
	mov	r2, r1			@ right arrow ($6A)
	and	r2, #1<<0xA
	orr	r0, r2, lsr #0xA-8

	ldr	r1, [r8, #keys-dseg]
	mov	r2, r1			@ W ($11)
	and	r2, #1<<0x11
	orr	r0, r2, lsl #19-0x11
	mov	r2, r1			@ I ($17)
	and	r2, #1<<0x17
	orr	r0, r2, lsr #0x17-15
	mov	r2, r1			@ A ($1E)
	and	r2, #1<<0x1E
	orr	r0, r2, lsr #0x1E-18
	mov	r2, r1			@ S ($1F)
	and	r2, #1<<0x1F
	orr	r0, r2, lsr #0x1F-17
	mov	r2, r1			@ Esc ($01)
	and	r2, #1<<1
	orr	r0, r2, lsl #31-1

	ldr	r1, [r8, #keys+1*4-dseg]
	mov	r2, r1			@ D ($20)
	and	r2, #1<<0
	orr	r0, r2, lsl #16-0
	mov	r2, r1			@ J ($24)
	and	r2, #1<<4
	orr	r0, r2, lsl #14-4
	mov	r2, r1			@ K ($25)
	and	r2, #1<<5
	orr	r0, r2, lsl #13-5
	mov	r2, r1			@ L ($26)
	and	r2, #1<<6
	orr	r0, r2, lsl #12-6
	pop	{pc}

kd15:	.word	0x80104518
pkeys:	.word	keys
	.align	2

	.data
	.align	2
kdfd:	.word	0			@ key down file descriptor
keys:	.word	0, 0, 0, 0		@ 4 words = 128 bits; Esc = bit 1, etc.
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to delay the number of microseconds in r0.
@  DELAYUS(54945)	\delay 1/18.2 seconds
@
intr94:	push	{r4, r5, lr}
	mov	r5, r0			@ save duration in r5
	bl	intr82			@ r4:= GetTime
	mov	r4, r0
dly10:	bl	intr82			@ repeat until GetTime-r4 >= r5
	sub	r0, r4
	cmp	r0, r5
	blt	dly10
	pop	{r4, r5, pc}

@-------------------------------------------------------------------------------
@ Intrinsic to return the current date and time.
@  BYTE_ARRAY:= GETDATETIME
@   0: year since 1900
@   1: month (1..12)
@   2: day (1..31)
@   3: hour (0..23)
@   4: minute (0..59)
@   5: second (0..59)
@   6: hundreths (0..99)
@   7: week day 0=Sun, 1=Mon...6=Sat
@
intr95:	push	{lr}

	ldr	r0, ptime_t		@ 
	mov	r1, #0			@ error = gettimeofday(&struct, NULL)
	bl	gettimeofday

	push	{r4}			@ for some insane reason the call to
	mov	r4, sp			@  localtime must have sp on a double-
	bic	sp, #7			@  word boundary!
	ldr	r0, ptime_t
	bl	localtime		@ r0 = localtime(&struct)
	mov	sp, r4
	pop	{r4}

	ldr	r1, [r0, #5*4]		@ get (integer) year since 1900
	strb	r1, [r8, #dtime-dseg+0]	@ save in our array as a byte
	ldr	r1, [r0, #4*4]		@ get month
	add	r1, #1			@ make range 1..12
	strb	r1, [r8, #dtime-dseg+1]
	ldr	r1, [r0, #3*4]		@ get day
	strb	r1, [r8, #dtime-dseg+2]
	ldr	r1, [r0, #2*4]		@ get hour
	strb	r1, [r8, #dtime-dseg+3]
	ldr	r1, [r0, #1*4]		@ get minute
	strb	r1, [r8, #dtime-dseg+4]
	ldr	r1, [r0, #0*4]		@ get second
	strb	r1, [r8, #dtime-dseg+5]
	ldr	r1, [r0, #6*4]		@ get day of week
	strb	r1, [r8, #dtime-dseg+7]

	ldr	r0, ptime_t		@ get additional microseconds
	ldr	r0, [r0, #4]
	mov	r1, #10000&0x00FF	@ convert microseconds to hundreths
	orr	r1, #10000&0xFF00	@  by dividing by 10000
	bl	udiv			@ r0 = r0/r1
	strb	r0, [r8, #dtime-dseg+6]

	ldr	r0, pdtime		@ return address of our byte array
	pop	{pc}

ptime_t:.word	time_t			@ point to time_t array
pdtime:	.word	dtime			@ point to dtime array

	.data
	.align	2
time_t:	.word	0, 0			@ seconds since epoch, additional usec
@		0  1  2  3  4  5  6  7
@		Y  M  D  H  M  S  h  wd
dtime:	.byte	0, 0, 0, 0, 0, 0, 0, 0
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to insert a character into the keyboard's input queue. The next key
@  read from ChIn(1) will be this character. Several characters can be inserted.
@  INSERTKEY(CHARACTER)
@
intr96:	push	{lr}
	push	{r0}
	mov	r0, #0			@ ioctl(0, TIOCSTI, &char)
	mov	r1, #0x5412&0x00FF
	orr	r1, #0x5412&0xFF00
	mov	r2, sp			@ point to char on stack
	bl	ioctl
	pop	{r0}			@ balance stack
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to return the address of an array containing the currently displayed
@  frame buffer with its dimensions (in pixels) and starting address.
@  ADDR:= GETFB
@  ADDR points to an integer array with:
@   WIDTH=ADDR(0), HEIGHT=ADDR(1), DEPTH=ADDR(2), FRAMEBUFFER=ADDR(3).
@
intr97:	ldr	r0, pframebuf
	mov	pc, lr

pframebuf:.word	scrwidth

@-------------------------------------------------------------------------------
@ Intrinsic to wait for vertical sync signal.
@  WAITFORVSYNC
@
intr98:	push	{lr}
	ldr	r0, [r8, #fbfd-dseg]	@ ioctl(fbfd, FBIO_WAITFORVSYNC, 0);
	ldr	r1, wait4vs
	mov	r2, #0
	bl	ioctl
	pop	{pc}

wait4vs:.word	0x40044620		@ FBIO_WAITFORVSYNC

@-------------------------------------------------------------------------------
@ Intrinsic set the currently displayed video page.
@  SHOWPAGE(PAGE)
@
intr99:	push	{lr}
	and	r0, #1			@ restrict pages to 0 and 1
	ldr	r1, [r8, #yres-dseg]	@ yoff = page * yres
	mul	r0, r1
	str	r0, [r8, #yoff-dseg]

	ldr	r0, [r8, #fbfd-dseg]	@ ioctl(fbfd, FBIOPAN_DISPLAY, &vinfo);
	mov	r1, #0x4606&0x00FF
	orr	r1, #0x4606&0xFF00
	ldr	r2, pvinfo
	bl	ioctl
	pop	{pc}

pvinfo:	.word	xres			@ pointer to vinfo struct

	.data
	.align	2
@ structure for frame buffer driver variable information
xres:	.word	1280		@  0 horizontal visible resolution
yres:	.word	720		@  4 vertical visible resolution
xvres:	.word	1280		@  8 horizontal virtual resolution
yvres:	.word	720		@ 12 vertical virtual resolution
xoff:	.word	0		@ 16 offset from virtual area to visible area
yoff:	.word	0		@ 20
pixbits:.word	16		@ 24 bits per pixel
grays:	.word	0		@ 28 if not 0 then use gray levels, not colors
@ . . .
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to quickly copy an array of bytes. This is logically equivalent to:
@  for I:= 0 to SIZE-1 do DST(I):= SRC(I); \ but is many times faster, and it
@  works for overlapping memory areas.
@  COPYMEM(DST, SRC, SIZE)
@   regs:  r1   r0   r2
@
intr100:mov	r2, r0			@ get the number of bytes to copy
	pop	{r0}			@ get source address
	pop	{r1}			@ get destination address

	cmp	r1, r0			@ skip if dest addr is > source addr
	bhi	cm100
	beq	cm90			@ exit if addresses are the same

cm10:	tst	r0, #0xF		@ copy array up to 1st 16-byte boundary
	beq	cm20			@  of source's address
	cmp	r2, #0			@ if remaining Size = 0 then all done
	beq	cm90
	ldrb	r3, [r0]		@ get byte at source address
	add	r0, #1			@ bump pointer
	strb	r3, [r1]		@ store byte into destination address
	add	r1, #1			@ bump pointer
	sub	r2, #1			@ decrement Size counter
	b	cm10			@ loop
cm20:
	lsrs	r3, r2, #4		@ r3 = Size/16 = number of 16-byte
	beq	cm30			@  chunks to copy; skip if less than 16
	push	{r4-r7}			@ save registers
cm25:	ldmia	r0!, {r4-r7}		@ copy 16 bytes at a time
	subs	r3, #1			@ count down number of 16-byte chunks
	stmia	r1!, {r4-r7}
	bne	cm25			@ tight loop only takes 16 cycles
	pop	{r4-r7}			@ restore registers
cm30:
	ands	r2, #0xF		@ copy any remaining amount that's less
	beq	cm90			@  than 16 bytes
cm35:	ldrb	r3, [r0]
	add	r0, #1
	strb	r3, [r1]
	add	r1, #1
	subs	r2, #1
	bne	cm35			@ loop
cm90:	mov	pc, lr			@ return

@ Handle case where destination is > source and areas might be overlapping
@ for I:= Size-1 downto 0 do Dst(I):= Src(I)
cm100:
	add	r0, r2			@ point to end of source area +1
	add	r1, r2			@ point to end of destination area +1

cm110:	tst	r0, #0xF		@ copy array up to 1st 16-byte boundary
	beq	cm120			@  of source's address
	cmp	r2, #0			@ if remaining Size = 0 then all done
	beq	cm190
	sub	r0, #1			@ bump pointer
	ldrb	r3, [r0]		@ get byte at source address
	sub	r1, #1			@ bump pointer
	strb	r3, [r1]		@ store byte into destination address
	sub	r2, #1			@ decrement Size counter
	b	cm110			@ loop
cm120:
	lsrs	r3, r2, #4		@ r3 = Size/16 = number of 16-byte
	beq	cm130			@  chunks to copy; skip if less than 16
	push	{r4-r7}			@ save registers
cm125:	ldmdb	r0!, {r4-r7}		@ copy 16 bytes at a time
	subs	r3, #1			@ count down number of 16-byte chunks
	stmdb	r1!, {r4-r7}
	bne	cm125			@ tight loop only takes 16 cycles
	pop	{r4-r7}			@ restore registers
cm130:
	ands	r2, #0xF		@ copy any remaining amount that's less
	beq	cm190			@  than 16 bytes
cm135:	sub	r0, #1
	ldrb	r3, [r0]
	sub	r1, #1
	strb	r3, [r1]
	subs	r2, #1
	bne	cm135			@ loop
cm190:	mov	pc, lr			@ return

@-------------------------------------------------------------------------------
@ Intrinsic to quickly fill an array of bytes. This is logically equivalent to:
@  for I:= 0 to SIZE-1 do ADDRESS(I):= Value; \ but is many times faster.
@  FILLMEM(ADDRESS, VALUE, SIZE)
@   regs:  r2       r0     r1
@
intr101:mov	r1, r0			@ get the number of bytes to fill
	pop	{r0}			@ get value to set the bytes to
	pop	{r2}			@ get starting address of the array

fm10:	tst	r2, #0xF		@ fill array up to 1st 16-byte boundary
	beq	fm20
	cmp	r1, #0			@ if remaining Size = 0 then all done
	beq	fm90
	strb	r0, [r2]
	add	r2, #1
	sub	r1, #1
	b	fm10			@ loop
fm20:
	lsrs	r3, r1, #4		@ r3 = Size/16 = number of 16-byte
	beq	fm30			@  chunks to store; skip if none
	push	{r4-r6}			@ save registers
	and	r0, #0xFF		@ copy Value into 3 high bytes of r0
	orr	r0, r0, lsl #8
	orr	r0, r0, lsl #16
	mov	r4, r0			@ copy Value into 3 more registers
	mov	r5, r0			@  giving a total of 16 bytes of Value
	mov	r6, r0
fm25:	stmia	r2!, {r0, r4-r6}	@ store 16 bytes and bump up r2
	subs	r3, #1			@ count down number of 16-byte chunks
	bne	fm25			@ tight loop only takes 8 cycles
	pop	{r4-r6}			@ (unrolled loop goes no faster)
fm30:
	ands	r1, #0xF		@ fill any remaining amount that's less
	beq	fm90			@  than 16 bytes
fm35:	strb	r0, [r2]
	add	r2, #1
	subs	r1, #1
	bne	fm35			@ loop
fm90:	mov	pc, lr			@ return

@-------------------------------------------------------------------------------
@ Intrinsic to disable or enable the video screen display. This unfortunately
@  also changes the palette registers and has some bizarre timing issues!
@  Raspbian Sept 8, 2014 version.
@  BLANKSCREEN(true)		\turns screen off, i.e. blanks it
@
intr102:push	{lr}
	cmp	r0, #0			@ false turns screen on
	movne	r0, #1			@ not false turns screen off (blanks it)
	mov	r2, r0
	ldr	r0, [r8, #fbfd-dseg]	@ ioctl(fbfd, FBIOBLANK, blank);
	mov	r1, #0x4611&0x00FF
	orr	r1, #0x4611&0xFF00
	bl	ioctl
	pop	{pc}

@-------------------------------------------------------------------------------
@ Intrinsic to start playing a sound file, which can be one of these types:
@   voc wav raw or au
@  PLAYSOUNDFILE("/path/filename.ext")
@
intr103:push	{r0, lr}		@ save pointer to sound filename
	bl	fork			@ start another process (thread/task)
	cmp	r0, #0			@ the parent process returns now
	popne	{r0, pc}

	pop	{r0}			@ get pointer to sound file
	sub	sp, #0x80		@ make some temporary space for
	mov	r1, sp			@  concatenating strings
	adr	r2, aplay
psf10:	ldrb	r3, [r2]		@ copy first string to temporary space
	add	r2, #1
	cmp	r3, #0
	beq	psf20
	strb	r3, [r1]
	add	r1, #1
	b	psf10

psf20:	ldrb	r3, [r0]		@ append filename
	add	r0, #1
	tst	r3, #0x80		@ handle MSB terminator
	beq	psf25
	and	r3, #0x7F
	strb	r3, [r1]
	add	r1, #1
	mov	r3, #0			@ append zero terminator
psf25:
	strb	r3, [r1]
	add	r1, #1
	cmp	r3, #0			@ loop until zero terminator
	bne	psf20

	sub	r1, #1			@ back up over 0 terminator
	adr	r2, aplay2		@ append second string
psf30:	ldrb	r3, [r2]
	add	r2, #1
	strb	r3, [r1]
	add	r1, #1
	cmp	r3, #0
	bne	psf30

	mov	r0, sp			@ point to "aplay..."
	bl	system			@ and execute it
@	mov	r1, #0
@	bl	execlp
	mov	r0, #0			@ indicate normal termination
@	b	_exit			@ (child shouldn't flush buffers, etc.)
	b	exit

aplay:	.asciz	"aplay -q "		@ play sound in "quiet mode"
aplay2:	.asciz	" >/dev/null 2>&1 "	@ discard any output or error messages
	.align	2

@-------------------------------------------------------------------------------
@ Intrinsic to set the number of digits that HexOut (27) outputs.
@  SETHEXDIGITS(DIGITS)
@
intr104:cmp	r0, #17			@ if (unsigned) digits > 16 then
	movhs	r0, #1			@  digits:= 1
	str	r0, [r8, #hexdigits-dseg]
	mov	pc, lr			@ return

@===============================================================================
@			 FLOATING POINT I/O ROUTINES
@===============================================================================
@
@ Input a floating-point number and return it in d0. d1 and d2 are destroyed.
@  Register usage:
@	 r0 = character input
@	 r1 = exponent
@	 r2:0 = flag: number is negative
@	 r2:1 = flag: exponent is negative
@	 r3 = used to multiply by 10
@
flin:	push	{lr}
	mov	r2, #0			@ assume positive number & exponent
	mov	r1, #0			@ initialize exponent to 0
	adr	r0, f0_0		@ initialize d0 accumulator to 0.0
	fldd	d0, [r0]

@ Loop to handle any leading characters
flin10:	bl	getdig			@ get a character
	beq	flin20			@ branch if it's a digit
	cmp	r0, #'-'		@ minus sign?
	bne	flin15			@ skip if not
	 eor	r2, #1			@ set negative number flag
	 b	flin10			@ loop--get next character
flin15:
	cmp	r0, #EOF		@ exit if end of file
	beq	flin99
	cmp	r0, #'.'		@ is it a decimal point?
	bne	flin10			@ loop if not
	b	flin30			@ else go get fractional part

@ Read the integer part (the part before the decimal point)
flin20:	bl	addin			@ add in the digit to d0
	bl	getdig			@ get next character
	beq	flin20			@ loop until not a digit
	cmp	r0, #'.'		@ decimal point?
	bne	flin35			@ branch if not--end of number

@ Read the fractional part (the part following the decimal point)
flin30:	bl	getdig			@ get a digit
	bne	flin35			@ exit loop if not a digit
	 bl	addin			@ add in the digit
	 sub	r1, #1			@ decremnent exponent
	 b	flin30			@ loop until not digit

@ Here on end of number, read in the exponent if provided
flin35:	cmp	r0, #'E'		@ if exponent
	beq	flin40			@  then handle it
	 cmp	r0, #'e'		@ also check lowercase
	 bne	flin55			@ branch if no exponent--go finish up
flin40:	bl	getdig			@ get a character
	beq	flin45			@ skip if it's a digit
	 cmp	r0, #EOF		@ exit if end of file
	 beq	flin99
	 cmp	r0, #'-'		@ negative?
	 bne	flin40			@ loop if not
	 eor	r2, #2			@ set flag for negative exponent
	 b	flin40			@ loop until digit

@ Here when we've got an exponent digit in al, convert digits to binary in r3
flin45:	mov	r3, #0			@ initialize exponent accumulator (r3)
flin47:	add	r3, r3, r3, lsl #2	@ multiply r3 by 10 the fast and
	add	r3, r3			@  tricky way 8-)
	add	r3, r0			@ add in the digit
	bl	getdig			@ get next character
	beq	flin47			@ loop until not a digit

	tst	r2, #2			@ negative exponent?
	beq	flin50			@ skip if not
	 rsb	r3, #0			@ make it negative
flin50:	add	r1, r3			@ combine with fractional part exponent

@ Now adjust number according to exponent
flin55:	tst	r1, r1			@ test exponent
	beq	flin90			@ if it's zero then exit
	bmi	flin65			@ negative?

@ Here if exponent is greater than zero
	adr	r0, f10_0		@ multiply d0 by ten "exponent" times
	fldd	d1, [r0]
flin60:	fmuld	d0, d0, d1
	subs	r1, #1
	bne	flin60			@ loop
	b	flin90

@ Here if exponent is negative
flin65:	adr	r0, f10_0		@ divide d0 by ten "exponent" times
	fldd	d1, [r0]
flin67:	fdivd	d0, d0, d1
	adds	r1, #1
	bne	flin67			@ loop

flin90:	tst	r2, #1			@ was number negative?
	beq	flin99			@ skip if not
	 fnegd	d0, d0			@ make it negative
flin99:	pop	{pc}

@-------------------------------------------------------------------------------
@ Get a character into r0. If it's a digit (0-9) then set the Z flag and
@  convert the digit from ASCII to binary in r0.
@
getdig:	push	{r1-r3, lr}		@ make sure chin doesn't change these
gdig10:	bl	chin			@ get a character
	cmp	r0, #'_'		@ ignore underlines
	beq	gdig10
	cmp	r0, #'0'		@ test for digit
	blo	gdig90			@ skip if not
	 cmp	r0, #'9'		@ test again
	 bhi	gdig90			@ skip if not
	 and	r0, #0x0F		@ convert ASCII digit to binary
	 cmp	r0, r0			@ set Z flag to show char was a digit
gdig90:	pop	{r1-r3, pc}

@-------------------------------------------------------------------------------
@ Multiply d0 by 10.0 then add to it the digit in r0. d1 and d2 are destroyed.
@
addin:	push	{r1, lr}
	adr	r1, f10_0		@ multiply d0 by 10.0
	fldd	d1, [r1]
	fmuld	d0, d0, d1
	fmsr	s4, r0			@ add (integer) digit to d0
	fsitod	d1, s4
	faddd	d0, d0, d1
	pop	{r1, pc}

f10_0:	.double	10.0			@ constants
f1000_0:.double	1000.0
f0_5:	.double	0.5
f1_0:	.double	1.0
f0_0:	.double	0.0
infmsg:	.asciz	"INF"			@ infinite
nanmsg:	.asciz	"NAN"			@ not a number
	.align	2

@===============================================================================
@ Output floating point number in d0.
@  Inputs: form1 and form2 from the Format intrinsic
@  Register usage:
@	 r0 = scratch
@	 r1 = exponent
@	 r2 = status flags for number and number of digits before decimal point
@	 r3 = maxium number of significant digits that can be displayed
@	 r4 = number of digit places before decimal point
@	 r5 = sign of number to output
@	 r6 = FPU coprocessor status of number to output
@
sigfig	=	15			@ number of decimal digits of precision

flout:	push	{r4-r7, lr}		@ save registers
	fcmpd	d0, d0			@ test for not a number (NAN)
	fmstat				@ get status
	bvc	flout20			@ V flag set if NAN
	 adr	r0, nanmsg		@ display "NAN"
	 bl	intr12b
	 b	flout90			@ exit
flout20:
@ test for infinite value
	fmrdh	r0, d0			@ get exponent
	mov	r2, r0			@ save sign bit
	mov	r1, #0x0FF00000		@ if exponent = 7FF (and not NAN)
	orr	r1, #0x70000000	
	and	r0, r1
	cmp	r0, r1
	bne	flout22
	 tst	r2, #0x80000000		@ have infinite value
	 mov	r0, #'-'		@ display '-' in negative infinity
	 blne	fchout
	 adr	r0, infmsg		@ display "INF"
	 bl	intr12b
	 b	flout90			@ exit
flout22:
	adr	r0, f0_0		@ test the sign of the number
	fldd	d1, [r0]		@ by comparing it to 0.0
	fcmpd	d0, d1
	fmstat				@ copy FPU status to CPSR
	mrs	r0, cpsr		@ get status flags
	mov	r2, r0			@ save them for later
	and	r0, #0x80000000		@ get sign flag (N)
	mov	r5, r0			@ save it
	fabsd	d0, d0			@ force number to be positive

	ldr	r0, [r8, #form1-dseg]	@ determine format type
	cmp	r0, #0
	beq	flout25			@ =0 = scientific notation
	 bmi	flout50			@ <0 = engineering notation
					@ >0 = standard decimal notation
	 mov	r2, r0			@ number of digits before decimal point
	 bl	fixpt			@ do standard decimal notation output
	 b	flout90			@ exit

@ Output number in scientific notation
flout25:mov	r1, #0			@ initialize exponent to 0
	mov	r0, r2			@ get saved status
	tst	r0, #0x40000000
	bne	flout40			@ skip if number is zero (Z flag set)

@ Multiply number by ten until it's not a fraction
flout30:adr	r0, f1_0		@ is number < 1.0 ?
	fldd	d1, [r0]
	fcmpd	d0, d1
	fmstat
	bpl	flout35			@ exit loop if >= 1.0

	 adr	r0, f10_0		@ multiply by 10.0
	 fldd	d1, [r0]
	 fmuld	d0, d0, d1
	 sub	r1, #1			@ decrement exponent
	 b	flout30			@ loop

@ Divide number by ten until it's less than ten
flout35:adr	r0, f10_0		@ is number >= 10.0 ?
	fldd	d1, [r0]
	fcmpd	d0, d1
	fmstat
	bmi	flout40			@ exit loop if not

	 adr	r0, f10_0		@ divide by 10.0
	 fldd	d1, [r0]
	 fdivd	d0, d0, d1
	 add	r1, #1			@ increment exponent
	 b	flout35			@ loop

flout40:mov	r2, #2			@ set 2 places before decimal point
	b	flout70			@ output number followed by its exponent

@ Output the number in engineering notation
flout50:mov	r1, #0			@ initialize exponent to 0
	mov	r0, r2			@ get status
	tst	r0, #0x40000000
	bne	flout65			@ skip if number is zero (Z flag set)

@ Multiply number by one thousand until it's not a fraction (i.e. not < 1.0)
flout55:adr	r0, f1_0		@ is number < 1.0 ?
	fldd	d1, [r0]
	fcmpd	d0, d1
	fmstat
	bpl	flout60			@ exit loop when number is >= 1.0

	 adr	r0, f1000_0		@ multiply by 1000.0
	 fldd	d1, [r0]
	 fmuld	d0, d0, d1
	 sub	r1, #3			@ decrement exponent
	 b	flout55			@ loop

@ Divide by one thousand until number is less than a thousand
flout60:adr	r0, f1000_0		@ is number >= 1000.0 ?
	fldd	d1, [r0]
	fcmpd	d0, d1
	fmstat
	bmi	flout65			@ exit loop if not

	 adr	r0, f1000_0		@ divide by 1000.0
	 fldd	d1, [r0]
	 fdivd	d0, d0, d1
	 add	r1, #3			@ increment exponent
	 b	flout60			@ loop

flout65:mov	r2, #4			@ set 4 places before decimal point

@ Output number followed by exponent created by above notation adjustments
flout70:bl	fixpt			@ output fixed point part (i.e: 123.45)
	mov	r0, #'E'		@ output the 'E'
	bl	fchout
	mov	r0, #'+'		@ get a plus sign
	tst	r1, r1			@ negative exponent?
	bge	flout75			@ skip if not
	 rsb	r1, #0			@ make exponent positive
	 mov	r0, #'-'		@ get a minus sign
flout75:bl	fchout			@ output the sign

@ Output leading zeros of exponent, if any (i.e: 123.45E+003)
	cmp	r1, #100		@ no leading zero if exponent >= 100
	bge	flout85
	 cmp	r1, #10			@ one zero if >= 10
	 bge	flout80
	  mov	r0, #'0'		@ do leading zeros
	  bl	fchout
flout80: mov	r0, #'0'
	 bl	fchout
flout85:mov	r0, r1			@ output the exponent
	bl	intout
flout90:
	pop	{r4-r7, pc}		@ restore registers and return

@-------------------------------------------------------------------------------
@ Routine to output the floating-point number in d0 in standard decimal
@  format (i.e. no exponent: 123.456). The number of digits before the decimal
@  point is in r2 and the number of digits after it is in "form2".
@
fixpt:	push	{lr}
	adr	r0, f1_0		@ test for fraction (i.e: < 1.0)
	fldd	d1, [r0]
	fcmpd	d0, d1
	fmstat				@ and save resulting status flags in r6
	mrs	r6, cpsr

	adr	r0, f0_0		@ compare number to 0.0
	fldd	d1, [r0]
	fcmpd	d1, d0
	fmstat				@ is it zero?
	bne	fixpt10			@ skip if not
	 mov	r4, #0			@ no integer digits
	 b	fixpt15			@ go output the zero
fixpt10:
	bl	divten			@ convert number to a fraction < 1.0
	bl	fround			@ round to number of displayed digits
fixpt15:
@ Calculate and output the number of leading blanks, and output the sign
	sub	r2, r4
	mov	r0, r6			@ get number's status
	tst	r0, #0x80000000		@ is number < 1.0 ? (is N flag set?)
	beq	fixpt20			@ skip if not
	 sub	r2, #1			@ else count the leading zero
fixpt20:tst	r5, #0x80000000		@ test the sign of the number
	beq	fixpt25			@ skip if it's positive
	 sub	r2, #1			@ else count the minus sign
fixpt25:tst	r2, r2			@ deal with zero or negative value
	ble	fixpt35			@ skip if no blanks
fixpt30:mov	r0, #' '		@ output r2 number of blanks
	bl	fchout
	subs	r2, #1			@ loop
	bne	fixpt30

@ Test and handle the sign
fixpt35:tst	r5, #0x80000000		@ test the sign
	beq	fixpt40			@ skip if positive
	 mov	r0, #'-'		@ output a minus sign
	 bl	fchout

@ Output leading zero if number is a fraction (i.e: < 1.0)
fixpt40:mov	r0, r6			@ get number status
	tst	r0, #0x80000000		@ is number < 1.0 ? (is N flag set?)
	beq	fixpt45			@ skip if not
	 mov	r0, #'0'		@ output leading zero
	 bl	fchout
fixpt45:
@ Output the digits of the number
	mov	r3, #sigfig		@ set maximum digits to output
	mov	r2, r4			@ set number of digits in integer part
	bl	digout			@ output integer part, before decimal pt
	ldr	r0, [r8, #form2-dseg]	@ any digits after decimal point?
	cmp	r0, #0
	ble	fixpt90			@ branch if not--exit

	mov	r0, #'.'		@ output the decimal point
	bl	fchout
	ldr	r2, [r8, #form2-dseg]	@ set number of digits after decimal pt.
	bl	digout			@ display fractional part
fixpt90:pop	{pc}

@-------------------------------------------------------------------------------
@ Round the number by adding 0.5 * 10 ^ -(di+form2)
@
fround:	ldr	r0, [r8, #form2-dseg]	@ calculate number of digits to output
	add	r0, r4
	cmp	r0, #sigfig		@ beyond the max. significant figures?
	bls	frnd10			@ skip if not
	 mov	r0, #sigfig		@ else limit to sigfigs
frnd10:	adr	r7, rndtbl		@ add rounding factor
	add	r7, r0, lsl #3		@ index doubles
	fldd	d1, [r7]
	faddd	d0, d0, d1

	adr	r0, f1_0		@ overflow? (d0 >= 1.0?)
	fldd	d1, [r0]
	fcmpd	d0, d1
	fmstat
	bpl	frnd20			@ skip if so
	 mov	pc, lr			@ return

@ Handle rounding overflow by adjusting the number back to a fraction < 1.0.
@  and eliminating any leading zero.
frnd20:	mov	r6, #0			@ clear flags to eliminate leading zero

	ldr	r0, [r8, #form1-dseg]	@ scientific notation?
	cmp	r0, #0
	bne	frnd30			@ skip if not
	 sub	r4, #1			@ one less digit before decimal point
	 add	r1, #1			@ add 1 to E exponent
	 b	divten2			@ go adjust back below 1.0
frnd30:
	bge	divten2			@ skip if not engineering notation
	cmp	r4, #3			@ 3 digits ahead of decimal point?
	bne	divten2			@ skip if not
	mov	r4, #0			@ reset digit count so we end up with 1
	add	r1, #3			@ add 3 to E exponent
	b	divten2			@ go adjust back below 1.0

@- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@ Routine to divide d0 by ten until it is a fraction (< 1.0)
@
divten:	mov	r4, #0			@ zero digit counter
divten1:adr	r0, f1_0		@ if d0 < 1.0?
	fldd	d1, [r0]
	fcmpd	d0, d1
	fmstat
	bmi	divten9			@ then exit
divten2:
	adr	r0, f10_0		@ else divide by 10.0
	fldd	d1, [r0]
	fdivd	d0, d0, d1
	add	r4, #1			@ count digits before decimal point
	b	divten1			@ loop
divten9:
	mov	pc, lr			@ return

	.align	3
rndtbl:	.double	5.0E-1			@ rounding constants
	.double	5.0E-2
	.double	5.0E-3
	.double	5.0E-4
	.double	5.0E-5
	.double	5.0E-6
	.double	5.0E-7
	.double	5.0E-8
	.double	5.0E-9
	.double	5.0E-10
	.double	5.0E-11
	.double	5.0E-12
	.double	5.0E-13
	.double	5.0E-14
	.double	5.0E-15
	.double	5.0E-16

@-------------------------------------------------------------------------------
@ Output a string of floating point digits
@  Number of digits to output is in r2, which is destroyed
@  If maximum precision is exceeded then output zeros
@
digout:	push	{lr}
	tst	r2, r2			@ exit if no digits to output
	beq	digo90

digo10:	tst	r3, r3			@ more significant digits?
	beq	digo30			@ branch if not--output a zero
	adr	r0, f10_0		@ multiply d0 by 10.0
	fldd	d1, [r0]
	fmuld	d0, d0, d1

	fcpyd	d1, d0
	adr	r0, f0_5		@ subtract .5 to compensate for rounding
	fldd	d2, [r0]
	fsubd	d1, d1, d2
	ftosid	s4, d1			@ convert to integer digit
	fmrs	r0, s4			@ move s4 to r0
	orr	r0, #'0'
	cmp	r0, #'9'
	ble	digo20
	 mov	r0, #'9'		@ avoid ":" due to rounding error
digo20:	bl	fchout			@ output digit

	fsitod	d1, s4			@ load and float the digit
	fsubd	d0, d0, d1		@ subrtract it from the original number
	sub	r3, #1			@ count down the significant digits
	b	digo40
digo30:
	 mov	r0, #'0'		@ output a zero
	 bl	fchout
digo40:
	subs	r2, #1			@ loop
	bne	digo10
digo90:	pop	{pc}

@-------------------------------------------------------------------------------
@ Output a character while preserving all registers.
@
fchout:	push	{r0-r3, lr}
	bl	chout
	pop	{r0-r3, pc}

@===============================================================================
@ Table of routines to handle fundamental I/O functions
@ dev number:	  0	  1	  2	  3	  4	  5	  6	  7	  8
openit:	.word	openi0,	openi1,	dumio,	openi3,	dumio,	dumio,	openi6,	dumio,	psopni
openot:	.word	dumio,	dumio,	dumio,	openo3,	dumio,	dumio,	openo6,	dumio,	psopno
chint:	.word	chin0,	chin1,	dumio,	chin3,	dumio,	dumio,	chin6,	dumio,	psin
choutt:	.word	conout,	conout,	dumio,	chout3,	dumio,	dumio,	chout6,	dumio,	psout
closet:	.word	dumio,	dumio,	dumio,	close3,	dumio,	dumio,	dumio,	dumio,	dumio

@-------------------------------------------------------------------------------
@
openi:	ldr	r1, [r8, #nowdev-dseg]	@ get the device number
	cmp	r1, #8			@ legal?
	bhi	ioerr
	adr	r2, openit
	ldr	r2, [r2, r1, lsl #2]	@ fetch table entry
	mov	pc, r2			@ branch to where it's pointing

@-------------------------------------------------------------------------------
@
openo:	ldr	r1, [r8, #nowdev-dseg]	@ get the device number
	and	r1, #0xFF		@ $106 is legal
	cmp	r1, #8			@ legal?
	bhi	ioerr
	adr	r2, openot
	ldr	r2, [r2, r1, lsl #2]	@ fetch table entry
	mov	pc, r2			@ branch to where it's pointing

@-------------------------------------------------------------------------------
@
chin:	ldr	r1, [r8, #backfl-dseg]
	cmp	r1, #0
	bne	chin95
	 ldr	r1, [r8, #nowdev-dseg]
	 cmp	r1, #8
	 bhi	ioerr
	 adr	r2, chint
	 ldr	r2, [r2, r1, lsl #2]	@ fetch table entry
	 mov	pc, r2			@ branch to where it's pointing
chin95:
	mov	r1, #0
	str	r1, [r8, #backfl-dseg]
	ldr	r0, [r8, #lastch-dseg]
	mov	pc, lr			@ return

	.data
	.align	2
lastch:	.word	0			@ previous character read in by chin
backfl:	.word	0			@ flag: re-read last character
scancode:.word	0			@ non-ASCII char's scan code
	.text

@-------------------------------------------------------------------------------
@
chout:	ldr	r1, [r8, #nowdev-dseg]	@ get the device number
@	and	r1, #0xFF		@ ($106 is legal)
	bic	r1, #0x300		@ ($106 and $206 are legal)
	cmp	r1, #8			@ legal?
	bhi	ioerr
	adr	r2, choutt
	ldr	r2, [r2, r1, lsl #2]	@ fetch table entry
	mov	pc, r2			@ branch to where it's pointing

@-------------------------------------------------------------------------------
@
close:	ldr	r1, [r8, #nowdev-dseg]	@ get the device number
	cmp	r1, #8			@ legal?
	bhi	ioerr
	adr	r2, closet
	ldr	r2, [r2, r1, lsl #2]	@ fetch table entry
	mov	pc, r2			@ branch to where it's pointing

@-------------------------------------------------------------------------------
@ Report I/O error
@
ioerr:	mov	r0, #3			@ report RUN-TIME ERROR 3: I/O
	b	error			@ branch and return

@-------------------------------------------------------------------------------
@ Device 7, the null device
@
dumio:	mov	r0, #EOF		@ return end-of-file character (if used)
	mov	pc, lr

@===============================================================================

	.data
	.align	2
kbinx:	.word	0			@ index into kbbuf
kbbufsiz: .word	0			@ number of characters stored in buffer
kbbufptr: .word	kbbuf			@
kbbuf:	.skip	128			@ keyboard line input buffer
	.text

@-------------------------------------------------------------------------------
@ Discard any characters in the keyboard buffer and any pending keyboard input.
@
openi0:	mov	r0, #0
	str	r0, [r8, #kbbufsiz-dseg]
	str	r0, [r8, #kbinx-dseg]
	b	openi1			@ go discard any pending keystrokes

@-------------------------------------------------------------------------------
@ Get character from keyboard buffer. If buffer is empty then fill it with a
@  line of characters typed on the keyboard. The line is terminated with the
@  Enter (carriage return) key.
@
chin0:	push	{lr}
	ldr	r2, [r8, #kbinx-dseg]	@ branch if there's something in buffer
	ldr	r3, [r8, #kbbufsiz-dseg]
	cmp	r2, r3
	blt	ch030

@ Fill buffer with keystrokes terminated with Enter key (or LF or EOF)
	ldr	r3, [r8, #kbbufptr-dseg]
	mov	r2, #0			@ point to start of buffer
ch010:	push	{r2, r3}
ch011:	bl	chin1			@ get keystroke
	cmp	r0, #0			@ ignore non-ASCII keystrokes, those
	bne	ch012			@  that are preceded by a zero byte
	bl	chin1			@  eat the scan code
	b	ch011			@  loop back for next char
ch012:	pop	{r2, r3}

	cmp	r0, #0x08		@ is key a Backspace?
	bne	ch020			@ skip if not
	 cmp	r2, #0			@ at start of buffer?
	 ble	ch010			@ ignore key if so
	 sub	r2, #1			@ back up one character in buffer
	 bl	putcharnow		@ back up on screen (r0 = backspace)
	 ldrb	r1, [r2, r3]		@ get character from line buffer
	 cmp	r1, #0x20		@ if it's a ctrl char then back up
	 bllt	putcharnow		@  on screen a second time
	 mov	r0, #0x20		@ blank out character with a space
	 bl	putcharnow
	 cmp	r1, #0x20		@ if char was a ctrl char then
	 bllt	putcharnow		@  blank out second character
	 mov	r0, #0x08		@ back up on screen again
	 bl	putcharnow
	 cmp	r1, #0x20		@ if char was a ctrl char then
	 bllt	putcharnow		@  back up on screen a second time
	 b	ch010			@ loop back for next keystroke
ch020:
	strb	r0, [r2, r3]		@ always store char into line buffer
	cmp	r0, #CR			@ echo control characters as "^char"
	beq	ch025			@ except CR, LF, EOF, and Tab
	cmp	r0, #LF
	beq	ch025
	cmp	r0, #EOF
	beq	ch025
	cmp	r0, #0x09
	beq	ch025
	cmp	r0, #0x20
	bge	ch025
	orr	r1, r0, #0x40		@ convert ctrl char to displayable char
	mov	r0, #'^'
	bl	putcharnow		@ and echo character to display
	mov	r0, r1
ch025:
	cmp	r2, #127		@ bump buffer index if it's not too big
	addlt	r2, #1
	bllt	putcharnow		@ and echo character to display

	cmp	r0, #LF			@ exit loop on Linux's "new-line"
	beq	ch029			@ (because of redirected input files)
	cmp	r0, #EOF		@ also exit loop on Ctrl+Z (EOF)
	beq	ch028			@  but first move to start of new line
	cmp	r0, #CR			@ loop until Enter key is struck
	bne	ch010
ch028:	mov	r0, #LF			@ echo CR as a "new-line"
	bl	putcharnow
ch029:
	mov	r0, #0			@ make device 6 pen position track
	str	r0, [r8, #penx-dseg]	@  device 0's cursor position
	ldr	r0, [r8, #peny-dseg]
	add	r0, #16			@ assume device 0 height is always 16
	str	r0, [r8, #peny-dseg]

	str	r2, [r8, #kbbufsiz-dseg] @ number of characters in the buffer
	mov	r2, #0			@ reset index to start of line buffer
ch030:
	ldr	r3, [r8, #kbbufptr-dseg] @ get character from line buffer
	ldrb	r0, [r2, r3]
	add	r2, #1			@ bump index
	str	r2, [r8, #kbinx-dseg]
	str	r0, [r8, #lastch-dseg]	@ save copy in case of BackUp intrinsic
	pop	{pc}			@ return character in r0

@-------------------------------------------------------------------------------
@ Output byte in al to console. (Can be redirected to a file using ">" on the
@  command line.)
@
conout:	push	{lr}
	cmp	r0, #0x0C		@ form feed?
	bne	co20			@ skip if not
	 adr	r0, clrmsg		@ else clear screen
	 bl	text
	 mov	r0, #0			@ also move pen to upper-left corner
	 str	r0, [r8, #penx-dseg]
	 str	r0, [r8, #peny-dseg]
	 b	co90
co20:
	cmp	r0, #LF			@ Linux newline (does CR and LF)?
	bne	co30			@ skip if not
	 mov	r1, #0			@ make device 6 pen position track
	 str	r1, [r8, #penx-dseg]	@  device 0's cursor position
	 ldr	r1, [r8, #peny-dseg]
@	 ldr	r2, [r8, #fontheight-dseg]
@	 add	r1, r2
	 add	r1, #16			@ assume device 0 height is always 16
	 str	r1, [r8, #peny-dseg]
co30:
	cmp	r0, #BEL		@ bell?
	bne	co40

@  SOUND(VOLUME, DURATION, PERIOD)
	mov	r0, #2			@ duration
	mov	r1, #1			@ volume = on
	push	{r0, r1}
	mov	r0, #1024		@ period
	bl	intr39
	b	co90
co40:
	bl	putcharnow		@ (r0 is preserved)

	ldr	r1, [r8, #penx-dseg]	@ make device 6's pen position track
	cmp	r0, #CR			@ if carriage return then
	moveq	r1, #0			@   move to start of line
	addne	r1, #8			@ else move to next character position
	cmp	r0, #0x09		@ if tab then
	subeq	r1, #8			@   undo addne above
	biceq	r1, #0x3F		@   truncate to previous tab position
	addeq	r1, #8*8		@   move to next tab position
	cmp	r0, #0x08		@ if backspace then
	subeq	r1, #8+8		@   back up and also undo addne above
	str	r1, [r8, #penx-dseg]
co90:
	pop	{pc}

clrmsg:	.byte	ESC			@ ANSI command to clear terminal screen
	.ascii	"[2J"
	.byte	ESC			@ and move cursor to home position
	.asciz	"[1;1H"
	.align	2

@===============================================================================
@ Discard any pending keyboard input.
@
openi1:	push	{lr}
	mov	r0, #0			@ discard any pending non-ASCII char
	str	r0, [r8, #scancode-dseg]

oi10:	bl	intr33			@ while keyhit do read key
	cmp	r0, #0
	popeq	{pc}

	mov	r0, #0			@ read(STDIN, &char, 1)
	push	{r0}			@ make temporary 0 space on top of stack
	mov	r1, sp			@ point r1 to it (for &char in low byte)
	mov	r2, #1			@ read in one character
	bl	read
	pop	{r0}			@ get character into r0
	b	oi10

@-------------------------------------------------------------------------------
@ Wait for then get a keystroke value from the keyboard. Don't echo it to the
@  display. Ctrl+C aborts. Non-ASCII keys return a zero followed by their scan
@  code. Input can be redirected from a file by using "<" on the command line.
@
chin1:	push	{lr}
@ if there is a scan code pending then return it, but first clear it
	ldr	r0, [r8, #scancode-dseg]
	cmp	r0, #0
	mov	r1, #0			@ clear scan code
	str	r1, [r8, #scancode-dseg]
	bne	chin190			@ go return scan code

	bl	readstd			@ get character from standard input dev
	cmp	r0, #0x00		@ if "keystroke" is a Ctrl+Break then
	moveq	r0, #EOF		@  change it to an EOF (Ctrl+Z); it's
					@  very handy for redirected input files
	cmp	r0, #0x7F		@ if keystroke is a Delete then
	moveq	r0, #0x08		@  change it back to the Backspace

	cmp	r0, #ESC		@ if keystroke is the Esc key
	bne	chin130			@   (skip if not)
	 bl	intr33			@ and if another key is pending then
	 cmp	r0, #0			@   (skip if not)
	 beq	chin120			@
	  bl	getscan			@   get the scan code
	  str	r0, [r8, #scancode-dseg]@   save it for next call to chin1
	  mov	r0, #0			@   return a zero this time
	  b	chin130
chin120: mov	r0, #ESC		@ restore the Esc key
chin130:
	str	r0, [r8, #lastch-dseg]	@ save copy in case of BackUp intrinsic
	cmp	r0, #0x03		@ is keystroke a Ctrl-C?
	bne	chin190			@ skip if not
	 ldr	r1, [r8, #ctcflag-dseg]	@ if Ctrl-C flag is 'false' then
	 cmp	r1, #0
	 beq	intr16			@  abort program
	 mov	r1, #-1			@ else set flag indicating Ctrl-C struck
	 str	r1, [r8, #cinflag-dseg]
chin190:
@ check for and handle possible Pause keystroke
	cmp	r0, #0			@ skip if not Pause key
	bne	chin199
	ldr	r1, [r8, #scancode-dseg]
	cmp	r1, #0x45
	bne	chin199

	mov	r0, #0			@ discard the Pause keystroke
	str	r0, [r8, #scancode-dseg]
	bl	chin1			@ wait for and discard next keystroke
	cmp	r0, #0
	mov	r0, #0			@ return 0 so it will be ignored
	bne	chin199

	mov	r0, #0			@ if next keystroke was a non-ASCII key
	str	r0, [r8, #scancode-dseg]@  eliminate it and return a 0
chin199:
	pop	{pc}			@ return with char in r0

@-------------------------------------------------------------------------------
@ Convert the Esc sequence that Linux generates for non-ASCII keys to its
@  equivalent PC scan code and return it in r0.
@
getscan:push	{r4, lr}
	mov	r4, #0			@   [Key:= 0;
gs10:	bl	intr33			@   while KeyHit do
	cmp	r0, #0			@     (branch if not)
	beq	gs20
	 lsl	r4, #8			@     Key:= Key<<8;
	 bl	readstd			@     Key:= STDIN
	 add	r4, r0
	 b	gs10
gs20:					@   for I:= 0 to $5B-$10 do
	adr	r2, scntbl		@     if Key = Tbl(I) then
	mov	r0, #0
gs25:	ldr	r1, [r2, r0, lsl #2]
	cmp	r4, r1
	bne	gs30
	 add	r0, #0x10		@       return I + $10;
	 b	gs99
gs30:	add	r0, #1
	cmp	r0, #0x5B-0x10
	ble	gs25

	mov	r0, #0			@   return 0; (no scan code found)
gs99:	pop	{r4, pc}

scntbl:	.word	0x00000071		@ 10 Alt-Q
	.word	0x00000077		@ 11 Alt-W
	.word	0x00000065		@ 12 Alt-E
	.word	0x00000072		@ 13 Alt-R
	.word	0x00000074		@ 14 Alt-T
	.word	0x00000079		@ 15 Alt-Y
	.word	0x00000075		@ 16 Alt-U
	.word	0x00000069		@ 17 Alt-I
	.word	0x0000006F		@ 18 Alt-O
	.word	0x00000070		@ 19 Alt-P
	.word	0x00000000		@ 1A -
	.word	0x00000000		@ 1B -
	.word	0x00000000		@ 1C -
	.word	0x00000000		@ 1D -
	.word	0x00000061		@ 1E Alt-A
	.word	0x00000073		@ 1F Alt-S
	.word	0x00000064		@ 20 Alt-D
	.word	0x00000066		@ 21 Alt-F
	.word	0x00000067		@ 22 Alt-G
	.word	0x00000068		@ 23 Alt-H
	.word	0x0000006A		@ 24 Alt-J
	.word	0x0000006B		@ 25 Alt-K
	.word	0x0000006C		@ 26 Alt-L
	.word	0x00000000		@ 27 -
	.word	0x00000000		@ 28 -
	.word	0x00000000		@ 29 -
	.word	0x00000000		@ 2A -
	.word	0x00000000		@ 2B -
	.word	0x0000007A		@ 2C Alt-Z
	.word	0x00000078		@ 2D Alt-X
	.word	0x00000063		@ 2E Alt-C
	.word	0x00000076		@ 2F Alt-V
	.word	0x00000062		@ 30 Alt-B
	.word	0x0000006E		@ 31 Alt-N
	.word	0x0000006D		@ 32 Alt-M
	.word	0x00000000		@ 33 -
	.word	0x00000000		@ 34 -
	.word	0x00000000		@ 35 -
	.word	0x00000000		@ 36 -
	.word	0x00000000		@ 37 -
	.word	0x00000000		@ 38 -
	.word	0x00000000		@ 39 -
	.word	0x00000000		@ 3A -
	.word	0x005B5B41		@ 3B F1
	.word	0x005B5B42		@ 3C F2
	.word	0x005B5B43		@ 3D F3
	.word	0x005B5B44		@ 3E F4
	.word	0x005B5B45		@ 3F F5
	.word	0x5B31377E		@ 40 F6
	.word	0x5B31387E		@ 41 F7
	.word	0x5B31397E		@ 42 F8
	.word	0x5B32307E		@ 43 F9
	.word	0x5B32317E		@ 44 F10
	.word	0x00005B50		@ 45 Pause
	.word	0x00000000		@ 46 -
	.word	0x005B317E		@ 47 Home
	.word	0x00005B41		@ 48 Up arrow
	.word	0x005B357E		@ 49 Page Up
	.word	0x00000000		@ 4A -
	.word	0x00005B44		@ 4B Left arrow
	.word	0x00005B47		@ 4C Center "5"
	.word	0x00005B43		@ 4D Right arrow
	.word	0x00000000		@ 4E -
	.word	0x005B347E		@ 4F End
	.word	0x00005B42		@ 50 Down arrow
	.word	0x005B367E		@ 51 Page Down
	.word	0x005B327E		@ 52 Insert
	.word	0x005B337E		@ 53 Delete
	.word	0x5B32357E		@ 54 Shift-F1
	.word	0x5B32367E		@ 55 Shift-F2
	.word	0x5B32387E		@ 56 Shift-F3
	.word	0x5B32397E		@ 57 Shift-F4
	.word	0x5B33317E		@ 58 Shift-F5
	.word	0x5B33327E		@ 59 Shift-F6
	.word	0x5B33337E		@ 5A Shift-F7
	.word	0x5B33347E		@ 5B Shift-F8

@-------------------------------------------------------------------------------
@ Read (get) character from standard input device and return it in r0.
@
readstd:push	{lr}
	mov	r0, #0			@ read(STDIN, &char, 1)
	push	{r0}			@ make temporary 0 space on top of stack
	mov	r1, sp			@ point r1 to it (for &char in low byte)
	mov	r2, #1			@ read in one character
	bl	read
	pop	{r0, pc}		@ return with char in r0

@===============================================================================

	.data
	.align	2
bufsiz	=	1024			@ I/O buffer sizes (bytes)
infd:		.word	0
outfd:		.word	0
inbufflg:	.word	0
outbufflg:	.word	0
inbufsiz:	.word	0
outbufsiz:	.word	0
ininx:		.word	0
outinx:		.word	0
charin:		.word	0		@ last character read from input file
inbufptr:	.word	inbuf
outbufptr:	.word	outbuf
	.bss
inbuf:	.skip	bufsiz			@ large input buffer for file
outbuf:	.skip	bufsiz			@ large output buffer for file
	.text

@-------------------------------------------------------------------------------
@ Initialize file for input.
@ OPENI(3)
@
openi3:	push	{lr}
	ldr	r0, [r8, #inbufflg-dseg]@ if inbufflg = 'I' then
	cmp	r0, #'I'
	bne	oi310
	 mov	r0, #bufsiz		@   inbufsiz = bufsiz
	 str	r0, [r8, #inbufsiz-dseg]
	 str	r0, [r8, #ininx-dseg]	@   ininx = inbufsiz
oi310:
	ldr	r0, [r8, #infd-dseg]	@ lseek(infd, 0, SEEK_SET)
	mov	r1, #0
	mov	r2, #0
	bl	lseek
	pop	{pc}

@-------------------------------------------------------------------------------
@ OPENO(3)
@
openo3:	push	{lr}
	ldr	r0, [r8, #outbufflg-dseg]@ if outbufflg = 'O' then
	cmp	r0, #'O'
	bne	oo310
	 mov	r0, #bufsiz		@   outbufsiz = bufsiz
	 str	r0, [r8, #outbufsiz-dseg]
	 mov	r0, #0
	 str	r0, [r8, #outinx-dseg]	@   outinx = 0
oo310:
	ldr	r0, [r8, #outfd-dseg]	@ lseek(outfd, 0, SEEK_SET)
	mov	r1, #0
	mov	r2, #0
	bl	lseek
	pop	{pc}

@-------------------------------------------------------------------------------
@ Read a byte from the input file assigned to device 3 (by FSET). Reading beyond
@ the end of file is normally an error, but the first attempt instead returns an
@ EOF character. Any additional attempts are flagged as an error.
@ CHIN(3)
@
chin3:	push	{lr}
	ldr	r0, [r8, #inbufflg-dseg]@ if (inbufflg == 'I')
	cmp	r0, #'I'
	bne	ci340

	mov	r0, #1			@    assume something is in the buffer
	ldr	r3, [r8, #ininx-dseg]	@    if (ininx >= inbufsiz)
	ldr	r2, [r8, #inbufsiz-dseg]
	cmp	r3, r2
	blt	ci330

	 ldr	r0, [r8, #infd-dseg]	@     n = read(infd, inbufptr, inbufsiz)
	 ldr	r1, [r8, #inbufptr-dseg]
	 bl	read			@     (r2 = inbufsiz)
	 str	r0, [r8, #inbufsiz-dseg]@     inbufsiz = n
	 mov	r1, #0
	 str	r1, [r8, #ininx-dseg]	@     ininx = 0
ci330:
	ldr	r2, [r8, #ininx-dseg]	@    ch = inbuf[ininx++]
	ldr	r3, [r8, #inbufptr-dseg]
	ldrb	r1, [r2, r3]
	add	r2, #1
	str	r2, [r8, #ininx-dseg]
	str	r1, [r8, #charin-dseg]
	b	ci350
ci340:
	ldr	r0, [r8, #infd-dseg]	@ else n = read(infd, &ch, 1);
	push	{r1}
	mov	r1, sp
	mov	r2, #1
	bl	read
	pop	{r1}
	str	r1, [r8, #charin-dseg]
ci350:
	cmp	r0, #0			@ if (n == 0)
	bne	ci390
	ldr	r0, [r8, #charin-dseg]	@   if (ch == EOF) Error(3);
	cmp	r0, #EOF
	bne	ci380
	 mov	r0, #3
	 bl	error
ci380:	mov	r0, #EOF		@   ch = EOF;
	str	r0, [r8, #charin-dseg]
ci390:
	ldr	r0, [r8, #charin-dseg]
	str	r0, [r8, #lastch-dseg]
	pop	{pc}			@ return ch (byte) in r0

@-------------------------------------------------------------------------------
@ Output the byte in r0 to the device 3 file.
@ CHOUT(3,CH)
@
chout3:	push	{lr}
	ldr	r1, [r8, #outbufflg-dseg] @ if (outbufflg == 'O')
	cmp	r1, #'O'
	bne	co350

	ldr	r2, [r8, #outbufptr-dseg] @   outbuf[outinx++] = ch;
	ldr	r3, [r8, #outinx-dseg]
	strb	r0, [r3, r2]
	add	r3, #1
	str	r3, [r8, #outinx-dseg]

	ldr	r2, [r8, #outbufsiz-dseg] @   if (outinx >= outbufsiz)
	cmp	r3, r2
	blt	co390

	ldr	r0, [r8, #outfd-dseg]	  @  n = write(outfd, outbuf, outbufsiz)
	ldr	r1, [r8, #outbufptr-dseg]
	bl	write
	mov	r1, #0			  @  outinx = 0;
	str	r1, [r8, #outinx-dseg]
	ldr	r1, [r8, #outbufsiz-dseg] @  if (n != outbufsiz) Error(3);
	cmp	r0, r1
	beq	co390
	b	co380
co350:
	push	{r0}			@ else	n = write(outfd, &ch, 1)
	ldr	r0, [r8, #outfd-dseg]
	mov	r1, sp
	mov	r2, #1
	bl	write
	pop	{r1}
	cmp	r0, #1			@ if (n != 1) Error(3);
	beq	co390
co380:	 mov	r0, #3
	 bl	error
co390:	pop	{pc}

@-------------------------------------------------------------------------------
@ CLOSE(3)
@
close3:	push	{lr}
	ldr	r0, [r8, #outbufflg-dseg] @ if outbufflg = 'O' then
	cmp	r0, #'O'
	bne	cl390

	ldr	r2, [r8, #outinx-dseg]	@   if outinx # 0 then
	cmp	r2, #0
	beq	cl390

	ldr	r0, [r8, #outfd-dseg]	@     n = write(outfd, outbuf, outinx)
	ldr	r1, [r8, #outbufptr-dseg]
	bl	write

	ldr	r2, [r8, #outinx-dseg]	@     if n # outinx then Error(3)
	cmp	r0, r2
	beq	cl380
	 mov	r0, #3
	 bl	error
	 pop	{pc}

cl380:	mov	r0, #0			@     outinx:= 0
	str	r0, [r8, #outinx-dseg]
cl390:	pop	{pc}

@===============================================================================
@ Discard any pending keyboard input.
@  OPENI(6)
openi6:	b	openi1

@-------------------------------------------------------------------------------
@ Set attribute for white characters on black background, move cursor to upper-
@  left corner of screen, enable scroll mode, and open window to full screen.
@  OPENO(6)
@
openo6:	ldr	r0, [r8, #scrdepth-dseg]
	cmp	r0, #16			@ set white letters on black for dev 6
	mov	r0, #0x00000007		@ 8 & 16-bit graphics use palette colors
	movgt	r0, #0x000000AA		@ 24- or 32-bit graphics R8:G8:B8
	orrgt	r0, #0x0000AA00
	orrgt	r0, #0x00AA0000
	orrgt	r0, #0xFF000000		@ full brightness
	str	r0, [r8, #attrib-dseg]

	mov	r0, #0
	str	r0, [r8, #penx-dseg]	@ move pen to upper-left corner and
	str	r0, [r8, #peny-dseg]
	str	r0, [r8, #windmode-dseg] @ enable scroll mode
					@ fall into fullwind
@-------------------------------------------------------------------------------
@ Set device 6's window to full screen size.
@
fullwind:
	push	{lr}
	mov	r0, #0
	str	r0, [r8, #windx0-dseg]
	str	r0, [r8, #windy0-dseg]

	ldr	r0, [r8, #scrwidth-dseg] @ windx1:= scrwidth/8 - 1
	lsr	r0, #3
	sub	r0, #1
	str	r0, [r8, #windx1-dseg]

	ldr	r0, [r8, #scrheight-dseg]
	ldr	r1, [r8, #fontheight-dseg]
	bl	udiv
	sub	r0, #1
	str	r0, [r8, #windy1-dseg]	@ windy1:= scrheight/fontheight - 1
	pop	{pc}

@-------------------------------------------------------------------------------
@ Wait for then get a keystroke value from the keyboard. Echo it to the display.
@ Any special character handling is disabled.
@
chin6:	push	{lr}
	mov	r0, #0			@ read(STDIN, &char, 1); get keystroke
	push	{r0}			@ make temporary 0 space on top of stack
	mov	r1, sp			@ point r1 to it (for &char in low byte)
	mov	r2, #1			@ read in one character
	bl	read
	pop	{r0}			@ get character into r0
	cmp	r0, #0x7F		@ if keystroke is a Delete then
	moveq	r0, #0x08		@  change it back to a Backspace
	str	r0, [r8, #lastch-dseg]	@ save copy in case of backup
	bl	chout6			@ echo it on the graphic display
	ldr	r0, [r8, #lastch-dseg]
	pop	{pc}			@ return with character in r0

@-------------------------------------------------------------------------------
@ Display character in r0 on screen at current cursor location. This outputs
@  the character immediately (whereas putchar buffers a line and only outputs
@  characters to the screen after a line is completed).
@
putcharnow:
	push	{lr}
	push	{r0-r3}			@ preserve registers
	push	{r0}			@ use top of stack for data location
	mov	r0, #1			@ write(STDOUT, &c, 1)
	mov	r1, sp			@ point to character to output
	mov	r2, #1			@ output one byte
	bl	write
	pop	{r0}
	pop	{r0-r3}			@ restore character in r0, etc.
	pop	{pc}

@-------------------------------------------------------------------------------
@ Display character in r0 at penx,peny.
@  Registers:
@   r4: char
@   r5: X coordinate
@   r6: Y coordinate
@   r7: byte from font table
@   r9: foreground color attribute
@   r10: background color attribute
@   r11: font height (pixels)
@   r12: address of font table
@
chout6:	push	{r4-r7, r9-r12, lr}	@ save registers because of bl call
	ldrb	r1, [r8, #nowdev+1-dseg] @ check high byte of device number
	cmp	r1, #0			@ adjust device $006's font height to be
	bne	cho601			@  compatible with the screen height
	ldr	r2, [r8, #scrheight-dseg]
	cmp	r2, #352		@ if screenheight = 350, use 8x14 font
	movle	r1, #2
	cmp	r2, #200		@ if screenheight = 200, use 8x8 font
	movle	r1, #1			@ (this makes it agree with the PC)
cho601:
	cmp	r1, #1
	ldrlt	r11, [r8, #fontheight6-dseg] @ device $006's font is adjustable
	moveq	r11, #8			@ device $106's font is 8 pixels high
	movgt	r11, #14		@ device $206's font is 14 pixels high
	str	r11, [r8, #fontheight-dseg]
	ldrlt	r12, [r8, #pfont16-dseg]@ point to table corresponding to height
	ldreq	r12, [r8, #pfont8-dseg]
	ldrgt	r12, [r8, #pfont14-dseg]

	ldr	r9, [r8, #attrib-dseg]	@ set foreground and background colors
	mov	r10, r9
	ldr	r1, [r8, #scrdepth-dseg]@	r9	r10
	cmp	r1, #8			@ depth	fgnd	bgnd
	andlt	r9, #0x0F		@  0	a&F	a>>4&F
	andlt	r10, #0xF0		@  4	a&F	a>>4&F
	lsrlt	r10, #4			@  8	a&FF	a>>8
	lsreq	r10, #8			@ 16	a&FFFF	a>>16
	lsrgt	r10, #16		@ 24	a	0
	andeq	r9, #0xFF		@ 32	a	0
	bge	cho602
	 cmp	r9, r10			@ depth < 8
	 orreq	r9, #0x80		@ use exclusive-or graphics
	 orreq	r10, #0x80
cho602:	cmp	r1, #16
	biceq	r9, #0xFF000000
	biceq	r9, #0x00FF0000
	cmp	r1, #24			@ set background to black (0) for
	movge	r10, #0			@  depths >= 24

	and	r0, #0xFF
	cmp	r0, #CR			@ if carriage return then
	bne	cho605
	 ldr	r0, [r8, #windx0-dseg]	@   penx:= windx0 * 8
	 lsl	r0, #3
	 str	r0, [r8, #penx-dseg]
	 b	cho690
cho605:
	cmp	r0, #LF			@ if LF then do newline and exit
	bne	cho607
	 bl	donl
	 b	cho690
cho607:					@ check if outside window
	push	{r0}			@ save character
	ldr	r0, [r8, #penx-dseg]	@ if penx > windx1*8 then
	ldr	r1, [r8, #windx1-dseg]
	lsl	r1, #3
	cmp	r0, r1
	ble	cho608
	 ldr	r0, [r8, #windmode-dseg] @   if windmode = clip then
	 cmp	r0, #2
	 popeq	{r0}			@      exit
	 beq	cho690
	 cmp	r0, #1			@   else if windmode = wrap then
	  ldreq	r0, [r8, #windx0-dseg]	@     do carriage reutrn
	  lsleq	r0, #3			@     i.e. move to beginning of window
	  streq	r0, [r8, #penx-dseg]	@     penx:= windx0 * 8
	  beq	cho608
	 bl	donl			@    else do newline
cho608:
	ldr	r0, [r8, #peny-dseg]	@ if peny > windy1*fontheight then
	ldr	r1, [r8, #windy1-dseg]
	mul	r1, r11
	cmp	r0, r1
	ble	cho609
	 ldr	r0, [r8, #windmode-dseg] @   if windmode = clip then
	 cmp	r0, #2
	 popeq	{r0}			@      exit
	 beq	cho690
cho609:
	pop	{r0}			@ display character
	mov	r4, r0			@ save character in r4
	ldr	r0, [r8, #penx-dseg]	@ save position of upper-left corner
	str	r0, [r8, #penx0-dseg]	@  of character cell
	ldr	r0, [r8, #peny-dseg]
	str	r0, [r8, #peny0-dseg]

	mov	r6, #0			@ for Y:= 0 to Height-1 do
cho610:	mov	r0, r12			@   byte:= fonttbl(char*height+Y)
	mla	r1, r4, r11, r6
	ldrb	r7, [r1, r0]

	cmp	r9, r10			@ if foregnd and backgnd colors are the
	beq	cho625			@  same then go do transparent backgnd

	mov	r5, #0			@   for X:= 0 to 7 do
cho620:	ldr	r2, [r8, #penx0-dseg]	@     Point(X+penx0, Y+peny0, ...
	add	r2, r5
	ldr	r1, [r8, #peny0-dseg]
	add	r1, r6
	mov	r0, r10			@     assume background color
	tst	r7, #0x80		@     if byte & $80 then
	movne	r0, r9			@       use foreground color
	bl	intr41a			@     plot point
	lsl	r7, #1			@     byte << 1
	add	r5, #1			@   next X
	cmp	r5, #7
	ble	cho620
	b	cho650

cho625:	mov	r5, #0			@   for X:= 0 to 7 do
cho630:	ldr	r2, [r8, #penx0-dseg]	@     Point(X+penx0, Y+peny0, ...
	add	r2, r5
	ldr	r1, [r8, #peny0-dseg]
	add	r1, r6
	mov	r0, r9			@     get foreground color
	tst	r7, #0x80		@     if byte & $80 then plot point
	blne	intr41a			@      (background is transparent)
	lsl	r7, #1			@     byte << 1
	add	r5, #1			@   next X
	cmp	r5, #7
	ble	cho630
cho650:
	add	r6, #1			@ next Y
	cmp	r6, r11
	blt	cho610

	ldr	r0, [r8, #penx0-dseg]	@ move cursor to next character cell
	add	r0, #8
	str	r0, [r8, #penx-dseg]
	ldr	r0, [r8, #peny0-dseg]
	str	r0, [r8, #peny-dseg]

	ldr	r0, [r8, #penx-dseg]	@ if penx > windx1*8 then
	ldr	r1, [r8, #windx1-dseg]
	lsl	r1, #3
	cmp	r0, r1
	ble	cho690
	 ldr	r0, [r8, #windmode-dseg] @   if windmode = scroll then
	 cmp	r0, #0
	 bleq	donl			@      scroll window if necessary
cho690:	pop	{r4-r7, r9-r12, pc}	@ restore registers and return

@ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@ Handle line feed for device 6
@
donl:	push	{lr}
	ldr	r0, [r8, #windx0-dseg]	@ do Linux newline
	lsl	r0, #3			@ move to beginning of line
	str	r0, [r8, #penx-dseg]	@ penx:= windx0 * 8
	ldr	r0, [r8, #peny-dseg]	@ move down one text line
	add	r0, r11			@ peny:= peny + fontheight
	str	r0, [r8, #peny-dseg]

@ for y:= windy0*fontheight to windy1*fontheight-1 do
	ldr	r7, [r8, #windy1-dseg]	@ limity:= windy1 * fontheight
	mul	r7, r7, r11
	cmp	r0, r7			@ if peny <= limity then return
	ble	donl90

@ moved off bottom of window
	ldr	r1, [r8, #windmode-dseg]
	cmp	r1, #1			@ if windmode = wrap then
	bne	scr00
	 ldr	r0, [r8, #windy0-dseg]	@   move to beginning of top line
	 mul	r0, r0, r11		@   peny:= windy0 * fontheight
	 str	r0, [r8, #peny-dseg]
	 b	donl90			@   return
scr00:
	cmp	r1, #0			@ if windmode # scroll then
	bne	donl90			@   do nothing (clip) return

@ scroll text in window up one line
	ldr	r6, [r8, #windy0-dseg]	@ y:= windy0 * fontheight
	mul	r6, r6, r11

@   for x:= windx0*8 to (windx1+1)*8-1 do
	ldr	r4, [r8, #windx1-dseg]	@ limitx:= (windx1+1) * 8
	add	r4, #1
	lsl	r4, #3

scr10:	ldr	r5, [r8, #windx0-dseg]	@ x:= windx0 * 8
	lsl	r5, #3

@     Point(x, y, ReadPix(x, y+fontheight));
scr20:	add	r0, r6, r11		@ y + fontheight;
	push	{r5}			@ x
	bl	intr44			@ ReadPix
	mov	r2, r5			@ x
	mov	r1, r6			@ y
	bl	intr41a			@ Point

	add	r5, #1			@ next x
	cmp	r5, r4
	blt	scr20

	add	r6, #1			@ next y
	cmp	r6, r7
	blt	scr10

@ blank bottom line of text in the window by filling it with background color
@ for y:= windy1*fontheight to (windy1+1)*fontheight-1 do
	mov	r6, r7			@ y:= windy1 * fontheight
	add	r7, r11			@ limity:= (windy1+1) * fontheight
	mov	r0, r10			@ background color
@   for x:= windx0*8 to (windx1+1)*8-1 do
scr30:	ldr	r5, [r8, #windx0-dseg]	@ x:= windx0 * 8
	lsl	r5, #3
					@ r4 = limitx = (windx1+1) * 8
@     Point(x, y, bkgndcolor);
scr40:	mov	r2, r5			@ x
	mov	r1, r6			@ y
	bl	intr41a			@ Point

	add	r5, #1			@ next x
	cmp	r5, r4
	blt	scr40

	add	r6, #1			@ next y
	cmp	r6, r7
	blt	scr30

@ set cursor to beginning of bottom line in window
	ldr	r0, [r8, #windx0-dseg]	@ penx:= windx0 * 8
	lsl	r0, #3
	str	r0, [r8, #penx-dseg]
	ldr	r0, [r8, #windy1-dseg]	@ peny:= windy1 * fontheight
	mul	r0, r0, r11
	str	r0, [r8, #peny-dseg]
donl90:	pop	{pc}			@ return

	.data
	.align	2
pfont16:.word	font16			@ pointer to 8x16 font table
fontheight6: .word	16		@ adjustable font height (pixels) 
fontheight: .word	16		@ font height (pixels) for current font

pfont8:	.word	font8
pfont14:.word	font14
penx0:	.word	0			@ initial position of pen
peny0:	.word	0
	.text

@-------------------------------------------------------------------------------
@ Intrinsic to return the address of the built-in 256-character font tables.
@  ADDRESS:= GETFONT(FACE)		\FACE: 0=8x16, 1=8x8, 2=8x14
@
intr91:	cmp	r0, #1
	adrlt	r0, font16
	ldreq	r0, [r8, #pfont8-dseg]
	ldrgt	r0, [r8, #pfont14-dseg]
	mov	pc, lr

	.include	"font16.s"

@-------------------------------------------------------------------------------
@ Intrinsic to set the height and address of a font table for device 6.
@  SETFONT(HEIGHT, ADDRESS)
@
intr92:	str	r0, [r8, #pfont16-dseg]
	pop	{r0}
	str	r0, [r8, #fontheight6-dseg]
	str	r0, [r8, #fontheight-dseg]
	mov	pc, lr

	.include	"font8.s"
	.include	"font14.s"

@===============================================================================
@ Pseudo device handler (device 8).
@ The pseudo device is a circular memory buffer that can be read from or written
@ to. It can be used to parse strings and to convert numbers to ASCII strings.
@
	.data
	.align	2
pssize	=	256			@ size of buffer in bytes
pswrite:.word	0			@ write index
psstart:.word	0			@ start index
psread:	.word	0			@ read index into psbuf
psbuf:	.word	pbuf
pbuf:	.skip	pssize			@ buffer
	.text

@-------------------------------------------------------------------------------
@ Open pseudo device for output.
@
psopno:	mov	r0, #0
	str	r0, [r8, #pswrite-dseg]	@ zero write index
	str	r0, [r8, #psstart-dseg]	@ zero start index
					@ (pfall) also open input
@-------------------------------------------------------------------------------
@ Open pseudo device for input.
@
psopni:	ldr	r0, [r8, #psstart-dseg]	@ set read index to start
	str	r0, [r8, #psread-dseg]
	mov	pc, lr			@ return

@-------------------------------------------------------------------------------
@ Input a byte from the pseudo device buffer and return it in r0.
@
psin:	ldr	r1, [r8, #psread-dseg]	@ get read index
	ldr	r2, [r8, #pswrite-dseg]	@ if read = write then buffer is empty
	cmp	r1, r2
	bne	psin10			@ skip if more to read
	 mov	r0, #EOF		@ signal end of file
	 b	psin90			@ exit
psin10:
	ldr	r2, [r8, #psbuf-dseg]	@ get char from buffer at index
	ldrb	r0, [r1, r2]		@ (high bytes are zeroed)
	add	r1, #1			@ bump read index
	cmp	r1, #pssize		@ wrap around?
	blt	psin20			@ skip if not
	 mov	r1, #0			@ start at beginning
psin20:	str	r1, [r8, #psread-dseg]
psin90:	str	r0, [r8, #lastch-dseg]
	mov	pc, lr			@ return with byte in r0

@-------------------------------------------------------------------------------
@ Output the byte in r0 to the pseudo device buffer.
@
psout:	ldr	r1, [r8, #pswrite-dseg]	@ store char into buffer at index
	ldr	r2, [r8, #psbuf-dseg]
	strb	r0, [r1, r2]
	add	r1, #1			@ bump index
	cmp	r1, #pssize		@ wrap around?
	blt	psout10			@ skip if not
	 mov	r1, #0			@ start at beginning
psout10:str	r1, [r8, #pswrite-dseg]

	ldr	r2, [r8, #psstart-dseg]	@ caught start index?
	cmp	r1, r2
	bne	psout20			@ skip if not
	 add	r2, #1			@ bump start index
	 cmp	r2, #pssize		@ wrap around?
	 blt	psout20			@ skip if not
	  mov	r2, #0			@ start at beginning
psout20:str	r2, [r8, #psstart-dseg]

	ldr	r2, [r8, #psread-dseg]	@ caught read index?
	cmp	r1, r2
	bne	psout30			@ skip if not
	 add	r2, #1			@ increment read index
	 cmp	r2, #pssize		@ wrap around?
	 blt	psout30			@ skip if not
	  mov	r2, #0			@ start at beginning
psout30:str	r2, [r8, #psread-dseg]
	mov	pc, lr			@ return

@-------------------------------------------------------------------------------

	.if	!natrpi		@ debug
	.bss
	.align	4
heaplo:	.skip	0x4000000		@ 64 MB of heap space
	.endif
